//
// C++ Interface: dbobj
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2008, 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef _DBOBJ_H
#define _DBOBJ_H

#include <iostream>
#include <list>
#include <glibmm.h>
#include <gdkmm.h>
#include <sqlite3x.hpp>
#include "geom.h"
#include "fplan.h"

class Conversions {
        public:
                static Glib::ustring time_str(time_t t);
                static Glib::ustring time_str(const Glib::ustring& fmt, time_t t);
                static time_t time_parse(const Glib::ustring& str);
                static Glib::ustring timefrac8_str(uint64_t t);
                static Glib::ustring timefrac8_str(const Glib::ustring& fmt, uint64_t t);
                static uint64_t timefrac8_parse(const Glib::ustring& str);
                static Glib::ustring freq_str(uint32_t f);
                static Glib::ustring alt_str(int32_t a, uint16_t flg);
                static Glib::ustring dist_str(float d);
                static Glib::ustring track_str(float t);
                static Glib::ustring velocity_str(float v);
                static Glib::ustring verticalspeed_str(float v);
                static uint32_t unsigned_feet_parse(const Glib::ustring& str, bool *valid = 0);
                static int32_t signed_feet_parse(const Glib::ustring& str, bool *valid = 0);
};

class DbBaseElements {
        public:
                class ElementAddress {
                        public:
                                typedef enum {
                                        table_main = 0,
                                        table_aux = 1
                                } table_t;

                                ElementAddress(uint64_t id = 0, table_t tbl = table_main) : m_id(id), m_table(tbl) {}
                                uint64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                bool operator==(const ElementAddress& a2) const { return get_id() == a2.get_id() && get_table() == a2.get_table(); }
                                bool operator!=(const ElementAddress& a2) const { return !operator==(a2); }
                                operator bool(void) const { return !!m_id; }

                        protected:
                                uint64_t m_id;
                                table_t m_table;
                };

                class TileNumber {
                        public:
                                typedef unsigned int tile_t;
                                typedef std::pair<tile_t,tile_t> tilerange_t;
                                static tile_t to_tilenumber(const Point& pt);
                                static tile_t to_tilenumber(const Rect& r);
                                static tilerange_t to_tilerange(const Rect& r);

                                static const unsigned int tile_bits = 16;

                                class Iterator {
                                        public:
                                                Iterator(const Rect& r);
                                                tile_t operator()(void) const;
                                                bool next(void);

                                        protected:
                                                tile_t m_tile1, m_tile2, m_tile;
                                                unsigned int m_bits;
                                };

                        protected:
                                static const Point::coord_t lon_offset = 0xf4000000;
                };

                class Navaid {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef enum {
                                        navaid_invalid,
                                        navaid_vor,
                                        navaid_vordme,
                                        navaid_dme,
                                        navaid_ndb
                                } navaid_type_t;

                                typedef enum {
                                        label_n,
                                        label_e,
                                        label_s,
                                        label_w,
                                        label_center,
                                        label_any,
                                        label_off
                                } label_placement_t;

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_all = subtables_none;

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Navaid(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all) {}
                                unsigned int get_subtables(void) const { return ~0; }
                                bool is_subtables_loaded(void) const { return true; }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                Glib::ustring get_sourceid(void) const { return m_sourceid; }
                                void set_sourceid(const Glib::ustring& id) { m_sourceid = id; }
                                Point get_coord(void) const { return m_coord; }
                                void set_coord(const Point& pt) { m_coord = pt; }
                                int32_t get_elev(void) const { return m_elev; }
                                void set_elev(int32_t e) { m_elev = e; }
                                int32_t get_range(void) const { return m_range; }
                                void set_range(int32_t rng = 0) { m_range = rng; }
                                uint32_t get_frequency(void) const { return m_freq; }
                                void set_frequency(uint32_t f = 0) { m_freq = f; }
                                Glib::ustring get_icao(void) const { return m_icao; }
                                void set_icao(const Glib::ustring& ic) { m_icao = ic; }
                                Glib::ustring get_name(void) const { return m_name; }
                                void set_name(const Glib::ustring& nm) { m_name = nm; }
                                Glib::ustring get_icao_name(void) const;
                                navaid_type_t get_navaid_type(void) const { return m_navaid_type; }
                                void set_navaid_type(navaid_type_t nt = navaid_invalid) { m_navaid_type = nt; }
                                static Glib::ustring get_navaid_typename(navaid_type_t nt);
                                Glib::ustring get_navaid_typename(void) const { return get_navaid_typename(m_navaid_type); }
                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                void set_label_placement(label_placement_t lp = label_off) { m_label_placement = lp; }
                                static Glib::ustring get_label_placement_name(label_placement_t lp);
                                Glib::ustring get_label_placement_name(void) const { return get_label_placement_name(m_label_placement); }
                                bool is_valid(void) const { return m_navaid_type != navaid_invalid; }
                                bool is_inservice(void) const { return m_inservice; }
                                void set_inservice(bool is = false) { m_inservice = is; }
                                time_t get_modtime(void) const { return m_modtime; }
                                void set_modtime(time_t t = 0) { m_modtime = t; }

                        protected:
                                int64_t m_id;
                                Point m_coord;
                                int32_t m_elev;
                                int32_t m_range;
                                uint32_t m_freq;
                                Glib::ustring m_sourceid;
                                Glib::ustring m_icao;
                                Glib::ustring m_name;
                                navaid_type_t m_navaid_type;
                                label_placement_t m_label_placement;
                                bool m_inservice;
                                table_t m_table;
                                time_t m_modtime;
                };

                class Waypoint {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef ::DbBaseElements::Navaid::label_placement_t label_placement_t;
                                static const label_placement_t label_n = ::DbBaseElements::Navaid::label_n;
                                static const label_placement_t label_e = ::DbBaseElements::Navaid::label_e;
                                static const label_placement_t label_s = ::DbBaseElements::Navaid::label_s;
                                static const label_placement_t label_w = ::DbBaseElements::Navaid::label_w;
                                static const label_placement_t label_center = ::DbBaseElements::Navaid::label_center;
                                static const label_placement_t label_any = ::DbBaseElements::Navaid::label_any;
                                static const label_placement_t label_off = ::DbBaseElements::Navaid::label_off;

                                static const char usage_invalid = 0;
                                static const char usage_highlevel = 'H';
                                static const char usage_lowlevel = 'L';
                                static const char usage_bothlevel = 'B';
                                static const char usage_rnav = 'R';
                                static const char usage_terminal = 'T';
                                static const char usage_vfr = 'V';
                                static const char usage_user = 'U';

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_all = subtables_none;

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Waypoint(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all) {}
                                unsigned int get_subtables(void) const { return ~0; }
                                bool is_subtables_loaded(void) const { return true; }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                Glib::ustring get_sourceid(void) const { return m_sourceid; }
                                void set_sourceid(const Glib::ustring& id) { m_sourceid = id; }
                                Point get_coord(void) const { return m_coord; }
                                void set_coord(const Point& pt) { m_coord = pt; }
                                Glib::ustring get_icao(void) const { return m_icao; }
                                void set_icao(const Glib::ustring& ic) { m_icao = ic; }
                                Glib::ustring get_name(void) const { return m_name; }
                                void set_name(const Glib::ustring& nm) { m_name = nm; }
                                Glib::ustring get_icao_name(void) const;
                                char get_usage(void) const { return m_usage; }
                                void set_usage(char u) { m_usage = u; }
                                static Glib::ustring get_usage_name(char usage);
                                Glib::ustring get_usage_name(void) const { return get_usage_name(m_usage); }
                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                void set_label_placement(label_placement_t lp = label_off) { m_label_placement = lp; }
                                static Glib::ustring get_label_placement_name(label_placement_t lp) { return DbBaseElements::Navaid::get_label_placement_name(lp); }
                                Glib::ustring get_label_placement_name(void) const { return get_label_placement_name(m_label_placement); }
                                bool is_valid(void) const { return !!m_usage; }
                                time_t get_modtime(void) const { return m_modtime; }
                                void set_modtime(time_t t = 0) { m_modtime = t; }

                        protected:
                                int64_t m_id;
                                Point m_coord;
                                Glib::ustring m_sourceid;
                                Glib::ustring m_icao;
                                Glib::ustring m_name;
                                char m_usage;
                                label_placement_t m_label_placement;
                                table_t m_table;
                                time_t m_modtime;
                };


                class Airway {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef ::DbBaseElements::Navaid::label_placement_t label_placement_t;
                                static const label_placement_t label_n = ::DbBaseElements::Navaid::label_n;
                                static const label_placement_t label_e = ::DbBaseElements::Navaid::label_e;
                                static const label_placement_t label_s = ::DbBaseElements::Navaid::label_s;
                                static const label_placement_t label_w = ::DbBaseElements::Navaid::label_w;
                                static const label_placement_t label_center = ::DbBaseElements::Navaid::label_center;
                                static const label_placement_t label_any = ::DbBaseElements::Navaid::label_any;
                                static const label_placement_t label_off = ::DbBaseElements::Navaid::label_off;

                                typedef enum {
                                        airway_invalid,
                                        airway_low,
                                        airway_high,
                                        airway_both,
                                        airway_user
                                } airway_type_t;

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_all = subtables_none;

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Airway(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all) {}
                                unsigned int get_subtables(void) const { return ~0; }
                                bool is_subtables_loaded(void) const { return true; }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                Glib::ustring get_sourceid(void) const { return m_sourceid; }
                                void set_sourceid(const Glib::ustring& id) { m_sourceid = id; }
                                Rect get_bbox(void) const;
                                Point get_labelcoord(void) const { return m_begin_coord.halfway(m_end_coord); }
                                Point get_begin_coord(void) const { return m_begin_coord; }
                                void set_begin_coord(const Point& pt) { m_begin_coord = pt; }
                                Glib::ustring get_begin_name(void) const { return m_begin_name; }
                                void set_begin_name(const Glib::ustring& nm) { m_begin_name = nm; }
                                Point get_end_coord(void) const { return m_end_coord; }
                                void set_end_coord(const Point& pt) { m_end_coord = pt; }
                                Glib::ustring get_end_name(void) const { return m_end_name; }
                                void set_end_name(const Glib::ustring& nm) { m_end_name = nm; }
                                Glib::ustring get_name(void) const { return m_name; }
                                void set_name(const Glib::ustring& nm) { m_name = nm; }
                                int16_t get_base_level(void) const { return m_base_level; }
                                void set_base_level(int16_t lvl) { m_base_level = lvl; }
                                int16_t get_top_level(void) const { return m_top_level; }
                                void set_top_level(int16_t lvl) { m_top_level = lvl; }
                                airway_type_t get_type(void) const { return m_type; }
                                void set_type(airway_type_t t) { m_type = t; }
                                static Glib::ustring get_type_name(airway_type_t t);
                                Glib::ustring get_type_name(void) const { return get_type_name(m_type); }
                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                void set_label_placement(label_placement_t lp = label_off) { m_label_placement = lp; }
                                static Glib::ustring get_label_placement_name(label_placement_t lp) { return DbBaseElements::Navaid::get_label_placement_name(lp); }
                                Glib::ustring get_label_placement_name(void) const { return get_label_placement_name(m_label_placement); }
                                bool is_valid(void) const { return m_type != airway_invalid; }
                                time_t get_modtime(void) const { return m_modtime; }
                                void set_modtime(time_t t = 0) { m_modtime = t; }

                        protected:
                                int64_t m_id;
                                Point m_begin_coord;
                                Point m_end_coord;
                                Glib::ustring m_sourceid;
                                Glib::ustring m_begin_name;
                                Glib::ustring m_end_name;
                                Glib::ustring m_name;
                                int16_t m_base_level;
                                int16_t m_top_level;
                                airway_type_t m_type;
                                label_placement_t m_label_placement;
                                table_t m_table;
                                time_t m_modtime;
                };

                class Airport {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef ::DbBaseElements::Navaid::label_placement_t label_placement_t;
                                static const label_placement_t label_n = ::DbBaseElements::Navaid::label_n;
                                static const label_placement_t label_e = ::DbBaseElements::Navaid::label_e;
                                static const label_placement_t label_s = ::DbBaseElements::Navaid::label_s;
                                static const label_placement_t label_w = ::DbBaseElements::Navaid::label_w;
                                static const label_placement_t label_center = ::DbBaseElements::Navaid::label_center;
                                static const label_placement_t label_any = ::DbBaseElements::Navaid::label_any;
                                static const label_placement_t label_off = ::DbBaseElements::Navaid::label_off;

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_runways = (1 << 0);
                                static const unsigned int subtables_helipads = (1 << 1);
                                static const unsigned int subtables_comms = (1 << 2);
                                static const unsigned int subtables_vfrroutes = (1 << 3);
                                static const unsigned int subtables_linefeatures = (1 << 4);
                                static const unsigned int subtables_pointfeatures = (1 << 5);
                                static const unsigned int subtables_all = (subtables_runways | subtables_helipads | subtables_comms | subtables_vfrroutes | subtables_linefeatures | subtables_pointfeatures);

                                class Runway {
                                        public:
                                                static const uint16_t flag_le_usable = (1 << 0);
                                                static const uint16_t flag_he_usable = (1 << 1);

                                                static const unsigned int nr_lights = 8;

                                                typedef uint8_t light_t;
                                                static const light_t light_off                     = 0;
                                                static const light_t light_pcl                     = 1; // Pilot Controlled Lighting
                                                static const light_t light_sf                      = 2; // Sequenced Flashing Lights
                                                static const light_t light_tdzl                    = 3; // Touchdown Zone Lighting
                                                static const light_t light_cl                      = 4; // Centerline Lighting System
                                                static const light_t light_hirl                    = 5; // High Intensity Runway Lights
                                                static const light_t light_mirl                    = 6; // Medium Intensity Runway Lighting System
                                                static const light_t light_lirl                    = 7; // Low Intensity Runway Lighting System
                                                static const light_t light_rail                    = 8; // Runway Alignment Lights
                                                static const light_t light_reil                    = 9; // Runway End Identifier Lights
                                                static const light_t light_alsf2                   = 10; // ALSF-2
                                                static const light_t light_alsf1                   = 11; // ALSF-1
                                                static const light_t light_sals_salsf              = 12; // SALS or SALSF
                                                static const light_t light_ssalr                   = 13; // SSALR
                                                static const light_t light_mals_malsf_ssals_ssalsf = 14; // MALS and MALSF or SSALS and SSALF
                                                static const light_t light_malsr                   = 15; // MALSR
                                                static const light_t light_overruncenterline       = 16; // Overrun Centerline
                                                static const light_t light_center                  = 17; // Centerline and Bar
                                                static const light_t light_usconfiguration         = 18; // US Configuration (b)
                                                static const light_t light_hongkongcurve           = 19; // Hong Kong Curve
                                                static const light_t light_centerrow               = 20; // Center row
                                                static const light_t light_leftsinglerow           = 21; // Left Single Row
                                                static const light_t light_formernatostandard      = 22; // Former NATO Standard ©
                                                static const light_t light_centerrow2              = 23; // Center Row
                                                static const light_t light_natostandard            = 24; // NATO standard
                                                static const light_t light_centerdoublerow         = 25; // Center and Double Row
                                                static const light_t light_portableapproach        = 26; // Portable Approach
                                                static const light_t light_centerrow3              = 27; // Center Row
                                                static const light_t light_hals                    = 28; // Helicopter Approach Lighting System (HALS)
                                                static const light_t light_2parallelrow            = 30; // Two Parallel row
                                                static const light_t light_leftrow                 = 31; // Left Row (High Intensity)
                                                static const light_t light_airforceoverrun         = 32; // Air Force Overrun
                                                static const light_t light_calvert                 = 33; // Calvert (British)
                                                static const light_t light_singlerowcenterline     = 34; // Single Row Centerline
                                                static const light_t light_narrowmulticross        = 35; // Narrow Multi-cross
                                                static const light_t light_centerlinehighintensity = 36; // Centerline High Intensity
                                                static const light_t light_alternatecenverline     = 37; // Alternate Centerline and Bar
                                                static const light_t light_cross                   = 38; // Cross
                                                static const light_t light_centerrow4              = 39; // Center Row
                                                static const light_t light_singaporecenterline     = 40; // Singapore Centerline
                                                static const light_t light_centerline2crossbars    = 41; // Centerline 2 Crossbars
                                                static const light_t light_odals                   = 42; // Omni-directional Approach Lighting System
                                                static const light_t light_vasi                    = 43; // Visual Approach Slope Indicator
                                                static const light_t light_tvasi                   = 44; // T-Visual Approach Slope Indicator
                                                static const light_t light_pvasi                   = 45; // Pulsating Visual Approach Slope Indicator
                                                static const light_t light_jumbo                   = 46; // VASI with a TCH to accommodate long bodied or jumbo aircraft
                                                static const light_t light_tricolor                = 47; // Tri-color Arrival Approach (TRICOLOR)
                                                static const light_t light_apap                    = 48; // Alignment of Elements System
                                                static const light_t light_papi                    = 50; // Precision Approach Path Indicator
                                                static const light_t light_ols                     = 51; // Optical landing System
                                                static const light_t light_waveoff                 = 52;
                                                static const light_t light_portable                = 53;
                                                static const light_t light_floods                  = 54;
                                                static const light_t light_lights                  = 55;
                                                static const light_t light_lcvasi                  = 56; // Low Cost Visual Approach Slope Indicator

                                                Runway(const Glib::ustring& idhe, const Glib::ustring& idle, uint16_t len, uint16_t wid,
                                                       const Glib::ustring& sfc, const Point& hecoord, const Point& lecoord,
                                                       uint16_t hetda, uint16_t helda, uint16_t hedisp, uint16_t hehdg, int16_t heelev,
                                                       uint16_t letda, uint16_t lelda, uint16_t ledisp, uint16_t lehdg, int16_t leelev,
                                                       uint16_t flags);
                                                Glib::ustring get_ident_he(void) const { return m_ident_he; }
                                                Glib::ustring get_ident_le(void) const { return m_ident_le; }
                                                uint16_t get_length(void) const { return m_length; }
                                                uint16_t get_width(void) const { return m_width; }
                                                Glib::ustring get_surface(void) const { return m_surface; }
                                                Point get_he_coord(void) const { return m_he_coord; }
                                                Point get_le_coord(void) const { return m_le_coord; }
                                                uint16_t get_he_tda(void) const { return m_he_tda; }
                                                uint16_t get_he_lda(void) const { return m_he_lda; }
                                                uint16_t get_he_disp(void) const { return m_he_disp; }
                                                uint16_t get_he_hdg(void) const { return m_he_hdg; }
                                                int16_t get_he_elev(void) const { return m_he_elev; }
                                                uint16_t get_le_tda(void) const { return m_le_tda; }
                                                uint16_t get_le_lda(void) const { return m_le_lda; }
                                                uint16_t get_le_disp(void) const { return m_le_disp; }
                                                uint16_t get_le_hdg(void) const { return m_le_hdg; }
                                                int16_t get_le_elev(void) const { return m_le_elev; }
                                                uint16_t get_flags(void) const { return m_flags; }
                                                light_t get_he_light(unsigned int idx) const { return m_he_lights[idx & 7]; }
                                                light_t get_le_light(unsigned int idx) const { return m_le_lights[idx & 7]; }
                                                void set_ident_he(const Glib::ustring& x) { m_ident_he = x; }
                                                void set_ident_le(const Glib::ustring& x) { m_ident_le = x; }
                                                void set_length(uint16_t x) { m_length = x; }
                                                void set_width(uint16_t x) { m_width = x; }
                                                void set_surface(const Glib::ustring& x) { m_surface = x; }
                                                void set_he_coord(const Point& pt) { m_he_coord = pt; }
                                                void set_le_coord(const Point& pt) { m_le_coord = pt; }
                                                void set_he_tda(uint16_t x) { m_he_tda = x; }
                                                void set_he_lda(uint16_t x) { m_he_lda = x; }
                                                void set_he_disp(uint16_t x) { m_he_disp = x; }
                                                void set_he_hdg(uint16_t x) { m_he_hdg = x; }
                                                void set_he_elev(int16_t x) { m_he_elev = x; }
                                                void set_le_tda(uint16_t x) { m_le_tda = x; }
                                                void set_le_lda(uint16_t x) { m_le_lda = x; }
                                                void set_le_disp(uint16_t x) { m_le_disp = x; }
                                                void set_le_hdg(uint16_t x) { m_le_hdg = x; }
                                                void set_le_elev(int16_t x) { m_le_elev = x; }
                                                void set_flags(uint16_t x) { m_flags = x; }
                                                void set_he_light(unsigned int idx, light_t x) { m_he_lights[idx & 7] = x; }
                                                void set_le_light(unsigned int idx, light_t x) { m_le_lights[idx & 7] = x; }
                                                bool is_he_usable(void) const { return !!(m_flags & flag_he_usable); }
                                                bool is_le_usable(void) const { return !!(m_flags & flag_le_usable); }
                                                bool is_concrete(void) const;
                                                bool is_water(void) const;

                                                static Glib::ustring get_light_string(light_t x);

                                                void compute_default_hdg(void);
                                                void compute_default_coord(const Point& arptcoord);

                                                void swap_he_le(void);
                                                bool normalize_he_le(void);

                                        protected:
                                                Glib::ustring m_ident_he;
                                                Glib::ustring m_ident_le;
                                                uint16_t m_length;
                                                uint16_t m_width;
                                                Glib::ustring m_surface;
                                                Point m_he_coord;
                                                Point m_le_coord;
                                                uint16_t m_he_tda;
                                                uint16_t m_he_lda;
                                                uint16_t m_he_disp;
                                                uint16_t m_he_hdg;
                                                int16_t m_he_elev;
                                                uint16_t m_le_tda;
                                                uint16_t m_le_lda;
                                                uint16_t m_le_disp;
                                                uint16_t m_le_hdg;
                                                int16_t m_le_elev;
                                                uint16_t m_flags;
                                                light_t m_he_lights[nr_lights];
                                                light_t m_le_lights[nr_lights];
                                };

                                class Helipad {
                                        public:
                                                static const uint16_t flag_usable = (1 << 0);

                                                Helipad(const Glib::ustring& id, uint16_t len, uint16_t wid,
                                                        const Glib::ustring& sfc, const Point& coord,
                                                        uint16_t hdg, int16_t elev, uint16_t flags);
                                                Glib::ustring get_ident(void) const { return m_ident; }
                                                uint16_t get_length(void) const { return m_length; }
                                                uint16_t get_width(void) const { return m_width; }
                                                Glib::ustring get_surface(void) const { return m_surface; }
                                                Point get_coord(void) const { return m_coord; }
                                                uint16_t get_hdg(void) const { return m_hdg; }
                                                int16_t get_elev(void) const { return m_elev; }
                                                uint16_t get_flags(void) const { return m_flags; }
                                                void set_ident(const Glib::ustring& x) { m_ident = x; }
                                                void set_length(uint16_t x) { m_length = x; }
                                                void set_width(uint16_t x) { m_width = x; }
                                                void set_surface(const Glib::ustring& x) { m_surface = x; }
                                                void set_coord(const Point& pt) { m_coord = pt; }
                                                void set_hdg(uint16_t x) { m_hdg = x; }
                                                void set_elev(int16_t x) { m_elev = x; }
                                                void set_flags(uint16_t x) { m_flags = x; }
                                                bool is_usable(void) const { return !!(m_flags & flag_usable); }
                                                bool is_concrete(void) const;

                                        protected:
                                                Glib::ustring m_ident;
                                                uint16_t m_length;
                                                uint16_t m_width;
                                                Glib::ustring m_surface;
                                                Point m_coord;
                                                uint16_t m_hdg;
                                                int16_t m_elev;
                                                uint16_t m_flags;
                                };

                                class Comm {
                                        public:
                                                Comm(const Glib::ustring& nm, const Glib::ustring& sect, const Glib::ustring& oprhr,
                                                                uint32_t f0 = 0, uint32_t f1 = 0, uint32_t f2 = 0, uint32_t f3 = 0, uint32_t f4 = 0);
                                                Glib::ustring get_name(void) const { return m_name; }
                                                Glib::ustring get_sector(void) const { return m_sector; }
                                                Glib::ustring get_oprhours(void) const { return m_oprhours; }
                                                uint32_t operator[](unsigned int x) const { if (x >= 5) return 0; return m_freq[x]; }
                                                void set_name(const Glib::ustring& x) { m_name = x; }
                                                void set_sector(const Glib::ustring& x) { m_sector = x; }
                                                void set_oprhours(const Glib::ustring& x) { m_oprhours = x; }
                                                void set_freq(unsigned int x, uint32_t y) { if (x < 5) m_freq[x] = y; }

                                        protected:
                                                uint32_t m_freq[5];
                                                Glib::ustring m_name;
                                                Glib::ustring m_sector;
                                                Glib::ustring m_oprhours;
                                };

                                class VFRRoute {
                                        public:
                                                class VFRRoutePoint {
                                                        public:
                                                                typedef enum {
                                                                        pathcode_arrival,
                                                                                pathcode_departure,
                                                                                pathcode_holding,
                                                                                pathcode_trafficpattern,
                                                                                pathcode_vfrtransition,
                                                                                pathcode_other,
                                                                                pathcode_invalid
                                                                } pathcode_t;

                                                                typedef enum {
                                                                        altcode_atorabove,
                                                                                altcode_atorbelow,
                                                                                altcode_assigned,
                                                                                altcode_altspecified,
                                                                                altcode_recommendedalt,
                                                                                altcode_invalid
                                                                } altcode_t;

                                                                VFRRoutePoint(const Glib::ustring& nm, const Point& pt, int16_t al, label_placement_t lp, char pts, pathcode_t pc, altcode_t ac, bool aa);
                                                                Glib::ustring get_name(void) const { return m_name; }
                                                                Point get_coord(void) const { return m_coord; }
                                                                int16_t get_altitude(void) const { return m_alt; }
                                                                Glib::ustring get_altitude_string(void) const;
                                                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                                                char get_ptsym(void) const { return m_ptsym; }
                                                                pathcode_t get_pathcode(void) const { return m_pathcode; }
                                                                static Glib::ustring get_pathcode_string(pathcode_t pc);
                                                                Glib::ustring get_pathcode_string(void) const { return get_pathcode_string(m_pathcode); }
                                                                altcode_t get_altcode(void) const { return m_altcode; }
                                                                static Glib::ustring get_altcode_string(altcode_t ac);
                                                                Glib::ustring get_altcode_string(void) const { return get_altcode_string(m_altcode); }
                                                                bool is_at_airport(void) const { return m_at_airport; }
                                                                bool is_compulsory(void) const { return get_ptsym() == 'C'; }
                                                                void set_name(const Glib::ustring& x) { m_name = x; }
                                                                void set_coord(const Point& x) { m_coord = x; }
                                                                void set_altitude(int16_t x) { m_alt = x; }
                                                                void set_label_placement(label_placement_t x) { m_label_placement = x; }
                                                                void set_ptsym(char x) { m_ptsym = x; }
                                                                void set_pathcode(pathcode_t x) { m_pathcode = x; }
                                                                void set_altcode(altcode_t x) { m_altcode = x; }
                                                                void set_at_airport(bool x) { m_at_airport = x; }

                                                        private:
                                                                Glib::ustring m_name;
                                                                Point m_coord;
                                                                int16_t m_alt;
                                                                label_placement_t m_label_placement;
                                                                char m_ptsym;
                                                                pathcode_t m_pathcode;
                                                                altcode_t m_altcode;
                                                                bool m_at_airport;
                                                };

                                                VFRRoute(const Glib::ustring& nm);
                                                Glib::ustring get_name(void) const { return m_name; }
                                                void set_name(const Glib::ustring& x) { m_name = x; }
                                                unsigned int size(void) const { return m_pts.size(); }
                                                const VFRRoutePoint& operator[](unsigned int x) const { return m_pts[x]; }
                                                VFRRoutePoint& operator[](unsigned int x) { return m_pts[x]; }
                                                void add_point(const VFRRoutePoint& pt) { m_pts.push_back(pt); }
                                                void insert_point(const VFRRoutePoint& x, int index) { m_pts.insert(m_pts.begin() + index, x); }
                                                void erase_point(int index) { m_pts.erase(m_pts.begin() + index); }
                                                void reverse(void);
                                                void name(void);

                                        protected:
                                                Glib::ustring m_name;
                                                std::vector<VFRRoutePoint> m_pts;
                                };

                                class PolylineNode {
                                        public:
                                                typedef enum {
                                                        feature_off = 0,                                // Nothing (can be used for a break in a chain of nodes without restarting a new chain)
                                                        feature_solid_yellow_line = 1,                  // Solid yellow line (eg. taxiway centreline or taxiway edge line).
                                                        feature_broken_yellow_line = 2,                 // Broken yellow line (used for taxiway-taxiway hold positions, AIM 2-3-6))
                                                        feature_double_solid_yellow_line = 3,           // Double solid yellow line (eg. taxiway edge lines)
                                                        feature_runway_hold_line = 4,                   // Runway hold lines (four yellow lines, two solid and two broken, solid line to right side along axis of chain) 
                                                        feature_other_hold_line = 5,                    // Other hold lines (one solid yellow and one broken, solid line to right along axis of chain) (used to define hold lines NOT related to runway, such as exits from aprons to active taxiways)
                                                        feature_ils_hold_line = 6,                      // ILS hold lines (two yellow lines with perpendicular yellow bars between them.
                                                        feature_yellow_line_with_broken_lines = 7,      // Yellow line with broken line on each side (used on taxiway centre lines for in the "ILS critical zone", on approach to a runway hold line).
                                                        feature_widely_broken_single_yellow_line = 8,   // Widely separated broken, single yellow lines (used to mark "lanes" for queueing  aeroplanes at hold points, AIM 2-3-6 Fig 2-3-18)
                                                        feature_widely_broken_double_yellow_line = 9,   // Widely separated broken, double yellow lines (used to mark "lanes" for queueing  aeroplanes at hold points, AIM 2-3-6 Fig 2-3-18)
                                                        feature_solid_white_line = 20,                  // Solid white line (to mark roadways on aprons)
                                                        feature_white_checkerboard = 21,                // White chequerboard (to mark roadways on aprons)
                                                        feature_broken_white_line = 22,                 // Broken white line (to mark roadways on aprons)
                                                        feature_solid_yellow_line_on_black = 51,        // Add 50 to the above codes to include a  black background, to improve visibility on light-coloured surfaces (such as concrete).
                                                        feature_broken_yellow_line_on_black = 52,
                                                        feature_double_solid_yellow_line_on_black = 53,
                                                        feature_runway_hold_line_on_black = 54,
                                                        feature_other_hold_line_on_black = 55,
                                                        feature_ils_hold_line_on_black = 56,
                                                        feature_yellow_line_with_broken_lines_on_black = 57,
                                                        feature_widely_broken_single_yellow_line_on_black = 58,
                                                        feature_widely_broken_double_yellow_line_on_black = 59,
                                                        feature_green_embedded_lights = 101,            // Green embedded centreline lights - bidirectional along (both ways) the string axis. 
                                                        feature_blue_edge_lights = 102,                 // Blue taxiway edge lights, omnidirectional.
                                                        feature_amber_embedded_lights = 103,            // Closely spaced Amber embedded lights, unidirectional, (one-way perpendicular to right of string axis), visible only to the left of the direction on the string (typically used for hold lines across taxiways).
                                                        feature_amber_pulsating_embedded_lights = 104,  // Closely spaced Amber embedded lights, unidirectional & alternately pulsating (one-way perpendicular to right of  string axis) and pulsating alternately (used for some hold short lines across wide taxiways, visible only to an aeroplane approaching the hold line towards the runway entrance).  Will pulse at between 30 and 60 Hz, and always "on".
                                                        feature_amber_green_embedded_lights = 105,      // Amber & green runway holding point embedded centreline lights:  Steady alternating amber and green lights, directional along both directions of the string axis.  Usually used approaching hold points in association with line type 7 in the "ILS critical zone" at the approach of a taxiway to a runway.
                                                        feature_red_edge_lights = 106                   // Red boundary edge lights, omnidirectional (eg. to mark edges of taxiway bridges instead of blue edge lights)

                                                } feature_t;

                                                static const unsigned int nr_features = 2;

                                                static const uint16_t flag_bezier_prev = (1 << 0);
                                                static const uint16_t flag_bezier_next = (1 << 1);
                                                static const uint16_t flag_endpath = (1 << 2);
                                                static const uint16_t flag_closepath = (1 << 3);
                                                static const uint16_t flag_bezier_splitcontrol = (1 << 15); // only for DB

                                                PolylineNode(const Point& pt = Point(), const Point& pcpt = Point(), const Point& ncpt = Point(), uint16_t flags = 0, feature_t f1 = feature_off, feature_t f2 = feature_off);
                                                Point get_point(void) const { return m_point; }
                                                Point get_prev_controlpoint(void) const { return m_prev_controlpoint; }
                                                Point get_next_controlpoint(void) const { return m_next_controlpoint; }
                                                uint16_t get_flags(void) const { return m_flags; }
                                                feature_t get_feature(unsigned int idx) const { return m_feature[idx % nr_features]; }
                                                void set_point(const Point& pt) { m_point = pt; }
                                                void set_prev_controlpoint(const Point& pt) { m_prev_controlpoint = pt; }
                                                void set_next_controlpoint(const Point& pt) { m_next_controlpoint = pt; }
                                                void set_flags(uint16_t flg) { m_flags = flg; }
                                                void set_feature(unsigned int idx, feature_t ft) { m_feature[idx % nr_features] = ft; }
                                                bool is_bezier_prev(void) const { return !!(m_flags & flag_bezier_prev); }
                                                bool is_bezier_next(void) const { return !!(m_flags & flag_bezier_next); }
                                                bool is_bezier(void) const { return !!(m_flags & (flag_bezier_prev | flag_bezier_next)); }
                                                bool is_endpath(void) const { return !!(m_flags & flag_endpath); }
                                                bool is_closepath(void) const { return !!(m_flags & flag_closepath); }

                                                unsigned int blobsize(void) const;
                                                uint8_t *blobencode(uint8_t *ptr) const;
                                                const uint8_t *blobdecode(const uint8_t *ptr, uint8_t const *end);

                                        protected:
                                                Point m_point;
                                                Point m_prev_controlpoint;
                                                Point m_next_controlpoint;
                                                uint16_t m_flags;
                                                feature_t m_feature[nr_features];
                                };

                                class Polyline : public std::vector<PolylineNode> {
                                        public:
                                                static const uint16_t flag_area = (1 << 0);
                                                static const uint16_t flag_airportboundary = (1 << 1);

                                                Polyline(void) : m_flags(0) {}

                                                Glib::ustring get_name(void) const { return m_name; }
                                                void set_name(const Glib::ustring& s) { m_name = s; }
                                                Glib::ustring get_surface(void) const { return m_surface; }
                                                void set_surface(const Glib::ustring& s) { m_surface = s; }
                                                uint16_t get_flags(void) const { return m_flags; }
                                                void set_flags(uint16_t flg) { m_flags = flg; }
                                                bool is_area(void) const { return !!(m_flags & flag_area); }
                                                bool is_airportboundary(void) const { return !!(m_flags & flag_airportboundary); }
                                                bool is_concrete(void) const;

                                                unsigned int blobsize(void) const;
                                                uint8_t *blobencode(uint8_t *ptr) const;
                                                const uint8_t *blobdecode(const uint8_t *ptr, uint8_t const *end);

                                                void bindblob(sqlite3x::sqlite3_command& cmd, int index) const;
                                                void getblob(sqlite3x::sqlite3_cursor& cursor, int index);

                                                unsigned int blobsize_polygon(void) const;
                                                uint8_t *blobencode_polygon(uint8_t *ptr) const;
                                                const uint8_t *blobdecode_polygon(const uint8_t *ptr, uint8_t const *end);

                                                void bindblob_polygon(sqlite3x::sqlite3_command& cmd, int index) const;
                                                void getblob_polygon(sqlite3x::sqlite3_cursor& cursor, int index);

                                                operator PolygonHole(void) const;

                                                Rect get_bbox(void) const;

                                        protected:
                                                Glib::ustring m_name;
                                                Glib::ustring m_surface;
                                                uint16_t m_flags;
                                };

                                class PointFeature {
                                        public:
                                                typedef enum {
                                                        feature_invalid = 0,
                                                        feature_vasi = 1,
                                                        feature_taxiwaysign = 2,
                                                        feature_startuploc = 3,
                                                        feature_tower = 4,
                                                        feature_lightbeacon = 5,
                                                        feature_windsock = 6
                                                } feature_t;

                                                PointFeature(feature_t ft = feature_invalid, const Point& pt = Point(), uint16_t hdg = 0, int16_t elev = 0,
                                                             unsigned int subtype = 0, int attr1 = 0, int attr2 = 0, const Glib::ustring& name = "", const Glib::ustring& rwyident = "");

                                                feature_t get_feature(void) const { return m_feature; }
                                                void set_feature(feature_t x) { m_feature = x; }
                                                Point get_coord(void) const { return m_coord; }
                                                void set_coord(const Point& pt) { m_coord = pt; }
                                                uint16_t get_hdg(void) const { return m_hdg; }
                                                void set_hdg(uint16_t x) { m_hdg = x; }
                                                int16_t get_elev(void) const { return m_elev; }
                                                void set_elev(int16_t x) { m_elev = x; }
                                                unsigned int get_subtype(void) const { return m_subtype; }
                                                void set_subtype(unsigned int x) { m_subtype = x; }
                                                int get_attr1(void) const { return m_attr1; }
                                                void set_attr1(int x) { m_attr1 = x; }
                                                int get_attr2(void) const { return m_attr2; }
                                                void set_attr2(int x) { m_attr2 = x; }
                                                Glib::ustring get_name(void) const { return m_name; }
                                                void set_name(const Glib::ustring& x) { m_name = x; }
                                                Glib::ustring get_rwyident(void) const { return m_rwyident; }
                                                void set_rwyident(const Glib::ustring& x) { m_rwyident = x; }

                                        protected:
                                                feature_t m_feature;
                                                Point m_coord;
                                                uint16_t m_hdg;
                                                int16_t m_elev;
                                                unsigned int m_subtype;
                                                int m_attr1;
                                                int m_attr2;
                                                Glib::ustring m_name;
                                                Glib::ustring m_rwyident;
                                };

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Airport(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all);
                                unsigned int get_subtables(void) const { return m_subtables; }
                                bool is_subtables_loaded(void) const { return !(subtables_all & ~m_subtables); }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                Glib::ustring get_sourceid(void) const { return m_sourceid; }
                                void set_sourceid(const Glib::ustring& id) { m_sourceid = id; }
                                Rect get_bbox(void) const { return m_bbox; }
                                Point get_coord(void) const { return m_coord; }
                                void set_coord(const Point& pt) { m_coord = pt; recompute_bbox(); }
                                Glib::ustring get_icao(void) const { return m_icao; }
                                void set_icao(const Glib::ustring& ic) { m_icao = ic; }
                                Glib::ustring get_name(void) const { return m_name; }
                                void set_name(const Glib::ustring& nm) { m_name = nm; }
                                Glib::ustring get_icao_name(void) const;
                                Glib::ustring get_vfrrmk(void) const { return m_vfrrmk; }
                                void set_vfrrmk(const Glib::ustring& vr) { m_vfrrmk = vr; }
                                int16_t get_elevation(void) const { return m_elev; }
                                void set_elevation(int16_t elev) { m_elev = elev; }
                                char get_typecode(void) const { return m_typecode; }
                                void set_typecode(char t) { m_typecode = t; }
                                static Glib::ustring get_type_string(char t);
                                Glib::ustring get_type_string(void) const { return get_type_string(m_typecode); }
                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                void set_label_placement(label_placement_t lp = label_off) { m_label_placement = lp; }
                                static Glib::ustring get_label_placement_name(label_placement_t lp) { return DbBaseElements::Navaid::get_label_placement_name(lp); }
                                Glib::ustring get_label_placement_name(void) const { return get_label_placement_name(m_label_placement); }
                                bool is_valid(void) const { return !!m_typecode; }
                                time_t get_modtime(void) const { return m_modtime; }
                                void set_modtime(time_t t = 0) { m_modtime = t; }
                                unsigned int get_nrrwy(void) const { return m_rwy.size(); }
                                const Runway& get_rwy(unsigned int x) const { return m_rwy[x]; }
                                Runway& get_rwy(unsigned int x) { return m_rwy[x]; }
                                unsigned int get_nrhelipads(void) const { return m_helipad.size(); }
                                const Helipad& get_helipad(unsigned int x) const { return m_helipad[x]; }
                                Helipad& get_helipad(unsigned int x) { return m_helipad[x]; }
                                unsigned int get_nrcomm(void) const { return m_comm.size(); }
                                const Comm& get_comm(unsigned int x) const { return m_comm[x]; }
                                Comm& get_comm(unsigned int x) { return m_comm[x]; }
                                unsigned int get_nrvfrrte(void) const { return m_vfrrte.size(); }
                                const VFRRoute& get_vfrrte(unsigned int x) const { return m_vfrrte[x]; }
                                VFRRoute& get_vfrrte(unsigned int x) { return m_vfrrte[x]; }
                                unsigned int get_nrlinefeatures(void) const { return m_linefeature.size(); }
                                const Polyline& get_linefeature(unsigned int x) const { return m_linefeature[x]; }
                                Polyline& get_linefeature(unsigned int x) { return m_linefeature[x]; }
                                unsigned int get_nrpointfeatures(void) const { return m_pointfeature.size(); }
                                const PointFeature& get_pointfeature(unsigned int x) const { return m_pointfeature[x]; }
                                PointFeature& get_pointfeature(unsigned int x) { return m_pointfeature[x]; }
                                void add_rwy(const Runway& x) { m_rwy.push_back(x); recompute_bbox(); }
                                void insert_rwy(const Runway& x, int index) { m_rwy.insert(m_rwy.begin() + index, x); recompute_bbox(); }
                                void erase_rwy(int index) { m_rwy.erase(m_rwy.begin() + index); recompute_bbox(); }
                                void add_helipad(const Helipad& x) { m_helipad.push_back(x); recompute_bbox(); }
                                void insert_helipad(const Helipad& x, int index) { m_helipad.insert(m_helipad.begin() + index, x); recompute_bbox(); }
                                void erase_helipad(int index) { m_helipad.erase(m_helipad.begin() + index); recompute_bbox(); }
                                void add_comm(const Comm& x) { m_comm.push_back(x); }
                                void insert_comm(const Comm& x, int index) { m_comm.insert(m_comm.begin() + index, x); }
                                void erase_comm(int index) { m_comm.erase(m_comm.begin() + index); }
                                void add_vfrrte(const VFRRoute& x) { m_vfrrte.push_back(x); recompute_bbox(); }
                                void insert_vfrrte(const VFRRoute& x, int index) { m_vfrrte.insert(m_vfrrte.begin() + index, x); recompute_bbox(); }
                                void erase_vfrrte(int index) { m_vfrrte.erase(m_vfrrte.begin() + index); recompute_bbox(); }
                                void add_linefeature(const Polyline& x) { m_linefeature.push_back(x); recompute_bbox(); }
                                void insert_linefeature(const Polyline& x, int index) { m_linefeature.insert(m_linefeature.begin() + index, x); recompute_bbox(); }
                                void erase_linefeature(int index) { m_linefeature.erase(m_linefeature.begin() + index); recompute_bbox(); }
                                void add_pointfeature(const PointFeature& x) { m_pointfeature.push_back(x); recompute_bbox(); recompute_bbox(); }
                                void insert_pointfeature(const PointFeature& x, int index) { m_pointfeature.insert(m_pointfeature.begin() + index, x); recompute_bbox(); }
                                void erase_pointfeature(int index) { m_pointfeature.erase(m_pointfeature.begin() + index); recompute_bbox(); }

                                void recompute_bbox(void);

                        protected:
                                int64_t m_id;
                                std::vector<Runway> m_rwy;
                                std::vector<Helipad> m_helipad;
                                std::vector<Comm> m_comm;
                                std::vector<VFRRoute> m_vfrrte;
                                std::vector<Polyline> m_linefeature;
                                std::vector<PointFeature> m_pointfeature;
                                Glib::ustring m_sourceid;
                                Glib::ustring m_icao;
                                Glib::ustring m_name;
                                Glib::ustring m_vfrrmk;
                                Rect m_bbox;
                                Point m_coord;
                                int16_t m_elev;
                                char m_typecode;
                                unsigned int m_subtables;
                                label_placement_t m_label_placement;
                                table_t m_table;
                                time_t m_modtime;
                };

                class Airspace {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef ::DbBaseElements::Navaid::label_placement_t label_placement_t;
                                static const label_placement_t label_n = ::DbBaseElements::Navaid::label_n;
                                static const label_placement_t label_e = ::DbBaseElements::Navaid::label_e;
                                static const label_placement_t label_s = ::DbBaseElements::Navaid::label_s;
                                static const label_placement_t label_w = ::DbBaseElements::Navaid::label_w;
                                static const label_placement_t label_center = ::DbBaseElements::Navaid::label_center;
                                static const label_placement_t label_any = ::DbBaseElements::Navaid::label_any;
                                static const label_placement_t label_off = ::DbBaseElements::Navaid::label_off;

                                static const uint8_t altflag_agl   = (1 << 0);
                                static const uint8_t altflag_fl    = (1 << 1);
                                static const uint8_t altflag_sfc   = (1 << 2);
                                static const uint8_t altflag_gnd   = (1 << 3);
                                static const uint8_t altflag_unlim = (1 << 4);
                                static const uint8_t altflag_notam = (1 << 5);
                                static const uint8_t altflag_unkn  = (1 << 6);

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_segments = (1 << 0);
                                static const unsigned int subtables_all = subtables_segments;

                                class Segment {
                                        public:
                                                Segment(void);
                                                Segment(const Point& coord1, const Point& coord2, const Point& coord0, char shape, int32_t r1, int32_t r2);
                                                Point get_coord1(void) const { return m_coord1; }
                                                void set_coord1(const Point& pt) { m_coord1 = pt; }
                                                Point get_coord2(void) const { return m_coord2; }
                                                void set_coord2(const Point& pt) { m_coord2 = pt; }
                                                Point get_coord0(void) const { return m_coord0; }
                                                void set_coord0(const Point& pt) { m_coord0 = pt; }
                                                char get_shape(void) const { return m_shape; }
                                                void set_shape(char shp) { m_shape = shp; }
                                                int32_t get_radius1(void) const { return m_radius1; }
                                                void set_radius1(int32_t r = 0) { m_radius1 = r; }
                                                int32_t get_radius2(void) const { return m_radius2; }
                                                void set_radius2(int32_t r = 0) { m_radius2 = r; }
                                                bool is_valid(void) const { return !!m_shape; }

                                        private:
                                                Point m_coord1;
                                                Point m_coord2;
                                                Point m_coord0;
                                                int32_t m_radius1;
                                                int32_t m_radius2;
                                                char m_shape;
                                };

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                typedef int32_t gndelev_t;

                                Airspace(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all);
                                unsigned int get_subtables(void) const { return m_subtables; }
                                bool is_subtables_loaded(void) const { return !(subtables_all & ~m_subtables); }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                Glib::ustring get_sourceid(void) const { return m_sourceid; }
                                void set_sourceid(const Glib::ustring& id) { m_sourceid = id; }
                                Rect get_bbox(void) const { return m_bbox; }
                                Point get_labelcoord(void) const { return m_labelcoord; }
                                void set_labelcoord(const Point& pt) { m_labelcoord = pt; recompute_bbox(); }
                                Glib::ustring get_icao(void) const { return m_icao; }
                                void set_icao(const Glib::ustring& ic) { m_icao = ic; }
                                Glib::ustring get_name(void) const { return m_name; }
                                void set_name(const Glib::ustring& nm) { m_name = nm; }
                                Glib::ustring get_icao_name(void) const;
                                Glib::ustring get_ident(void) const { return m_ident; }
                                void set_ident(const Glib::ustring& nm) { m_ident = nm; }
                                Glib::ustring get_classexcept(void) const { return m_classexcept; }
                                void set_classexcept(const Glib::ustring& s) { m_classexcept = s; }
                                Glib::ustring get_efftime(void) const { return m_efftime; }
                                void set_efftime(const Glib::ustring& s) { m_efftime = s; }
                                Glib::ustring get_note(void) const { return m_note; }
                                void set_note(const Glib::ustring& s) { m_note = s; }
                                Glib::ustring get_commname(void) const { return m_commname; }
                                void set_commname(const Glib::ustring& s) { m_commname = s; }
                                uint32_t get_commfreq(unsigned int x = 0) const { if (x >= 2) return 0; return m_commfreq[x]; }
                                void set_commfreq(unsigned int x = 0, uint32_t f = 0) { if (x < 2) m_commfreq[x] = f; }
                                char get_bdryclass(void) const { return m_bdryclass; }
                                void set_bdryclass(char bclass) { m_bdryclass = bclass; }
                                uint8_t get_typecode(void) const { return m_typecode; }
                                void set_typecode(uint8_t tc) { m_typecode = tc; }
                                gndelev_t get_gndelevmin(void) const { return m_gndelevmin; }
                                void set_gndelevmin(gndelev_t e) { m_gndelevmin = e; }
                                gndelev_t get_gndelevmax(void) const { return m_gndelevmax; }
                                void set_gndelevmax(gndelev_t e) { m_gndelevmax = e; }
                                int32_t get_altlwr(void) const { return m_altlwr; }
                                void set_altlwr(int32_t a) { m_altlwr = a; }
                                int32_t get_altupr(void) const { return m_altupr; }
                                void set_altupr(int32_t a) { m_altupr = a; }
                                uint8_t get_altlwrflags(void) const { return m_altlwrflags; }
                                void set_altlwrflags(uint8_t af) { m_altlwrflags = af; }
                                uint8_t get_altuprflags(void) const { return m_altuprflags; }
                                void set_altuprflags(uint8_t af) { m_altuprflags = af; }
                                int32_t get_altlwr_corr(void) const;
                                int32_t get_altupr_corr(void) const;
                                Glib::ustring get_altlwr_string(void) const;
                                Glib::ustring get_altupr_string(void) const;
                                static Gdk::Color get_color(char bclass, uint8_t tc);
                                Gdk::Color get_color(void) const { return get_color(m_bdryclass, m_typecode); }
                                static Glib::ustring get_class_string(char bc, uint8_t tc = 0);
                                Glib::ustring get_class_string(void) const { return get_class_string(m_bdryclass, m_typecode); }
                                static Glib::ustring get_type_string(uint8_t tc);
                                Glib::ustring get_type_string(void) const { return get_type_string(m_typecode); }
                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                void set_label_placement(label_placement_t lp = label_off) { m_label_placement = lp; }
                                static Glib::ustring get_label_placement_name(label_placement_t lp) { return DbBaseElements::Navaid::get_label_placement_name(lp); }
                                Glib::ustring get_label_placement_name(void) const { return get_label_placement_name(m_label_placement); }
                                bool is_valid(void) const { return !!m_bdryclass; }
                                time_t get_modtime(void) const { return m_modtime; }
                                void set_modtime(time_t t = 0) { m_modtime = t; }
                                unsigned int get_nrsegments(void) const { return m_seg.size(); }
                                Segment& operator[](unsigned int x) { return m_seg[x]; }
                                const Segment& operator[](unsigned int x) const { return m_seg[x]; }
                                void add_segment(const Segment& x) { m_seg.push_back(x); recompute_bbox(); }
                                void insert_segment(const Segment& x, int index) { m_seg.insert(m_seg.begin() + index, x); recompute_bbox(); }
                                void erase_segment(int index) { m_seg.erase(m_seg.begin() + index); recompute_bbox(); }
                                void clear_segments(void) { m_seg.clear(); recompute_bbox(); }
                                const Polygon& get_polygon(void) const { return m_polygon; }
                                void recompute_lineapprox(void);
                                void recompute_bbox(void);
                                void compute_initial_label_placement(void);
                                static float label_metric(const Point& p0, const Point& p1);
                                void compute_segments_from_poly(void);
                                void set_polygon(const Polygon& poly);
                                Point get_coord(void) const { return get_labelcoord(); }

                                static const gndelev_t gndelev_unknown;

                        protected:
                                int64_t m_id;
                                std::vector<Segment> m_seg;
                                Polygon m_polygon;
                                Glib::ustring m_sourceid;
                                Glib::ustring m_icao;
                                Glib::ustring m_name;
                                Glib::ustring m_ident;
                                Glib::ustring m_classexcept;
                                Glib::ustring m_efftime;
                                Glib::ustring m_note;
                                Glib::ustring m_commname;
                                uint32_t m_commfreq[2];
                                Rect m_bbox;
                                Point m_labelcoord;
                                label_placement_t m_label_placement;
                                uint8_t m_typecode;
                                char m_bdryclass;
                                int32_t m_altlwr, m_altupr;
                                uint8_t m_altlwrflags, m_altuprflags;
                                gndelev_t m_gndelevmin, m_gndelevmax;
                                time_t m_modtime;
                                table_t m_table;
                                unsigned int m_subtables;

                                static const float boundarycircleapprox = 10.0;
                                void recompute_lineapprox_circle(void);
                                static bool triangle_center(Point& ret, Point p0, Point p1, Point p2);
                };

                class Mapelement {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                static const uint8_t maptyp_invalid         = 0;

                                static const uint8_t maptyp_railway         = 47;
                                static const uint8_t maptyp_railway_d       = 57;
                                static const uint8_t maptyp_aerial_cable    = 48;

                                static const uint8_t maptyp_highway         = 45;
                                static const uint8_t maptyp_road            = 46;
                                static const uint8_t maptyp_trail           = 58;

                                static const uint8_t maptyp_river           = 50;
                                static const uint8_t maptyp_lake            = 49;
                                static const uint8_t maptyp_river_t         = 54;
                                static const uint8_t maptyp_lake_t          = 55;

                                static const uint8_t maptyp_canal           = 51;

                                static const uint8_t maptyp_pack_ice        = 56;
                                static const uint8_t maptyp_glacier         = 40;
                                static const uint8_t maptyp_forest          = 60;

                                static const uint8_t maptyp_city            = 42;
                                static const uint8_t maptyp_village         = 43;
                                static const uint8_t maptyp_spot            = 38;
                                static const uint8_t maptyp_landmark        = 44;

                                static const uint8_t maptyp_powerline       = 130;
                                static const uint8_t maptyp_border          = 134;

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_all = subtables_none;

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Mapelement(void);
                                Mapelement(uint8_t tc, const PolygonHole& po, const Glib::ustring nm = "");
                                Mapelement(uint8_t tc, const PolygonHole& po, const Point& lblcoord, const Glib::ustring nm = "");
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all) {}
                                unsigned int get_subtables(void) const { return ~0; }
                                bool is_subtables_loaded(void) const { return true; }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                bool is_valid(void) const { return m_typecode != maptyp_invalid; }
                                bool is_area(void) const;
                                Rect get_bbox(void) const { return m_bbox; }
                                Point get_labelcoord(void) const { return m_labelcoord; }
                                void set_labelcoord(const Point& pt) { m_labelcoord = pt; }
                                uint8_t get_typecode(void) const { return m_typecode; }
                                void set_typecode(uint8_t tc) { m_typecode = tc; }
                                static Glib::ustring get_typename(uint8_t tc);
                                Glib::ustring get_typename(void) const { return get_typename(get_typecode()); }
                                Glib::ustring get_name(void) const { return m_name; }
                                void set_name(const Glib::ustring& nm) { m_name = nm; }
                                const PolygonHole& get_polygon(void) const { return m_polygon; }
                                void set_polygon(const PolygonHole& poly) { m_polygon = poly; recompute_bbox_label(); }
                                const Polygon& get_exterior(void) const { return m_polygon.get_exterior(); }
                                unsigned int get_nrinterior(void) const { return m_polygon.get_nrinterior(); }
                                const Polygon& operator[](unsigned int x) const { return m_polygon[x]; }
                                Point get_center(void) const { return m_bbox.boxcenter(); }
                                Point get_coord(void) const { return get_labelcoord(); }

                        protected:
                                int64_t m_id;
                                Rect m_bbox;
                                uint8_t m_typecode;
                                table_t m_table;
                                Glib::ustring m_name;
                                Point m_labelcoord;
                                PolygonHole m_polygon;

                                void recompute_bbox_label(void);
                                int lon_center_poly(int32_t lat);
                                void recompute_label_coord(void);
                };

                class Track {
                        public:
                                class TrackPoint {
                                        public:
                                                typedef int16_t alt_t;
                                                static const unsigned int blob_size_v1 = 16;

                                                TrackPoint(void);
                                                time_t get_time(void) const { return m_time >> 8; }
                                                void set_time(time_t t) { m_time = ((uint64_t)t) << 8; }
                                                uint64_t get_timefrac8(void) const { return m_time; }
                                                void set_timefrac8(uint64_t t) { m_time = t; }
                                                Point get_coord(void) const { return m_coord; }
                                                void set_coord(const Point& pt) { m_coord = pt; }
                                                alt_t get_alt(void) const { return m_alt; }
                                                void set_alt(alt_t alt) { m_alt = alt; }
                                                bool is_alt_valid(void) const { return m_alt != altinvalid; }
                                                void set_alt_invalid(void) { m_alt = altinvalid; }

                                                bool operator<(const TrackPoint& b) const { return m_time < b.m_time; }

                                                void read_blob(const uint8_t *data);
                                                void write_blob(uint8_t *data) const;
                                                unsigned int blob_size(void) const { return blob_size_v1; }

                                        protected:
                                                static const alt_t altinvalid;
                                                Point m_coord;
                                                uint64_t m_time;
                                                alt_t m_alt;
                                };

                        protected:
                                typedef std::vector<TrackPoint> trackpoints_t;

                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef trackpoints_t::size_type size_type;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef enum {
                                        format_invalid = 0,
                                        format_empty = 1,
                                        format_gzip = 2
                                } format_t;

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_all = subtables_none;

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Track(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all) {}
                                unsigned int get_subtables(void) const { return ~0; }
                                bool is_subtables_loaded(void) const { return true; }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                Glib::ustring get_sourceid(void) const { return m_sourceid; }
                                void set_sourceid(const Glib::ustring& id) { m_sourceid = id; }
                                Glib::ustring get_fromicao(void) const { return m_fromicao; }
                                void set_fromicao(const Glib::ustring& ic) { m_fromicao = ic; }
                                Glib::ustring get_fromname(void) const { return m_fromname; }
                                void set_fromname(const Glib::ustring& nm) { m_fromname = nm; }
                                Glib::ustring get_fromicaoname(void) const;
                                Glib::ustring get_toicao(void) const { return m_toicao; }
                                void set_toicao(const Glib::ustring& ic) { m_toicao = ic; }
                                Glib::ustring get_toname(void) const { return m_toname; }
                                void set_toname(const Glib::ustring& nm) { m_toname = nm; }
                                Glib::ustring get_toicaoname(void) const;
                                Glib::ustring get_notes(void) const { return m_notes; }
                                void set_notes(const Glib::ustring& n) { m_notes = n; }
                                time_t get_offblocktime(void) const { return m_offblocktime; }
                                void set_offblocktime(time_t t) { m_offblocktime = t; }
                                time_t get_takeofftime(void) const { return m_takeofftime; }
                                void set_takeofftime(time_t t) { m_takeofftime = t; }
                                time_t get_landingtime(void) const { return m_landingtime; }
                                void set_landingtime(time_t t) { m_landingtime = t; }
                                time_t get_onblocktime(void) const { return m_onblocktime; }
                                void set_onblocktime(time_t t) { m_onblocktime = t; }
                                size_type size(void) const { return m_trackpoints.size(); }
                                const TrackPoint& operator[](size_type x) const { return m_trackpoints[x]; }
                                TrackPoint& operator[](size_type x) { return m_trackpoints[x]; }
                                TrackPoint& append(const TrackPoint& tp);
                                TrackPoint& insert(const TrackPoint& tp, int index);
                                void erase(int index);
                                void sort(void) { std::sort(m_trackpoints.begin(), m_trackpoints.end()); }
                                bool is_valid(void) const { return m_format != format_invalid; }
                                void make_valid(void);
                                Rect get_bbox(void) const { return m_bbox; }
                                void recompute_bbox(void);
                                void save_appendlog(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                size_type get_unsaved_points(void) const { return m_trackpoints.size() - m_savedtrackpoints; }
                                time_t get_modtime(void) const { return m_modtime; }
                                void set_modtime(time_t t = 0) { m_modtime = t; }

                        protected:
                                int64_t m_id;
                                Glib::ustring m_sourceid;
                                Glib::ustring m_fromicao;
                                Glib::ustring m_fromname;
                                Glib::ustring m_toicao;
                                Glib::ustring m_toname;
                                Glib::ustring m_notes;
                                time_t m_offblocktime;
                                time_t m_takeofftime;
                                time_t m_landingtime;
                                time_t m_onblocktime;
                                Rect m_bbox;
                                trackpoints_t m_trackpoints;
                                size_type m_savedtrackpoints;
                                format_t m_format;
                                table_t m_table;
                                time_t m_modtime;
                };

                class Label {
                        public:
                                typedef ::DbBaseElements::ElementAddress Address;
                                typedef Address::table_t table_t;
                                static const table_t table_main = Address::table_main;
                                static const table_t table_aux = Address::table_aux;
                                typedef ::DbBaseElements::TileNumber TileNumber;
                                typedef TileNumber::tile_t tile_t;
                                typedef TileNumber::tilerange_t tilerange_t;

                                typedef ::DbBaseElements::Navaid::label_placement_t label_placement_t;
                                static const label_placement_t label_n = ::DbBaseElements::Navaid::label_n;
                                static const label_placement_t label_e = ::DbBaseElements::Navaid::label_e;
                                static const label_placement_t label_s = ::DbBaseElements::Navaid::label_s;
                                static const label_placement_t label_w = ::DbBaseElements::Navaid::label_w;
                                static const label_placement_t label_center = ::DbBaseElements::Navaid::label_center;
                                static const label_placement_t label_any = ::DbBaseElements::Navaid::label_any;
                                static const label_placement_t label_off = ::DbBaseElements::Navaid::label_off;

                                static const unsigned int subtables_none = 0;
                                static const unsigned int subtables_all = subtables_none;

                                static const char *db_query_string;
                                static const char *db_aux_query_string;
                                static const char *db_text_fields[];
                                static const char *db_time_fields[];

                                Label(void);
                                void load(sqlite3x::sqlite3_cursor& cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables = subtables_all);
                                void load_subtables(sqlite3x::sqlite3_connection& db, unsigned int loadsubtables = subtables_all) {}
                                unsigned int get_subtables(void) const { return ~0; }
                                bool is_subtables_loaded(void) const { return true; }
                                void save(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void erase(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                void update_index(sqlite3x::sqlite3_connection& db, bool rtree, bool aux_rtree);
                                int64_t get_id(void) const { return m_id; }
                                table_t get_table(void) const { return m_table; }
                                Address get_address(void) const { return Address(m_id, m_table); }
                                bool get_mutable(void) const { return m_mutable; }
                                void set_mutable(bool mut) { m_mutable = mut; }
                                Point get_coord(void) const { return m_coord; }
                                void set_coord(const Point& pt) { m_coord = pt; }
                                label_placement_t get_label_placement(void) const { return m_label_placement; }
                                void set_label_placement(label_placement_t lp = label_off) { m_label_placement = lp; }
                                double get_metric(void) const { return m_metric; }
                                void set_metric(double metric) { m_metric = metric; }
                                int64_t get_subid(void) const { return m_subid; }
                                void set_subid(int64_t id) { m_subid = id; }
                                table_t get_subtable(void) const { return m_subtable; }
                                void set_subtable(table_t tbl) { m_subtable = tbl; }
                                bool is_valid(void) const { return m_label_placement != label_off; }

                        protected:
                                int64_t m_id;
                                table_t m_table;
                                bool m_mutable;
                                Point m_coord;
                                label_placement_t m_label_placement;
                                double m_metric;
                                int64_t m_subid;
                                table_t m_subtable;
                };
};

std::ostream& operator<<(std::ostream& os, const DbBaseElements::ElementAddress& addr);

class DbBaseCommon : protected DbBaseElements {
        public:
                typedef enum {
                        comp_startswith = FPlan::comp_startswith,
                        comp_contains = FPlan::comp_contains,
                        comp_exact = FPlan::comp_exact,
                        comp_exact_casesensitive,
                        comp_like
                } comp_t;

                DbBaseCommon(void);
                ~DbBaseCommon(void);
                void close(void);
                void detach(void);
                void sync_off(void);
                void analyze(void);
                void vacuum(void);
                unsigned int query_cache_size(void);
                void set_cache_size(unsigned int cs);
                bool is_open(void) const { return m_open != openstate_closed; }
                bool is_aux_attached(void) const { return m_open == openstate_auxopen; }
                void interrupt(void);

        protected:
                typedef enum {
                        openstate_closed,
                        openstate_mainopen,
                        openstate_auxopen
                } openstate_t;
                sqlite3x::sqlite3_connection m_db;
                openstate_t m_open;
                bool m_has_rtree;
                bool m_has_aux_rtree;

                static void dbfunc_simpledist(sqlite3_context *ctxt, int, sqlite3_value **values);
                static void dbfunc_simplerectdist(sqlite3_context *ctxt, int, sqlite3_value **values);
                static void dbfunc_upperbound(sqlite3_context *ctxt, int, sqlite3_value **values);

                static Glib::ustring quote_text_for_like(const Glib::ustring& s, char e = 0);
};

template <class T> class DbBase : public DbBaseCommon {
        protected:
                DbBase();

        public:
                typedef T element_t;
                typedef std::vector<element_t> elementvector_t;
                typedef enum element_t::table_t table_t;
                typedef typename element_t::Address Address;

                void open(const Glib::ustring& path = "");
                void attach(const Glib::ustring& path = "");
                static bool is_dbfile_exists(const Glib::ustring& path = "");
                void take(sqlite3 *dbh);
                void purgedb(void);
                element_t operator()(int64_t id, table_t tbl = element_t::table_main, unsigned int loadsubtables = element_t::subtables_all);
                element_t operator()(Address addr, unsigned int loadsubtables = element_t::subtables_all) { return operator()(addr.get_id(), addr.get_table(), loadsubtables); }
                void loadfirst(element_t& e, bool include_aux = true, unsigned int loadsubtables = element_t::subtables_all);
                void loadnext(element_t& e, bool include_aux = true, unsigned int loadsubtables = element_t::subtables_all);
                void save(element_t& e) { e.save(m_db, m_has_rtree, m_has_aux_rtree); }
                void erase(element_t& e) { e.erase(m_db, m_has_rtree, m_has_aux_rtree); }
                void update_index(element_t& e) { e.update_index(m_db, m_has_rtree, m_has_aux_rtree); }

                class ForEach {
                        public:
                                virtual ~ForEach() {}
                                virtual bool operator()(const element_t&) = 0;
                                virtual bool operator()(const std::string&) = 0;
                };

                void for_each(ForEach& cb, bool include_aux = true, unsigned int loadsubtables = element_t::subtables_all);
                void for_each_by_rect(ForEach& cb, const Rect& r, bool include_aux = true, unsigned int loadsubtables = element_t::subtables_all);
                elementvector_t find_by_text(const std::string& pattern, char escape, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all);
                elementvector_t find_by_time(time_t fromtime, time_t totime, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all);
                elementvector_t find_by_rect(const Rect& r, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all);
                elementvector_t find_nearest(const Point& pt, const Rect& r, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all);

        protected:
                static const char *main_table_name;
                static const char *database_file_name;
                static const char *order_field;
                static const char *delete_field;
                static const bool area_data;

                void drop_tables(void);
                void create_tables(void);
                void drop_indices(void);
                void create_indices(void);

                std::string get_area_select_string(bool auxtable, const Rect& r, const char *sortcol = 0);
                elementvector_t find(const std::string& fldname, const std::string& pattern, char escape, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all);
                elementvector_t loadid(uint64_t startid = 0, table_t table = element_t::table_main, const std::string& order = "", unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all);
};

class NavaidsDb : public DbBase<DbBaseElements::Navaid> {
        public:
                typedef element_t Navaid;

                void take(NavaidsDb& db) { DbBase<element_t>::take(db.m_db.take()); }

                elementvector_t find_by_sourceid(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("SRCID", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_icao(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("ICAO", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("NAME", pattern, escape, comp, limit, loadsubtables); }
};

class WaypointsDb : public DbBase<DbBaseElements::Waypoint> {
        public:
                typedef element_t Waypoint;

                void take(WaypointsDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                elementvector_t find_by_sourceid(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("SRCID", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_icao(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("ICAO", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("NAME", pattern, escape, comp, limit, loadsubtables); }
};

class AirwaysDb : public DbBase<DbBaseElements::Airway> {
        public:
                typedef element_t Airway;

                void take(AirwaysDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                elementvector_t find_by_sourceid(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("SRCID", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_begin_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("BNAME", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_end_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("ENAME", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("NAME", pattern, escape, comp, limit, loadsubtables); }
};

class AirportsDb : public DbBase<DbBaseElements::Airport> {
        public:
                typedef element_t Airport;

                void take(AirportsDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                elementvector_t find_by_sourceid(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("SRCID", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_icao(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("ICAO", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("NAME", pattern, escape, comp, limit, loadsubtables); }
};

class AirspacesDb : public DbBase<DbBaseElements::Airspace> {
        public:
                typedef element_t Airspace;

                void take(AirspacesDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                elementvector_t find_by_sourceid(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("SRCID", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_icao(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("ICAO", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("NAME", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_ident(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("IDENT", pattern, escape, comp, limit, loadsubtables); }
};

class MapelementsDb : public DbBase<DbBaseElements::Mapelement> {
        public:
                typedef element_t Mapelement;

                void take(MapelementsDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                elementvector_t find_by_name(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("NAME", pattern, escape, comp, limit, loadsubtables); }
};

class TracksDb : public DbBase<DbBaseElements::Track> {
        public:
                typedef element_t Track;

                void open(const Glib::ustring& path = "");
                void clear_appendlogs(void);
                void take(TracksDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                void save_appendlog(element_t& e) { e.save_appendlog(m_db, m_has_rtree, m_has_aux_rtree); }
                elementvector_t find_by_sourceid(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("SRCID", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_fromicao(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("FROMICAO", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_fromname(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("FROMNAME", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_toicao(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("TOICAO", pattern, escape, comp, limit, loadsubtables); }
                elementvector_t find_by_toname(const std::string& pattern, char escape = 0, comp_t comp = comp_exact, unsigned int limit = 0, unsigned int loadsubtables = element_t::subtables_all) { return find("TONAME", pattern, escape, comp, limit, loadsubtables); }
};

/* a simple label database, used for computing label placements */
class LabelsDb : public DbBase<DbBaseElements::Label> {
        public:
                typedef element_t Label;

                void take(LabelsDb& db) { DbBase<element_t>::take(db.m_db.take()); }
                elementvector_t find_mutable_by_metric(unsigned int limit = 0, unsigned int offset = 0, unsigned int loadsubtables = element_t::subtables_all);
};

template<int resolution, int tilesize>
class TopoDbN {
        public:
                typedef unsigned int tile_index_t;
                typedef unsigned int pixel_index_t;
                typedef std::pair<tile_index_t,pixel_index_t> tile_pixel_index_t;
                static const pixel_index_t tile_size = tilesize;

                class TopoCoordinate;

                class TopoTileCoordinate {
                        public:
                                static const uint16_t lon_tiles = 360 * 60 * 60 / resolution / tilesize;
                                static const uint16_t lat_tiles = lon_tiles / 2 + 1;

                                TopoTileCoordinate(uint16_t lon_tile = 0, uint16_t lat_tile = 0, uint16_t lon_offs = 0, uint16_t lat_offs = 0) : m_lon_tile(lon_tile), m_lon_offs(lon_offs), m_lat_tile(lat_tile), m_lat_offs(lat_offs) {}
                                TopoTileCoordinate(const TopoCoordinate& tc);
                                operator TopoCoordinate(void) const { return TopoCoordinate(m_lon_tile * tilesize + m_lon_offs, m_lat_tile * tilesize + m_lat_offs); }
                                uint16_t get_lon_tile(void) const { return m_lon_tile; }
                                uint16_t get_lon_offs(void) const { return m_lon_offs; }
                                uint16_t get_lat_tile(void) const { return m_lat_tile; }
                                uint16_t get_lat_offs(void) const { return m_lat_offs; }
                                void set_lon_tile(uint16_t lon_tile) { m_lon_tile = lon_tile; }
                                void set_lon_offs(uint16_t lon_offs) { m_lon_offs = lon_offs; }
                                void set_lat_tile(uint16_t lat_tile) { m_lat_tile = lat_tile; }
                                void set_lat_offs(uint16_t lat_offs) { m_lat_offs = lat_offs; }
                                tile_index_t get_tile_index(void) const { return m_lat_tile * (tile_index_t)lon_tiles + m_lon_tile; }
                                pixel_index_t get_pixel_index(void) const { return m_lat_offs * tilesize + m_lon_offs; }
                                tile_pixel_index_t get_tile_pixel_index(void) const { return tile_pixel_index_t(get_tile_index(), get_pixel_index()); }
                                void advance_tile_east(void);
                                void advance_tile_west(void);
                                void advance_tile_north(void);
                                void advance_tile_south(void);

                        private:
                                uint16_t m_lon_tile;
                                uint16_t m_lon_offs;
                                uint16_t m_lat_tile;
                                uint16_t m_lat_offs;
                };

                class TopoCoordinate {
                        public:
                                typedef uint32_t coord_t;
                                TopoCoordinate(coord_t lon = 0, coord_t lat = 0) : m_lon(lon), m_lat(lat) {}
                                TopoCoordinate(const Point& pt);
                                operator Point(void) const;
                                coord_t get_lon(void) const { return m_lon; }
                                coord_t get_lat(void) const { return m_lat; }
                                void set_lon(coord_t lon) { m_lon = lon; }
                                void set_lat(coord_t lat) { m_lat = lat; }
                                coord_t lon_dist(const TopoCoordinate& tc) const;
                                coord_t lat_dist(const TopoCoordinate& tc) const;
#if 0
                                bool operator==(const TopoCoordinate &p) const { return m_lon == p.m_lon && m_lat == p.m_lat; }
                                bool operator!=(const TopoCoordinate &p) const { return !operator==(p); }
                                TopoCoordinate operator+(const TopoCoordinate &p) const { return TopoCoordinate(m_lon + p.m_lon, m_lat + p.m_lat); }
                                TopoCoordinate operator-(const TopoCoordinate &p) const { return TopoCoordinate(m_lon - p.m_lon, m_lat - p.m_lat); }
                                TopoCoordinate operator-() const { return Point(-m_lon, -m_lat); }
                                TopoCoordinate& operator+=(const TopoCoordinate &p) { m_lon += p.m_lon; m_lat += p.m_lat; return *this; }
                                TopoCoordinate& operator-=(const TopoCoordinate &p) { m_lon -= p.m_lon; m_lat -= p.m_lat; return *this; }
                                bool operator<(const TopoCoordinate &p) const { return m_lon < p.m_lon && m_lat < p.m_lat; }
                                bool operator<=(const TopoCoordinate &p) const { return m_lon <= p.m_lon && m_lat <= p.m_lat; }
                                bool operator>(const TopoCoordinate &p) const { return m_lon > p.m_lon && m_lat > p.m_lat; }
                                bool operator>=(const TopoCoordinate &p) const { return m_lon >= p.m_lon && m_lat >= p.m_lat; }
#endif
                                void advance_east(void);
                                void advance_west(void);
                                void advance_north(void);
                                void advance_south(void);
                                static Point get_pointsize(void) { return (Point)TopoCoordinate(1, 1) - (Point)TopoCoordinate(0, 0); }
                                std::ostream& print(std::ostream& os);

                        protected:
                                coord_t m_lon;
                                coord_t m_lat;

                                static const coord_t from_point = 360 * 60 * 60 / resolution;
                                static const uint64_t to_point = (((((uint64_t)1) << 63) + from_point / 2) / from_point) << 1;
                };

                typedef int16_t elev_t;
                typedef std::pair<elev_t,elev_t> minmax_elev_t;
                static const elev_t nodata;
                static const elev_t ocean;

                class Tile {
                        public:
                                Tile(tile_index_t index);
                                Tile(const Tile& tile);
                                Tile(sqlite3x::sqlite3_connection& db, tile_index_t index);
                                ~Tile(void);
                                void save(sqlite3x::sqlite3_connection& db);
                                tile_index_t get_index(void) const { return m_index; }
                                elev_t get_elev(pixel_index_t index) const;
                                void set_elev(pixel_index_t index, elev_t elev);
                                elev_t get_minelev(void) { if (m_minmaxinvalid) update_minmax(); return m_minelev; }
                                elev_t get_maxelev(void) { if (m_minmaxinvalid) update_minmax(); return m_maxelev; }
                                bool is_dirty(void) const { return m_dirty; }

                        protected:
                                tile_index_t m_index;
                                elev_t m_minelev;
                                elev_t m_maxelev;
                                elev_t *m_elev;
                                bool m_dirty;
                                bool m_loaded;
                                bool m_minmaxinvalid;

                                void update_minmax(void);
                };

                class RGB {
                        public:
                                RGB(uint8_t r = 0, uint8_t g = 0, uint8_t b = 0) : m_col(0xff000000 | (r << 16) | (g << 8) | b) {}
                                operator Gdk::Color(void) const;
                                operator uint32_t(void) const { return m_col; }

                        private:
                                uint32_t m_col;
                };

                TopoDbN();
                void take(sqlite3 *dbh);
                void close(void);
                void purgedb(void);
                void sync_off(void);
                void analyze(void);
                void vacuum(void);
                void interrupt(void);
                Tile *find(tile_index_t index);
                void cache_commit(void);
                elev_t get_elev(const TopoTileCoordinate& tc);
                void set_elev(const TopoTileCoordinate& tc, elev_t elev);
                elev_t get_elev(const TopoCoordinate& pt) { return get_elev((TopoTileCoordinate)pt); }
                void set_elev(const TopoCoordinate& pt, elev_t elev) { set_elev((TopoTileCoordinate)pt, elev); }
                elev_t get_elev(const Point& pt) { return get_elev((TopoCoordinate)pt); }
                void set_elev(const Point& pt, elev_t elev) { set_elev((TopoCoordinate)pt, elev); }
                minmax_elev_t get_tile_minmax_elev(tile_index_t index);
                minmax_elev_t get_tile_minmax_elev(const TopoTileCoordinate& tc) { return get_tile_minmax_elev(tc.get_tile_index()); }
                minmax_elev_t get_minmax_elev(const Rect& r);
                minmax_elev_t get_minmax_elev(const Polygon& p);
                static RGB color(elev_t elev);
                static int m_to_ft(elev_t elev);

        protected:
                static const unsigned int tile_cache_size = 16;
                static const tile_index_t lon_tiles = 360 * 60 * 60 / resolution / tilesize;
                static const tile_index_t lat_tiles = lon_tiles / 2 + 1;
                static const tile_index_t nr_tiles = lon_tiles * lat_tiles;
                sqlite3x::sqlite3_connection m_db;
                typedef std::list<Tile> tile_cache_t;
                tile_cache_t m_tile_cache;

                void open(const Glib::ustring& path, const Glib::ustring& dbname);
                void drop_tables(void);
                void create_tables(void);
                void drop_indices(void);
                void create_indices(void);
};

template<int resolution, int tilesize>
std::ostream& operator<<(std::ostream& os, const typename TopoDbN<resolution,tilesize>::TopoCoordinate& tc);

class TopoDb30 : public TopoDbN<30,480> {
        private:
                typedef TopoDbN<30,480> TopoDbNN;

        public:
                //typedef TopoDbNN::TopoCoordinate TopoCoordinate;
                //typedef TopoDbNN::elev_t elev_t;
                //static const elev_t nodata;
                void open(const Glib::ustring& path);
};

#endif /* _DBOBJ_H */
