//
// C++ Implementation: fplan
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include <libxml++/libxml++.h>
#include <glibmm.h>
#ifdef HAVE_GIOMM
#include <giomm.h>
#endif
#include <unistd.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <sqlite3x.hpp>
#include <sqlite3.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdexcept>
#include <stdarg.h>
#include <limits>
#include <cstring>
#include <glib.h>
//#include <glib/gstdio.h>

#include "fplan.h"
#include "baro.h"
#include "wmm.h"

FPlanLeg::FPlanLeg( void )
        : m_fromicao(), m_toicao(), m_fromname(), m_toname(), m_fromcoord(), m_tocoord(),
          m_fromalt(0), m_toalt(0), m_fromtruealt(0), m_totruealt(0), m_fromflags(0), m_toflags(0),
          m_fromfreq(0), m_tofreq(0), m_fromtime(0), m_legtime(0),
          m_dist(0), m_vdist(0), m_truetrack(0), m_var(0), m_wca(0),
          m_legtype(legtype_invalid), m_future(false),
          m_crosstrack(0), m_disttodest(0), m_vgnd(0), m_legfraction(0),
          m_altdiff(0), m_referencealt(0), m_mapalt(0),
          m_mapcoord(), m_coursetodest(0), m_gpsvalid(false), m_gpsaltvalid(false)
{
}

FPlanLeg::FPlanLeg(const Glib::ustring& fromicao, const Glib::ustring & fromname, const Point & fromcoord,
                   int16_t fromalt, uint16_t fromflags, uint32_t fromtime, uint32_t fromfreq,
                   const Glib::ustring& toicao, const Glib::ustring & toname, const Point & tocoord,
                   int16_t toalt, uint16_t toflags, uint32_t totime, uint32_t tofreq,
                   legtype_t legtype, bool future)
        : m_fromicao(fromicao), m_toicao(toicao), m_fromname(fromname), m_toname(toname), m_fromcoord(fromcoord), m_tocoord(tocoord),
          m_fromalt(fromalt), m_toalt(toalt), m_fromtruealt(fromalt), m_totruealt(toalt), m_fromflags(fromflags), m_toflags(toflags),
          m_fromfreq(fromfreq), m_tofreq(tofreq), m_fromtime(fromtime), m_legtime(totime - fromtime),
          m_dist(0), m_vdist(0), m_truetrack(0), m_var(0), m_wca(0),
          m_legtype(legtype), m_future(future),
          m_crosstrack(0), m_disttodest(0), m_vgnd(0), m_legfraction(0),
          m_altdiff(0), m_referencealt(0), m_mapalt(0),
          m_mapcoord(), m_coursetodest(0), m_gpsvalid(false), m_gpsaltvalid(false)
{
        m_dist = m_fromcoord.spheric_distance_nmi(m_tocoord);
        m_truetrack = (uint16_t)(m_fromcoord.spheric_true_course(m_tocoord) * from_deg);
        WMM wmm;
        if (wmm.compute(get_fromaltitude() * (FPlanWaypoint::ft_to_m * 1e-3), get_fromcoord(), get_fromtime_unix()))
                m_var = (uint16_t)(wmm.get_dec() * from_deg);
}

Glib::ustring FPlanLeg::get_fromicaoname( void ) const
{
        Glib::ustring r(m_fromicao);
        if (!m_fromicao.empty() && !m_fromname.empty())
                r += " ";
        r += m_fromname;
        return r;
}

Glib::ustring FPlanLeg::get_toicaoname( void ) const
{
        Glib::ustring r(m_toicao);
        if (!m_toicao.empty() && !m_toname.empty())
                r += " ";
        r += m_toname;
        return r;
}

void FPlanLeg::baro_correction(IcaoAtmosphere& qnh)
{
        m_fromtruealt = m_fromalt;
        m_totruealt = m_toalt;
        if (!((m_fromflags | m_toflags) & FPlanWaypoint::altflag_standard))
                return;
        IcaoAtmosphere std;
        float press, alt;
        if (m_fromflags & FPlanWaypoint::altflag_standard) {
                std.altitute_to_pressure(&press, 0, get_fromaltitude() * FPlanWaypoint::ft_to_m);
                qnh.pressure_to_altitute(&alt, 0, press);
                m_fromtruealt = (int16_t)(alt * FPlanWaypoint::m_to_ft);
        }
        if (m_toflags & FPlanWaypoint::altflag_standard) {
                std.altitute_to_pressure(&press, 0, get_toaltitude() * FPlanWaypoint::ft_to_m);
                qnh.pressure_to_altitute(&alt, 0, press);
                m_totruealt = (int16_t)(alt * FPlanWaypoint::m_to_ft);
        }
}

int16_t FPlanLeg::baro_correction(IcaoAtmosphere & qnh, int16_t alt, uint16_t altflags)
{
        if (!(altflags & FPlanWaypoint::altflag_standard))
                return alt;
        IcaoAtmosphere std;
        float press, alt1;
        std.altitute_to_pressure(&press, 0, alt * FPlanWaypoint::ft_to_m);
        qnh.pressure_to_altitute(&alt1, 0, press);
        return Point::round<int16_t,float>(alt1 * FPlanWaypoint::m_to_ft);
}

/*
 * Winddreieck:
 * a = vdist
 * b = vwind
 * c = vcruise
 * alpha = ?
 * beta = WCA
 * gamma = 180 + TT - winddir
 *
 * Sinussatz: vwind/sin(WCA) = vcruise/sin(winddir-TT) = vdist/sin(winddir-TT-WCA)
 * Cosinussatz: vdist^2 = vwind^2 + vcruise^2 - 2*vwind*vcruise*cos(winddir-TT-WCA)
 */

void FPlanLeg::wind_correction(float vcruise_kts, float winddir_deg, float vwind_kts)
{
        uint16_t wdirminustt = (uint16_t)(winddir_deg * from_deg);
        wdirminustt -= m_truetrack;
        m_wca = (uint16_t)(from_rad * asin(sin(wdirminustt * to_rad) * vwind_kts / vcruise_kts));
        m_vdist = sqrt(vwind_kts * vwind_kts + vcruise_kts * vcruise_kts -
                       2.0 * vwind_kts * vcruise_kts * cos(to_rad * (wdirminustt - m_wca)));
        m_vgnd = m_vdist;
        switch (m_legtype) {
                case legtype_offblock_takeoff:
                case legtype_landing_onblock:
                        m_legtime = 5 * 60;
                        break;

                case legtype_normal:
                        m_legtime = (uint32_t)(3600 * m_dist / std::max(m_vdist, 0.1f));
                        break;

                default:
                        m_legtime = 0;
                        break;
        }
}

void FPlanLeg::update( time_t curtime, const Point & coord, bool coordvalid, int16_t alt, bool altvalid, float vgnd )
{
        m_crosstrack = m_disttodest = 0.0;
        m_legfraction = 0;
        m_altdiff = m_mapalt = 0;
        m_coursetodest = m_truetrack;
        m_gpsvalid = m_gpsaltvalid = false;
        CartesianVector3D cfrom = PolarVector3D(m_fromcoord, 0);
        CartesianVector3D cto = PolarVector3D(m_tocoord, 0);
        m_gpsvalid = coordvalid;
        m_gpsaltvalid = coordvalid && altvalid;
        float t = 0;
        if (m_gpsvalid) {
                m_mapcoord = coord;
                m_vgnd = vgnd;
                if (get_legtype() == legtype_normal) {
                        CartesianVector3D cm = PolarVector3D(m_mapcoord, 0);
                        t = cm.nearest_fraction(cfrom, cto);
                        CartesianVector3D cproj = cfrom + (cto - cfrom) * t;
                        m_crosstrack = (cm - cproj).length();
                        m_coursetodest = (uint16_t)(m_mapcoord.spheric_true_course(m_tocoord) * from_deg);
                        if ((int16_t)(m_coursetodest - m_truetrack) < 0)
                                m_crosstrack = -m_crosstrack;
                }
        } else {
                if (!m_legtime) {
                        t = 0.5;
                } else {
                        t = curtime - get_fromtime_unix();
                        t = std::min(std::max(t / m_legtime, 0.0f), 1.0f);
                }
                m_mapcoord = ((PolarVector3D)(cfrom + (cto - cfrom) * t)).getcoord();
                m_vgnd = m_vdist;
        }
        m_referencealt = m_fromtruealt + (int16_t)(t * (m_totruealt - m_fromtruealt));
        m_legfraction = (uint16_t)((1U << 15) * t);
        m_mapalt = m_gpsaltvalid ? alt : m_referencealt;
        m_altdiff = m_mapalt - m_referencealt;
        m_disttodest = m_mapcoord.spheric_distance_nmi(m_tocoord);
        WMM wmm;
        if (wmm.compute(get_mapalt() * (FPlanWaypoint::ft_to_m * 1e-3), get_mapcoord(), curtime))
                m_var = (uint16_t)(wmm.get_dec() * from_deg);
}

float FPlanLeg::get_verticalspeedtodest(void) const
{
        return (get_totruealtitude() - get_mapalt()) / std::max(1.0f / 60.0f, 60.0f * get_disttodest() / std::max(get_vgnd(), 1.0f));
}

float FPlanLeg::get_verticalspeed(void) const
{
        return (get_totruealtitude() - get_fromtruealtitude()) / std::max(1.0f / 60.0f, 60.0f * get_dist() / std::max(get_vgnd(), 1.0f));
}

FPlanWaypoint::FPlanWaypoint(void)
        : icao(), name(), time(0), savetime(0), frequency(0), coord(), alt(0), flags(0), dirty(false)
{
}

const std::string FPlanWaypoint::reader_fields = "ICAO,NAME,LON,LAT,ALT,FREQ,TIME,SAVETIME,FLAGS,NOTE";

FPlanWaypoint::FPlanWaypoint( sqlite3x::sqlite3_cursor & cursor )
        : icao(cursor.getstring(0)), name(cursor.getstring(1)), note(cursor.getstring(9)),
          time(cursor.getint64(6)), savetime(cursor.getint64(7)), frequency(cursor.getint(5)),
          coord(Point(cursor.getint(2), cursor.getint(3))), alt(cursor.getint(4)), flags(cursor.getint(8)), dirty(false)
{
}

Glib::ustring FPlanWaypoint::get_icao_name( void ) const
{
        Glib::ustring r(icao);
        if (!icao.empty() && !name.empty())
                r += " ";
        r += name;
        return r;
}

FPlanRoute::FPlanRoute(FPlan & fpp)
        : fp(fpp), id(0), note(), time_offblock(0), savetime_offblock(0), time_onblock(0), savetime_onblock(0), curwpt(0), wpts(), dirty(false)
{
}

FPlanWaypoint& FPlanRoute::operator[](uint32_t nr)
{
        if (nr >= wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        return wpts[nr];
}

const FPlanWaypoint& FPlanRoute::operator[](uint32_t nr) const
{
        if (nr >= wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        return wpts[nr];
}

void FPlanRoute::insert_wpt(uint32_t nr, const FPlanWaypoint & wpt)
{
        if (nr >= wpts.size()) {
                wpts.push_back(wpt);
                return;
        }
        wpts.insert(wpts.begin() + nr, wpt);
        dirty = true;
        if (id) {
                sqlite3x::sqlite3_transaction tran(fp.get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "UPDATE waypoints SET NR = NR + 100000 WHERE PLANID = ? AND NR >= ?;");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "UPDATE waypoints SET NR = NR - 99999 WHERE PLANID = ? AND NR >= ?;");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "INSERT INTO waypoints (PLANID,NR) VALUES (?,?);");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                tran.commit();
        }
}

void FPlanRoute::delete_wpt(uint32_t nr)
{
        if (nr >= wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        if (id) {
                sqlite3x::sqlite3_transaction tran(fp.get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "DELETE FROM waypoints WHERE PLANID = ? AND NR = ?;");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "UPDATE waypoints SET NR = NR + 99999 WHERE PLANID = ? AND NR > ?;");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "UPDATE waypoints SET NR = NR - 100000 WHERE PLANID = ? AND NR > ?;");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        wpts.erase(wpts.begin() + nr);
        dirty = true;
}

bool FPlanRoute::is_dirty(void) const
{
        if (dirty)
                return true;
        for (waypoints_t::const_iterator i(wpts.begin()), e(wpts.end()); i != e; i++)
                if ((*i).dirty)
                        return true;
        return false;
}

void FPlanRoute::set_dirty(bool d)
{
        dirty = d;
        for (waypoints_t::iterator i(wpts.begin()), e(wpts.end()); i != e; i++)
                (*i).dirty = d;
}

void FPlanRoute::load_fp(id_t nid)
{
        id = nid;
	load_fp();
}

void FPlanRoute::load_fp(void)
{
        wpts.clear();
        {
                sqlite3x::sqlite3_command cmd(fp.get_db(), "SELECT TIMEOFFBLOCK,SAVETIMEOFFBLOCK,TIMEONBLOCK,SAVETIMEONBLOCK,"
                                              "CURWPT,NOTE FROM fplan WHERE ID = ?;");
                cmd.bind(1, id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        throw std::runtime_error("load_fp: plan ID not found");
                time_offblock = cursor.getint64(0);
                savetime_offblock = cursor.getint64(1);
                time_onblock = cursor.getint64(2);
                savetime_onblock = cursor.getint64(3);
                curwpt = cursor.getint(4);
                note = cursor.getstring(5);

        }
        {
                sqlite3x::sqlite3_command cmd(fp.get_db(), "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE PLANID = ? ORDER BY NR ASC;");
                cmd.bind(1, id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        dirty = false;
}

void FPlanRoute::save_fp(void)
{
        if (!is_dirty())
                return;
        if (!id) {
                {
                        //sqlite3x::sqlite3_command cmd(fp.get_db(), "SELECT ID FROM fplan ORDER BY ID DESC LIMIT 1;");
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "SELECT MAX(ID) FROM fplan;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                id = cursor.getint(0) + 1;
                        else
                                id = 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "INSERT INTO fplan (ID) VALUES(?);");
                        cmd.bind(1, id);
                        cmd.executenonquery();
                }
        }
        sqlite3x::sqlite3_transaction tran(fp.get_db(), true);
        {
                sqlite3x::sqlite3_command cmd(fp.get_db(), "DELETE FROM waypoints WHERE PLANID = ? AND (NR < 0 OR NR >= ?);");
                cmd.bind(1, id);
                cmd.bind(2, get_nrwpt());
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(fp.get_db(), "UPDATE fplan SET TIMEOFFBLOCK = ?,"
                                              "SAVETIMEOFFBLOCK = ?, TIMEONBLOCK = ?, SAVETIMEONBLOCK = ?,"
                                              "CURWPT = ?, NOTE = ? WHERE ID = ?;");
                cmd.bind(1, (long long int)time_offblock);
                cmd.bind(2, (long long int)savetime_offblock);
                cmd.bind(3, (long long int)time_onblock);
                cmd.bind(4, (long long int)savetime_onblock);
                cmd.bind(5, (int)curwpt);
                cmd.bind(6, note);
                cmd.bind(7, id);
                cmd.executenonquery();
        }
        for (unsigned int nr = 0; nr < get_nrwpt(); nr++) {
                FPlanWaypoint& w(wpts[nr]);
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "INSERT OR IGNORE INTO waypoints (PLANID,NR) VALUES(?,?);");
                        cmd.bind(1, id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "UPDATE waypoints SET ICAO  = ?, NAME = ?, "
                                                      "LON = ?, LAT = ?, ALT = ?, FREQ = ?, TIME = ?, SAVETIME = ?,"
                                                      "FLAGS = ?, NOTE = ? WHERE PLANID = ? AND NR = ?;");
                        cmd.bind(1, w.get_icao());
                        cmd.bind(2, w.get_name());
                        cmd.bind(3, w.get_lon());
                        cmd.bind(4, w.get_lat());
                        cmd.bind(5, w.get_altitude());
                        cmd.bind(6, (int)w.get_frequency());
                        cmd.bind(7, (long long int)w.get_time());
                        cmd.bind(8, (long long int)w.get_save_time());
                        cmd.bind(9, w.get_flags());
                        cmd.bind(10, w.get_note());
                        cmd.bind(11, id);
                        cmd.bind(12, (int)nr);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        set_dirty(false);
}

void FPlanRoute::duplicate_fp(void)
{
        id = 0;
        dirty = true;
}

void FPlanRoute::new_fp(void)
{
        duplicate_fp();
        wpts.clear();
        note.clear();
        time_offblock = 0;
        savetime_offblock = 0;
        time_onblock = 0;
        savetime_onblock = 0;
        curwpt = 0;
        set_dirty(false);
}

void FPlanRoute::delete_fp(void)
{
        if (id) {
                sqlite3x::sqlite3_transaction tran(fp.get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "DELETE FROM waypoints WHERE PLANID = ?;");
                        cmd.bind(1, id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(fp.get_db(), "DELETE FROM fplan WHERE ID = ?;");
                        cmd.bind(1, id);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        new_fp();
}

void FPlanRoute::reverse_fp(void )
{
        if (wpts.empty())
                return;
        reverse(wpts.begin(), wpts.end());
        set_dirty(true);
}

bool FPlanRoute::load_first_fp( void )
{
        {
                sqlite3x::sqlite3_command cmd(fp.get_db(), "SELECT ID FROM fplan ORDER BY ID ASC LIMIT 1;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        return false;
                id = cursor.getint(0);
        }
        load_fp(id);
        return true;
}

bool FPlanRoute::load_next_fp( void )
{
        {
                sqlite3x::sqlite3_command cmd(fp.get_db(), "SELECT ID FROM fplan WHERE ID > ? ORDER BY ID ASC LIMIT 1;");
                cmd.bind(1, id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        return false;
                id = cursor.getint(0);
        }
        load_fp(id);
        return true;
}

float FPlanRoute::total_distance_km( void ) const
{
        float sum = 0;
        for (int i = wpts.size() - 1; i > 0; i--)
               sum += wpts[i-1].get_coord().spheric_distance_km(wpts[i].get_coord());
        return sum;
}

int16_t FPlanRoute::max_altitude( void ) const
{
        int16_t a = std::numeric_limits<int16_t>::min();
        for (int i = wpts.size() - 1; i >= 0; i--)
                a = std::max(a, wpts[i].get_altitude());
        return a;
}

void FPlanRoute::save_time( uint32_t nr )
{
        if (!nr) {
                save_time_offblock();
                return;
        }
        if (nr > wpts.size()) {
                save_time_onblock();
                return;
        }
        wpts[nr-1].save_time();
}

void FPlanRoute::restore_time( uint32_t nr )
{
        if (!nr) {
                restore_time_offblock();
                return;
        }
        if (nr > wpts.size()) {
                restore_time_onblock();
                return;
        }
        wpts[nr-1].restore_time();
}

void FPlanRoute::set_time( uint32_t nr, uint32_t t )
{
        if (!nr) {
                set_time_offblock(t);
                return;
        }
        if (nr > wpts.size()) {
                set_time_onblock(t);
                return;
        }
        wpts[nr-1].set_time(t);
}

uint32_t FPlanRoute::get_time( uint32_t nr ) const
{
        if (!nr)
                return get_time_offblock();
        if (nr > wpts.size())
                return get_time_onblock();
        return wpts[nr-1].get_time();
}

FPlanLeg FPlanRoute::get_leg( uint32_t nr )
{
        bool future = nr > get_curwpt();
        const FPlanWaypoint *wpt0(0), *wpt1(0);
        if (nr > 0 && nr <= get_nrwpt())
                wpt0 = &(*this)[nr - 1];
        if (nr < get_nrwpt())
                wpt1 = &(*this)[nr];
        if (!wpt0 && !wpt1) {
                if (nr > 0 && get_nrwpt()) {
                        wpt0 = &(*this)[get_nrwpt()-1];
                        return FPlanLeg("", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                        "", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                        FPlanLeg::legtype_invalid, future);
                }
                return FPlanLeg("", _("Off Block"), Point(), 0, 0, 0, 0,
                                "", _("On Block"), Point(), 0, 0, 0, 0,
                                FPlanLeg::legtype_invalid, future);
        }
        if (!wpt0)
                return FPlanLeg("", _("Off Block"), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_offblock(), wpt1->get_frequency(),
                                wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), wpt1->get_time(), wpt1->get_frequency(),
                                FPlanLeg::legtype_offblock_takeoff, future);
        if (!wpt1)
                return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                "", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                FPlanLeg::legtype_landing_onblock, future);
        return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                        wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_onblock(), wpt1->get_frequency(),
                        FPlanLeg::legtype_normal, future);
}

FPlanLeg FPlanRoute::get_directleg(void )
{
        uint32_t nr(get_nrwpt());
        bool future = false;
        if (!nr)
                return FPlanLeg("", _("Off Block"), Point(), 0, 0, 0, 0,
                                "", _("On Block"), Point(), 0, 0, 0, 0,
                                FPlanLeg::legtype_invalid, future);
        const FPlanWaypoint *wpt0(&(*this)[0]), *wpt1(&(*this)[nr - 1]);
        return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                        wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_onblock(), wpt1->get_frequency(),
                        FPlanLeg::legtype_normal, future);
}

Glib::ustring FPlan::get_userdbpath(void)
{
        Glib::ustring dbname(Glib::build_filename(Glib::get_home_dir(), USERDBPATH));
#if defined(HAVE_GIOMM) && 0
        {
                Glib::RefPtr<Gio::File> dir(Gio::File::create_for_path(dbname));
                try {
#ifdef HAVE_GIOMM_MAKE_DIRECTORY_WITH_PARENTS
                        if (dir->make_directory_with_parents())
#else
                        if (dir->make_directory())
#endif
                                std::cerr << "vfrnav directory (" << dbname << ") created" << std::endl;
                        else
                                throw std::runtime_error("vfrnav directory creation failed");
                } catch (const Gio::Error& e) {
                        if (e.code() != Gio::Error::EXISTS)
                                throw;
                        std::cerr << "vfrnav directory (" << dbname << ") already exists" << std::endl;
                }
        }
#else
        if (g_mkdir_with_parents(dbname.c_str(), 0755)) {
                throw std::runtime_error("vfrnav directory (" + dbname +")creation failed");
        }
#endif
        return dbname;
}

FPlan::FPlan(void)
{
        Glib::ustring dbname(Glib::build_filename(get_userdbpath(), "fplan.db"));
        db.open(dbname);
        if (sqlite3_create_function(db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("FPlanDb: cannot add user function simpledist");
        init_database();
}

FPlan::FPlan(sqlite3 *dbh)
{
        db.take(dbh);
        if (sqlite3_create_function(db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("FPlanDb: cannot add user function simpledist");
        init_database();
}

FPlan::~FPlan() throw()
{
}

void FPlan::dbfunc_simpledist(sqlite3_context *ctxt, int, sqlite3_value **values)
{
        sqlite3_result_double(ctxt, Point(sqlite3_value_int(values[0]), sqlite3_value_int(values[1]))
                        .spheric_distance_km(Point(sqlite3_value_int(values[2]), sqlite3_value_int(values[3]))));
}

void FPlan::init_database(void)
{
        {
                sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS fplan (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "TIMEOFFBLOCK INTEGER,"
                                              "SAVETIMEOFFBLOCK INTEGER,"
                                              "TIMEONBLOCK INTEGER,"
                                              "SAVETIMEONBLOCK INTEGER,"
                                              "CURWPT INTEGER,"
                                              "NOTE TEXT);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS waypoints (PLANID INTEGER NOT NULL CONSTRAINT fk_planid REFERENCES fplan(ID) ON DELETE CASCADE, "
                                              "NR INTEGER NOT NULL,"
                                              "ICAO TEXT,"
                                              "NAME TEXT,"
                                              "LON INTEGER,"
                                              "LAT INTEGER,"
                                              "ALT INTEGER,"
                                              "FREQ INTEGER,"
                                              "TIME INTEGER,"
                                              "SAVETIME INTEGER,"
                                              "FLAGS INTEGER,"
                                              "NOTE TEXT,"
                                              "PRIMARY KEY(PLANID,NR));");
                cmd.executenonquery();
        }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptplanindex ON waypoints (PLANID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptplanidnr ON waypoints (PLANID, NR);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptlon ON waypoints (LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptlat ON waypoints (LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptname ON waypoints (NAME);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wpticao ON waypoints (ICAO);"); cmd.executenonquery(); }
}

void FPlan::delete_database(void)
{
        { sqlite3x::sqlite3_command cmd(db, "DROP TABLE IF EXISTS fplan;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP TABLE IF EXISTS waypoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptplanindex;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptplanidnr;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptlat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptname;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wpticao;"); cmd.executenonquery(); }
        init_database();
}

std::string FPlan::escape_string( const std::string & s )
{
        std::string r;
        r.reserve(s.size());
        for (std::string::const_iterator i1(s.begin()), ie(s.end()); i1 != ie; i1++) {
                std::string::value_type c(*i1);
                if (c == '_' || c == '%')
                        r.push_back('%');
                r.push_back(c);
        }
        return r;
}

std::vector< FPlanWaypoint > FPlan::find_by_name( const Glib::ustring & nm, unsigned int limit, unsigned int skip, comp_t comp )
{
        std::vector<FPlanWaypoint> wpts;
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE NAME LIKE ? ORDER BY NAME ASC LIMIT ? OFFSET ?;");
                Glib::ustring pattern(escape_string(nm));
                switch (comp) {
                        case comp_contains:
                                pattern = '%' + pattern;
                                /* fall through */

                        case comp_startswith:
                                pattern += '%';
                                break;

                        case comp_exact:
                        default:
                                break;
                }
                cmd.bind(1, pattern);
                cmd.bind(2, (int)limit);
                cmd.bind(3, (int)skip);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}

std::vector< FPlanWaypoint > FPlan::find_by_icao( const Glib::ustring & ic, unsigned int limit, unsigned int skip, comp_t comp )
{
        std::vector<FPlanWaypoint> wpts;
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE ICAO LIKE ? ORDER BY ICAO ASC LIMIT ? OFFSET ?;");
                Glib::ustring pattern(escape_string(ic));
                switch (comp) {
                        case comp_contains:
                                pattern = '%' + pattern;
                                /* fall through */

                        case comp_startswith:
                                pattern += '%';
                                break;

                        case comp_exact:
                        default:
                                break;
                }
                cmd.bind(1, pattern);
                cmd.bind(2, (int)limit);
                cmd.bind(3, (int)skip);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}

std::vector< FPlanWaypoint > FPlan::find_nearest( const Point & pt, unsigned int limit, unsigned int skip, const Rect & rect )
{
        std::vector<FPlanWaypoint> wpts;
        {
                std::string rectexpr("LAT >= ?6 AND LAT <= ?8 AND LON >= ?5 AND LON <= ?7");
                if (rect.get_northeast().get_lon() < rect.get_southwest().get_lon())
                        std::string rectexpr("LAT >= ?6 AND LAT <= ?8 AND (LON <= ?5 OR LON >= ?7)");
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE " + rectexpr + " ORDER BY simpledist(LON,LAT,?1,?2) ASC LIMIT ?3 OFFSET ?4;");
                cmd.bind(1, pt.get_lon());
                cmd.bind(2, pt.get_lat());
                cmd.bind(3, (int)limit);
                cmd.bind(4, (int)skip);
                cmd.bind(5, rect.get_southwest().get_lon());
                cmd.bind(6, rect.get_southwest().get_lat());
                cmd.bind(7, rect.get_northeast().get_lon());
                cmd.bind(8, rect.get_northeast().get_lat());
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}
