/***************************************************************************
 *   Copyright (C) 2007, 2008, 2009, 2011 by Thomas Sailer                 *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"

#include <sys/time.h>
#include <sstream>
#include <gdk/gdkkeysyms.h>
#include <set>

#include "maps.h"

MapRenderer::MapRenderer( Engine & eng, const ScreenCoord & scrsize, const Point & center, int alt )
        : m_center(center), m_altitude(alt), m_engine(eng), m_route(0), m_track(0), m_screensize(scrsize)
{
}

void MapRenderer::draw( Glib::RefPtr< Gdk::Drawable > & drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs )
{
        Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(drawable);
        gc->set_fill(Gdk::SOLID);
        Gdk::Color white;
        white.set_grey(0xffff);
        Glib::RefPtr<Gdk::Colormap> colmap(Gdk::Colormap::get_system());
        colmap->alloc_color(white);
        gc->set_background(white);
        gc->set_foreground(white);
        drawable->draw_rectangle(gc, true, xy.getx(), xy.gety(), wh.getx(), wh.gety());
}

void MapRenderer::pan_up( void )
{
        ScreenCoord sc(get_screensize());
        sc.setx(sc.getx() / 2);
        sc.sety(sc.gety() * 3 / 4);
        set_center(screen_coord(sc), m_altitude);
}

void MapRenderer::pan_down( void )
{
        ScreenCoord sc(get_screensize());
        sc.setx(sc.getx() / 2);
        sc.sety(sc.gety() / 4);
        set_center(screen_coord(sc), m_altitude);
}

void MapRenderer::pan_left( void )
{
        ScreenCoord sc(get_screensize());
        sc.setx(sc.getx() * 3 / 4);
        sc.sety(sc.gety() / 2);
        set_center(screen_coord(sc), m_altitude);
}

void MapRenderer::pan_right( void )
{
        ScreenCoord sc(get_screensize());
        sc.setx(sc.getx() / 4);
        sc.sety(sc.gety() / 2);
        set_center(screen_coord(sc), m_altitude);
}

VectorMapRenderer::TimeMeasurement::TimeMeasurement( bool start )
        : m_running(start)
{
        m_start.assign_current_time();
        m_stop = m_start;
}

void VectorMapRenderer::TimeMeasurement::start( void )
{
        if (m_running)
                return;
        m_start.assign_current_time();
        m_stop = m_start;
        m_running = true;
}

void VectorMapRenderer::TimeMeasurement::stop( void )
{
        if (!m_running)
                return;
        m_stop.assign_current_time();
        m_running = false;
}

VectorMapRenderer::TimeMeasurement::operator unsigned int( )
{
        stop();
        Glib::TimeVal v(m_stop);
        v -= m_start;
        return v.tv_sec * 1000 + v.tv_usec / 1000;
}

std::ostream & VectorMapRenderer::TimeMeasurement::print( std::ostream & os )
{
        os << (unsigned int)(*this) << "ms";
        return os;
}

VectorMapRenderer::VectorMapRenderer( Engine & eng, int depth, const ScreenCoord & scrsize, const Point & center, renderer_t renderer )
        : MapRenderer(eng, scrsize, center), m_thread(0), m_drawflags(drawflags_all)
{
        switch (renderer) {
                default:
                case renderer_vmap:
                        m_thread = new DrawThreadVMap(eng, depth, scrsize, center);
                        break;

                case renderer_terrain:
                        m_thread = new DrawThreadTerrain(eng, depth, scrsize, center);
                        break;

                case renderer_airportdiagram:
                        m_thread = new DrawThreadAirportDiagram(eng, depth, scrsize, center);
                        break;
        }
        set_map_scale(0.1);
        std::cerr << "VectorMapRenderer: create depth " << depth << std::endl;
}

VectorMapRenderer::~VectorMapRenderer()
{
        delete m_thread;
}

void VectorMapRenderer::set_center( const Point & pt, int alt )
{
        //std::cerr << "VectorMapRenderer::set_center: " << pt << std::endl;
        MapRenderer::set_center(pt, alt);
        if (m_thread)
                m_thread->set_center(pt, alt);
}

void VectorMapRenderer::set_screensize( const ScreenCoord & scrsize )
{
        MapRenderer::set_screensize(scrsize);
        if (m_thread)
                m_thread->set_screensize(scrsize);
}

MapRenderer& VectorMapRenderer::operator =( DrawFlags f )
{
        if (m_thread)
                m_thread->set_drawflags(m_drawflags = f);
        //std::cerr << "drawflags: " << m_drawflags << std::endl;
        return *this;
}

MapRenderer& VectorMapRenderer::operator &=( DrawFlags f )
{
        if (m_thread)
                m_thread->set_drawflags(m_drawflags &= f);
        return *this;
}

MapRenderer& VectorMapRenderer::operator |=( DrawFlags f )
{
        if (m_thread)
                m_thread->set_drawflags(m_drawflags |= f);
        return *this;
}

MapRenderer& VectorMapRenderer::operator ^=( DrawFlags f )
{
        if (m_thread)
                m_thread->set_drawflags(m_drawflags ^= f);
        return *this;
}

void VectorMapRenderer::airspaces_changed(void )
{
        if (m_thread)
                m_thread->airspaces_changed();
}

void VectorMapRenderer::airports_changed(void )
{
        if (m_thread)
                m_thread->airports_changed();
}

void VectorMapRenderer::navaids_changed(void )
{
        if (m_thread)
                m_thread->navaids_changed();
}

void VectorMapRenderer::waypoints_changed(void )
{
        if (m_thread)
                m_thread->waypoints_changed();
}

void VectorMapRenderer::airways_changed(void)
{
        if (m_thread)
                m_thread->airways_changed();
}

void VectorMapRenderer::draw( Glib::RefPtr< Gdk::Drawable > & drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs )
{
        if (m_thread)
                m_thread->draw(drawable, xy, wh, offs);
        {
                Cairo::RefPtr<Cairo::Context> cr = drawable->create_cairo_context();
                cr->rectangle(xy.getx(), xy.gety(), wh.getx(), wh.gety());
                cr->clip();
                // draw route
                if ((m_drawflags & drawflags_route) && m_thread && m_route)
                        m_thread->draw_route(cr.operator->(), m_route, xy, offs);
                // draw track
                if ((m_drawflags & drawflags_track) && m_thread && m_track)
                        m_thread->draw_track(cr.operator->(), m_track, xy, offs);
                cr->translate(10, m_screensize.gety() - 10);
                cr->set_font_size(12);
                // draw scale
                if (m_thread)
                        m_thread->draw_scale(cr.operator->());
        }
}

void VectorMapRenderer::redraw_route( void )
{
}

void VectorMapRenderer::redraw_track( void )
{
}

template<typename T> std::ostream & operator <<( std::ostream & os, const MapRenderer::ScreenCoordTemplate<T>& sc )
{
        return os << '(' << sc.getx() << ',' << sc.gety() << ')';
}

template std::ostream& operator<< <int>(std::ostream & os, const MapRenderer::ScreenCoordTemplate<int>& sc);
template std::ostream& operator<< <float>(std::ostream & os, const MapRenderer::ScreenCoordTemplate<float>& sc);





VectorMapArea::VectorMapArea( BaseObjectType * castitem, const Glib::RefPtr< Gnome::Glade::Xml > & refxml )
        : Gtk::DrawingArea(castitem), m_engine(0), m_renderer(0), m_rendertype(renderer_none),
          m_dragbutton(2), m_dragstart(0, 0), m_dragoffset(0, 0), m_draginprogress(false),
          m_cursorvalid(false), m_cursoranglevalid(false), m_cursorangle(0),
          m_altitude(0), m_mapscale(0.1), m_route(0), m_track(0), m_drawflags(MapRenderer::drawflags_none)
{
        set_map_scale(0.1);
        signal_expose_event().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_expose_event));
        signal_size_allocate().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_size_allocate));
        signal_show().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_show));
        signal_hide().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_hide));
        signal_button_press_event().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_button_press_event));
        signal_button_release_event().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_button_release_event));
        signal_motion_notify_event().connect(sigc::mem_fun(*this, &VectorMapArea::on_my_motion_notify_event));
}

VectorMapArea::~VectorMapArea( )
{
        delete m_renderer;
}

void VectorMapArea::set_engine( Engine & eng )
{
        m_engine = &eng;
        renderer_alloc();
}

void VectorMapArea::set_renderer(renderer_t render)
{
        m_rendertype = render;
        renderer_alloc();
}

VectorMapArea::renderer_t VectorMapArea::get_renderer(void ) const
{
        return m_rendertype;
}

void VectorMapArea::renderer_alloc( void )
{
        delete m_renderer;
        m_renderer = 0;
        if (!m_engine)
                return;
        Glib::RefPtr<Gdk::Window> window = Gtk::Widget::get_root_window();
        if (!window)
                return;
        switch (m_rendertype) {
                default:
                case renderer_none:
                        break;

                case renderer_vmap:
                        m_renderer = new VectorMapRenderer(*m_engine, window->get_depth(), VectorMapRenderer::ScreenCoord(0, 0), Point(), VectorMapRenderer::renderer_vmap);
                        break;

                case renderer_terrain:
                        m_renderer = new VectorMapRenderer(*m_engine, window->get_depth(), VectorMapRenderer::ScreenCoord(0, 0), Point(), VectorMapRenderer::renderer_terrain);
                        break;

                case renderer_airportdiagram:
                        m_renderer = new VectorMapRenderer(*m_engine, window->get_depth(), VectorMapRenderer::ScreenCoord(0, 0), Point(), VectorMapRenderer::renderer_airportdiagram);
                        break;
        }
        if (m_renderer) {
                m_renderer->connect_update(sigc::mem_fun(*this, &VectorMapArea::renderer_update));
                m_renderer->set_maxalt(m_engine->get_prefs().maxalt * 100);
                m_renderer->set_route(m_route);
                m_renderer->set_track(m_track);
                m_renderer->set_center(m_center, m_altitude);
                m_renderer->set_map_scale(m_mapscale);
                *m_renderer = m_drawflags;
                m_renderer->set_screensize(MapRenderer::ScreenCoord(get_width(), get_height()));
        }
}

void VectorMapArea::set_route( const FPlanRoute & route )
{
        m_route = &route;
        if (!m_renderer)
                return;
        m_renderer->set_route(m_route);
}

void VectorMapArea::set_track(const TracksDb::Track * track)
{
        m_track = track;
        if (!m_renderer)
                return;
        m_renderer->set_track(m_track);
}

void VectorMapArea::renderer_update( void )
{
        queue_draw();
}

void VectorMapArea::redraw_route( void )
{
        if (!m_renderer)
                return;
        m_renderer->redraw_route();
        queue_draw();
}

void VectorMapArea::redraw_track( void )
{
        if (!m_renderer)
                return;
        m_renderer->redraw_track();
        queue_draw();
}

void VectorMapArea::draw_pixmap( GdkEventExpose * event )
{
        if (!m_renderer)
                return;
        Glib::RefPtr<Gdk::Drawable> drawable = get_window();
        if (!drawable)
                return;
        m_renderer->draw(drawable, MapRenderer::ScreenCoord(event->area.x, event->area.y),
                         MapRenderer::ScreenCoord(event->area.width, event->area.height),
                         m_dragoffset);
}

bool VectorMapArea::on_my_expose_event( GdkEventExpose * event )
{
        if (!m_renderer)
                return false;
        draw_pixmap(event);
        if (m_cursorvalid) {
                Glib::RefPtr<Gdk::Drawable> drawable = get_window();
                if (!drawable)
                        return true;
                MapRenderer::ScreenCoord sc(m_renderer->coord_screen(m_cursor) - m_dragoffset);
                Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(drawable);
                gc->set_fill(Gdk::SOLID);
                Gdk::Color col1, col2;
                col1.set_rgb(0xffff, 0, 0);
                col2.set_rgb(0xffff, 0xffff, 0xffff);
                Glib::RefPtr<Gdk::Colormap> colmap(Gdk::Colormap::get_system());
                colmap->alloc_color(col1);
                colmap->alloc_color(col2);
                if (m_cursoranglevalid) {
                        gc->set_line_attributes(3, Gdk::LINE_SOLID, Gdk::CAP_NOT_LAST, Gdk::JOIN_MITER);
                        gc->set_foreground(col2);
                        //drawable->draw_arc(gc, true, sc.getx() - 15, sc.gety() - 15, 30, 30, (90 + 180 - 30) * 64 - m_cursorangle, 60 * 64);
                        drawable->draw_arc(gc, false, sc.getx() - 15, sc.gety() - 15, 30, 30, 0, 360 * 64);
                        gc->set_line_attributes(1, Gdk::LINE_SOLID, Gdk::CAP_NOT_LAST, Gdk::JOIN_MITER);
                        gc->set_foreground(col1);
                        drawable->draw_arc(gc, true, sc.getx() - 15, sc.gety() - 15, 30, 30, (90 + 180 - 30) * 64 - m_cursorangle, 60 * 64);
                        drawable->draw_arc(gc, false, sc.getx() - 15, sc.gety() - 15, 30, 30, 0, 360 * 64);
                } else if (false) {
                        gc->set_foreground(col2);
                        drawable->draw_rectangle(gc, true, sc.getx() - 16, sc.gety() - 2, 32, 5);
                        drawable->draw_rectangle(gc, true, sc.getx() - 2, sc.gety() - 16, 5, 32);
                        gc->set_foreground(col1);
                        drawable->draw_rectangle(gc, true, sc.getx() - 15, sc.gety() - 1, 30, 3);
                        drawable->draw_rectangle(gc, true, sc.getx() - 1, sc.gety() - 15, 3, 30);
                } else {
                        gc->set_foreground(col2);
                        drawable->draw_rectangle(gc, true, sc.getx() - 16, sc.gety() - 1, 32, 3);
                        drawable->draw_rectangle(gc, true, sc.getx() - 1, sc.gety() - 16, 3, 32);
                        gc->set_foreground(col1);
                        drawable->draw_line(gc, sc.getx() - 15, sc.gety(), sc.getx() + 15, sc.gety());
                        drawable->draw_line(gc, sc.getx(), sc.gety() - 15, sc.getx(), sc.gety() + 15);
                }
        }
        return true;
}

void VectorMapArea::on_my_size_allocate( Gtk::Allocation & allocation )
{
	Gtk::DrawingArea::on_size_allocate(allocation);
        if (m_renderer)
                m_renderer->set_screensize(MapRenderer::ScreenCoord(allocation.get_width(), allocation.get_height()));
        //std::cerr << "map size allocate" << std::endl;
}

void VectorMapArea::on_my_show( void )
{
        std::cerr << "vmaparea: show" << std::endl;
        if (m_renderer)
                m_renderer->show();
}

void VectorMapArea::on_my_hide( void )
{
        std::cerr << "vmaparea: hide" << std::endl;
        if (m_renderer)
                m_renderer->hide();
}

unsigned int VectorMapArea::button_translate(GdkEventButton * event)
{
        unsigned int button = event->button;
        if (button == 1) {
                guint state = event->state;
                if (state & GDK_SHIFT_MASK)
                        button = 2;
                else if (state & GDK_CONTROL_MASK)
                        button = 3;
        }
        return button;
}

bool VectorMapArea::on_my_button_press_event( GdkEventButton * event )
{
        MapRenderer::ScreenCoordFloat sc(event->x, event->y);
        unsigned int button = button_translate(event);
        if (button == m_dragbutton) {
                m_dragstart = sc;
                m_dragoffset = MapRenderer::ScreenCoord(0, 0);
                m_draginprogress = true;
        } else if (button == 1) {
                if (m_renderer) {
                        m_cursor = m_renderer->screen_coord(sc);
                        m_cursorvalid = false;
                        signal_cursor()(m_cursor, cursor_invalid);
                        redraw();
                }
        }
        std::cerr << "map button_press " << sc << ' ' << button << std::endl;
        return false;
}

bool VectorMapArea::on_my_button_release_event( GdkEventButton * event )
{
        MapRenderer::ScreenCoordFloat sc(event->x, event->y);
        unsigned int button = button_translate(event);
        if (m_draginprogress && (button == m_dragbutton)) {
                if (m_renderer) {
                        Point pt = m_renderer->screen_coord(m_dragstart) - m_renderer->screen_coord(sc);
                        m_renderer->set_center(m_renderer->get_center() + pt, m_renderer->get_altitude());
                }
                m_dragoffset = MapRenderer::ScreenCoord(0, 0);
                m_draginprogress = false;
                redraw();
        } else if (button == 1) {
                if (m_renderer) {
                        m_cursor = m_renderer->screen_coord(sc);
                        m_cursorvalid = true;
                        signal_cursor()(m_cursor, cursor_mouse);
                        redraw();
                }
        }
        std::cerr << "map button_release " << sc << ' ' << button << std::endl;
        return false;
}

bool VectorMapArea::on_my_motion_notify_event( GdkEventMotion * event )
{
        VectorMapRenderer::ScreenCoordFloat sc(event->x, event->y);
        unsigned int button = 0 /* event->button */;
        Point pt;
        if (m_renderer) {
                pt = m_renderer->screen_coord(sc);
                signal_pointer()(pt);
        }
        if (m_draginprogress) {
                if (m_renderer) {
                        MapRenderer::ScreenCoordFloat sc1 = m_dragstart - sc;
                        m_dragoffset = MapRenderer::ScreenCoord((int)sc1.getx(), (int)sc1.gety());
                        redraw();
                }
        }
        //std::cerr << "map motion: screen " << sc << " coord " << pt << std::endl;
        return false;
}

void VectorMapArea::pan_up( void )
{
        if (m_renderer) {
                m_renderer->pan_up();
                m_center = m_renderer->get_center();
        }
        redraw();
}

void VectorMapArea::pan_down( void )
{
        if (m_renderer) {
                m_renderer->pan_down();
                m_center = m_renderer->get_center();
        }
        redraw();
}

void VectorMapArea::pan_left( void )
{
        if (m_renderer) {
                m_renderer->pan_left();
                m_center = m_renderer->get_center();
        }
        redraw();
}

void VectorMapArea::pan_right( void )
{
        if (m_renderer) {
                m_renderer->pan_right();
                m_center = m_renderer->get_center();
        }
        redraw();
}

void VectorMapArea::set_map_scale( float nmi_per_pixel )
{
        m_mapscale = nmi_per_pixel;
        if (!m_renderer)
                return;
        m_renderer->set_map_scale(m_mapscale);
        m_mapscale = m_renderer->get_map_scale();
        redraw();
}

float VectorMapArea::get_map_scale( void ) const
{
        if (!m_renderer)
                return m_mapscale;
        return m_renderer->get_map_scale();
}

void VectorMapArea::set_center( const Point & pt, int alt )
{
        m_center = pt;
        m_altitude = alt;
        if (!m_renderer)
                return;
        if (m_engine)
                m_renderer->set_maxalt(m_engine->get_prefs().maxalt * 100);
        m_renderer->set_center(m_center, m_altitude);
        redraw();
        if (!is_visible())
                m_renderer->hide();
}

Point VectorMapArea::get_center( void ) const
{
        if (!m_renderer)
                return m_center;
        return m_renderer->get_center();
}

int VectorMapArea::get_altitude(void) const
{
        if (!m_renderer)
                return m_altitude;
        return m_renderer->get_altitude();
}

void VectorMapArea::set_dragbutton( unsigned int dragbutton )
{
        m_dragbutton = dragbutton;
        m_dragoffset = MapRenderer::ScreenCoord(0, 0);
        m_draginprogress = false;
}

VectorMapArea::operator MapRenderer::DrawFlags( ) const
{
        if (!m_renderer)
                return m_drawflags;
        return *m_renderer;
}

VectorMapArea & VectorMapArea::operator =( MapRenderer::DrawFlags f )
{
        m_drawflags = f;
        if (m_renderer)
                *m_renderer = f;
        return *this;
}

VectorMapArea & VectorMapArea::operator &=( MapRenderer::DrawFlags f )
{
        m_drawflags &= f;
        if (m_renderer)
                *m_renderer &= f;
        return *this;
}

VectorMapArea & VectorMapArea::operator |=( MapRenderer::DrawFlags f )
{
        m_drawflags |= f;
        if (m_renderer)
                *m_renderer |= f;
        return *this;
}

VectorMapArea & VectorMapArea::operator ^=( MapRenderer::DrawFlags f )
{
        m_drawflags ^= f;
        if (m_renderer)
                *m_renderer ^= f;
        return *this;
}

void VectorMapArea::airspaces_changed(void)
{
        if (m_renderer)
                m_renderer->airspaces_changed();
}

void VectorMapArea::airports_changed(void)
{
        if (m_renderer)
                m_renderer->airports_changed();
}

void VectorMapArea::navaids_changed(void)
{
        if (m_renderer)
                m_renderer->navaids_changed();
}

void VectorMapArea::waypoints_changed(void)
{
        if (m_renderer)
                m_renderer->waypoints_changed();
}

void VectorMapArea::airways_changed(void)
{
        if (m_renderer)
                m_renderer->airways_changed();
}

void VectorMapArea::invalidate_cursor( void )
{
        if (m_cursorvalid) {
                m_cursorvalid = false;
                redraw();
                signal_cursor()(m_cursor, cursor_invalid);
        }
}

void VectorMapArea::set_cursor( const Point & pt )
{
         m_cursor = pt;
         m_cursorvalid = true;
         redraw();
         signal_cursor()(m_cursor, cursor_set);
}

Rect VectorMapArea::get_cursor_rect( void ) const
{
        if (!m_cursorvalid || !m_renderer)
                return Rect();
        MapRenderer::ScreenCoord sc(m_renderer->coord_screen(m_cursor));
        return Rect(m_renderer->screen_coord(sc + MapRenderer::ScreenCoord(-5, 5)),
                    m_renderer->screen_coord(sc + MapRenderer::ScreenCoord(5, -5)));
}

void VectorMapArea::set_cursor_angle(float angle)
{
        m_cursorangle = Point::round<int,float>(angle * 64.0f);
        m_cursoranglevalid = true;
        redraw();
}

void VectorMapArea::invalidate_cursor_angle(void)
{
        m_cursorangle = 0;
        if (m_cursoranglevalid) {
                m_cursoranglevalid = false;
                redraw();
        }
}

void VectorMapArea::zoom_out( void )
{
        if (m_renderer) {
                m_renderer->zoom_out();
                m_mapscale = m_renderer->get_map_scale();
        }
        redraw();
}

void VectorMapArea::zoom_in( void )
{
        if (m_renderer) {
                m_renderer->zoom_in();
                m_mapscale = m_renderer->get_map_scale();
        }
        redraw();
}

VectorMapRenderer::DrawThread::DrawThread(Engine & eng, int depth, const ScreenCoord & scrsize, const Point & center, int alt)
        : m_engine(eng), m_center(center), m_altitude(alt), m_depth(depth), m_screensize(scrsize),
          m_mapscale(0.0), m_invmapscale(0.0), m_mapscalelon(0.0), m_invmapscalelon(0.0), m_drawflags(drawflags_all),
          m_surfacecenter(center), m_done(false), m_abort(true), m_hidden(true), m_terminate(false), m_mapscalechange(false), m_thread(0),
          m_current_surfacecenter(center), m_current_surfacesize(0, 0), m_maxalt(std::numeric_limits<int>::max())
{
        set_map_scale(0.1);
        m_thread = Glib::Thread::create(sigc::mem_fun(*this, &VectorMapRenderer::DrawThread::thread), 0, true, true, Glib::THREAD_PRIORITY_NORMAL);
}

VectorMapRenderer::DrawThread::~DrawThread()
{
        if (m_thread) {
                m_terminate = true;
                m_abort = true;
                {
                        Glib::Mutex::Lock lock(m_mutex);
                        m_cond.broadcast();
                }
                m_thread->join();
                m_thread = 0;
        }
}

void VectorMapRenderer::DrawThread::set_screensize(const ScreenCoord& scrsize)
{
        m_screensize = scrsize;
        restart();
}

void VectorMapRenderer::DrawThread::set_center(const Point & pt, int alt)
{
        m_center = pt;
        m_altitude = alt;
        update_mapscalelon();
        restart();
}

VectorMapRenderer::DrawFlags VectorMapRenderer::DrawThread::set_drawflags(DrawFlags df)
{
        bool need_restart(m_drawflags != df);
        m_drawflags = df;
        if (need_restart)
                restart();
        return m_drawflags;
}

void VectorMapRenderer::DrawThread::airspaces_changed(void)
{
        if (m_drawflags & drawflags_airspaces)
                restart();
}

void VectorMapRenderer::DrawThread::airports_changed(void)
{
        if (m_drawflags & drawflags_airports)
                restart();
}

void VectorMapRenderer::DrawThread::navaids_changed(void)
{
        if (m_drawflags & drawflags_navaids)
                restart();
}

void VectorMapRenderer::DrawThread::waypoints_changed(void)
{
        if (m_drawflags & drawflags_waypoints)
                restart();
}

void VectorMapRenderer::DrawThread::airways_changed(void)
{
        if (m_drawflags & (drawflags_airways_low | drawflags_airways_high))
                restart();
}

void VectorMapRenderer::DrawThread::set_map_scale(float nmi_per_pixel)
{
        //std::cerr << "VectorMapRenderer::DrawThread::set_map_scale: " << nmi_per_pixel << std::endl;
        m_invmapscale = std::max(0.1f/20, std::min(1000.0f/20, nmi_per_pixel)) * (Point::from_deg / 60.0);
        m_mapscale = 1.0 / m_invmapscale;
        update_mapscalelon();
        m_mapscalechange = true;
        restart();
}

void VectorMapRenderer::DrawThread::update_mapscalelon(void)
{
        /*
         * mercator lat formula: ln(tan(pi/4+lat/2))
         * mercator lat scale factor (1/2+1/2*tan(1/2*lat+1/4*pi)^2)*tan(1/2*lat+1/4*pi)^(-1)
         * lon scale factor approx: 1 - 1/2*lat^2 + 1/24*lat^4 - 1/720 * lat^6 ...
         */
        /*
         * alternative mercator lat formula: asinh(tan(lat))
         * mercator lat scale factor sqrt(1+tan(lat)^2)
         * lon scale factor approximation: 1 - 1/2*lat^2 + 1/24*lat^4 - 1/720*lat^6 + 1/40320*lat^8 - ...
         */
        float l = m_center.get_lat_rad();
        l *= l;
        float c = 1 - l * (1.0f / 2.0f);
        l *= l;
        c += l * (1.0f / 24.0f);
        m_mapscalelon = m_mapscale * c;
        m_invmapscalelon = 1.0f / m_mapscalelon;
        std::cerr << "m_mapscale " << m_mapscale << " m_invmapscale " << m_invmapscale
                  << " m_mapscalelon " << m_mapscalelon << " m_invmapscalelon " << m_invmapscalelon << std::endl;
}

Cairo::RefPtr<Cairo::ImageSurface> VectorMapRenderer::DrawThread::get_surface(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        return m_surface;
}

void VectorMapRenderer::DrawThread::hide(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_abort = true;
        m_hidden = true;
        m_cond.broadcast();
}

void VectorMapRenderer::DrawThread::restart(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_abort = true;
        m_cond.broadcast();
}

void VectorMapRenderer::DrawThread::draw(Glib::RefPtr< Gdk::Drawable > & drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs)
{
        Cairo::RefPtr<Cairo::ImageSurface> surface;
        {
                Glib::Mutex::Lock lock(m_mutex);
                bool abrt(m_hidden);
                m_hidden = false;
                surface = m_surface;
                if (!surface)
                        abrt = true;
                if (abrt) {
                        m_abort = true;
                        m_cond.broadcast();
                }
        }
        if (surface) {
                Cairo::RefPtr<Cairo::Context> cr = drawable->create_cairo_context();
                Cairo::RefPtr<Cairo::SurfacePattern> pattern = Cairo::SurfacePattern::create(surface);
                pattern->set_extend(Cairo::EXTEND_NONE);
                pattern->set_filter(Cairo::FILTER_GOOD);
                ScreenCoord sc(get_offset() + offs);
                cr->rectangle(xy.getx(), xy.gety(), wh.getx(), wh.gety());
                cr->clip();
                cr->translate(-sc.getx(), -sc.gety());
                //cr->reset_clip();
                cr->set_source(pattern);
                cr->paint();
#if 0
                std::cerr << "draw: x " << xy.getx() << " y " << xy.gety() << " width " << wh.getx() << " height " << wh.gety()
                          << " surface w " << surface->get_width() << " h " << surface->get_height() << " screencoord " << sc
                          << " center " << m_center << " pixmap center " << m_surfacecenter << std::endl;
#endif
                return;
        }
        {
                Glib::RefPtr<Gdk::GC> gc = Gdk::GC::create(drawable);
                gc->set_fill(Gdk::SOLID);
                Gdk::Color white;
                white.set_grey(0xffff);
                Glib::RefPtr<Gdk::Colormap> colmap(Gdk::Colormap::get_system());
                colmap->alloc_color(white);
                gc->set_background(white);
                gc->set_foreground(white);
                drawable->draw_rectangle(gc, true, xy.getx(), xy.gety(), wh.getx(), wh.gety());
                return;
        }
}

void VectorMapRenderer::DrawThread::thread( void )
{
        while (!m_terminate) {
                //std::cerr << "draw thread: waiting" << std::endl;
                {
                        Glib::Mutex::Lock lock(m_mutex);
                        while (!m_terminate && !m_abort)
                                m_cond.wait(m_mutex);
                }
                //std::cerr << "draw thread: action: term " << m_terminate << " abort " << m_abort << " hidden " << m_hidden << std::endl;
                if (m_terminate)
                        break;
                if (m_abort) {
                        if (m_hidden) {
                                {
                                        Glib::Mutex::Lock lock(m_mutex);
                                        m_surface = Cairo::RefPtr<Cairo::ImageSurface>();
                                        m_abort = false;
                                }
                        } else {
                                do {
                                        compute_pixmap();
                                } while (!m_done && !(m_terminate || m_hidden));
                        }
                }
        }
        //std::cerr << "draw thread: terminating" << std::endl;
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::current_ctrans(const Point & p) const
{
        Point pp(p - m_current_surfacecenter);
        ScreenCoord ps(Point::round<int,float>(pp.get_lon() * m_mapscalelon), Point::round<int,float>(pp.get_lat() * m_mapscale));
        ScreenCoord cc(get_current_center_coord());
        return ScreenCoord(cc.getx() + ps.getx(), cc.gety() - ps.gety());
}

VectorMapRenderer::ScreenCoordFloat VectorMapRenderer::DrawThread::current_cftrans(const Point & p) const
{
        Point pp(p - m_current_surfacecenter);
        ScreenCoordFloat ps(pp.get_lon() * m_mapscalelon, pp.get_lat() * m_mapscale);
        ScreenCoord cc(get_current_center_coord());
        return ScreenCoordFloat(cc.getx() + ps.getx(), cc.gety() - ps.gety());
}

Point VectorMapRenderer::DrawThread::current_ctrans(const ScreenCoord & p, const Point & surfacecenter) const
{
        ScreenCoord cc(get_current_center_coord());
        ScreenCoord pp(p.getx() - cc.getx(), cc.gety() - p.gety());
        return Point(Point::round<Point::coord_t,float>(pp.getx() * m_invmapscalelon), Point::round<Point::coord_t,float>(pp.gety() * m_invmapscale)) + surfacecenter;
}

Point VectorMapRenderer::DrawThread::current_cftrans(const ScreenCoordFloat & p, const Point & surfacecenter) const
{
        ScreenCoord cc(get_current_center_coord());
        ScreenCoordFloat pp(p.getx() - cc.getx(), cc.gety() - p.gety());
        return Point(Point::round<Point::coord_t,float>(pp.getx() * m_invmapscalelon), Point::round<Point::coord_t,float>(pp.gety() * m_invmapscale)) + surfacecenter;
}

Point VectorMapRenderer::DrawThread::current_ctrans(const ScreenCoord & p) const
{
        return current_ctrans(p, m_current_surfacecenter);
}

Point VectorMapRenderer::DrawThread::current_cftrans(const ScreenCoordFloat & p) const
{
        return current_cftrans(p, m_current_surfacecenter);
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::get_current_center_coord(void) const
{
        return get_current_pixmap_size() / 2;
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::get_pixmap_size(void) const
{
        if (m_surface)
                return ScreenCoord(m_surface->get_width(), m_surface->get_height());
        return ScreenCoord(0, 0);
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::get_center_coord(void) const
{
        return get_pixmap_size() / 2;
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::ctrans(const Point & p) const
{
        Point pp(p - m_surfacecenter);
        ScreenCoord ps(Point::round<int,float>(pp.get_lon() * m_mapscalelon), Point::round<int,float>(pp.get_lat() * m_mapscale));
        ScreenCoord cc(get_center_coord());
        return ScreenCoord(cc.getx() + ps.getx(), cc.gety() - ps.gety());
}

VectorMapRenderer::ScreenCoordFloat VectorMapRenderer::DrawThread::cftrans(const Point & p) const
{
        Point pp(p - m_surfacecenter);
        ScreenCoordFloat ps(pp.get_lon() * m_mapscalelon, pp.get_lat() * m_mapscale);
        ScreenCoord cc(get_center_coord());
        return ScreenCoordFloat(cc.getx() + ps.getx(), cc.gety() - ps.gety());
}

Point VectorMapRenderer::DrawThread::ctrans(const ScreenCoord & p) const
{
        ScreenCoord cc(get_center_coord());
        ScreenCoord pp(p.getx() - cc.getx(), cc.gety() - p.gety());
        return Point(Point::round<int32_t,float>(pp.getx() * m_invmapscalelon), Point::round<int32_t,float>(pp.gety() * m_invmapscale)) + m_surfacecenter;
}

Point VectorMapRenderer::DrawThread::cftrans(const ScreenCoordFloat & p) const
{
        ScreenCoord cc(get_center_coord());
        ScreenCoordFloat pp(p.getx() - cc.getx(), cc.gety() - p.gety());
        return Point(Point::round<int32_t,float>(pp.getx() * m_invmapscalelon), Point::round<int32_t,float>(pp.gety() * m_invmapscale)) + m_surfacecenter;
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::get_offset(void) const
{
        Point pp(m_center - m_surfacecenter);
        ScreenCoord sc(Point::round<int,float>(pp.get_lon() * m_mapscalelon), Point::round<int,float>(pp.get_lat() * -m_mapscale));
        sc += get_center_coord();
        sc -= m_screensize / 2;
        return sc;
}

VectorMapRenderer::ScreenCoordFloat VectorMapRenderer::DrawThread::get_offset_float(void) const
{
        Point pp(m_center - m_surfacecenter);
        ScreenCoord sc1(get_center_coord() - m_screensize / 2);
        ScreenCoordFloat sc(pp.get_lon() * m_mapscalelon + sc1.getx(), pp.get_lat() * -m_mapscale + sc1.gety());
        return sc;
}

float VectorMapRenderer::DrawThread::nmi_to_pixel(float nmi) const
{
        return nmi * (1.0f / 60.0f * Point::from_deg) * m_mapscale;
}

Point VectorMapRenderer::DrawThread::screen_coord(const ScreenCoord & sc)
{
        Glib::Mutex::Lock lock(m_mutex);
        ScreenCoord c(sc + get_offset());
        return ctrans(c);
}

Point VectorMapRenderer::DrawThread::screen_coord(const ScreenCoordFloat & sc)
{
        Glib::Mutex::Lock lock(m_mutex);
        ScreenCoordFloat c(sc + get_offset_float());
        return cftrans(c);
}

VectorMapRenderer::ScreenCoord VectorMapRenderer::DrawThread::coord_screen(const Point & pt)
{
        Glib::Mutex::Lock lock(m_mutex);
        return ctrans(pt) - get_offset();
}

Glib::ustring VectorMapRenderer::DrawThread::freqstr( uint32_t f )
{
        if (f < 1000)
                return Glib::ustring();
        char buf[16];
        f /= 1000;
        if (f < 1000) {
                snprintf(buf, sizeof(buf), "%u", f);
                return buf;
        }
        snprintf(buf, sizeof(buf), "%.3f", f * (1.0 / 1000));
        return buf;
}

void VectorMapRenderer::DrawThread::polypath(Cairo::Context * cr, const Polygon& poly, bool closeit, const ScreenCoordFloat& offset)
{
        if (poly.empty())
                return;
        std::vector<Point>::const_iterator i1(poly.begin()), ie(poly.end());
        for (bool first = true; i1 != ie; i1++, first = false) {
                ScreenCoordFloat px(current_cftrans(*i1) + offset);
                //std::cerr << "pt: " << (*i1) << " -> " << px << std::endl;
                if (first)
                        cr->move_to(px.getx(), px.gety());
                else
                        cr->line_to(px.getx(), px.gety());
        }
        if (closeit)
                cr->close_path();
}

void VectorMapRenderer::DrawThread::polypath(Cairo::Context * cr, const PolygonHole & poly, bool closeit, const ScreenCoordFloat & offset)
{
        polypath(cr, poly.get_exterior(), closeit, offset);
        for (unsigned int i = 0; i < poly.get_nrinterior(); i++)
                polypath(cr, poly[i], closeit, offset);
}

void VectorMapRenderer::DrawThread::draw_gtopo(Cairo::Context * cr, const Rect & cbbox, const Cairo::RefPtr< Cairo::ImageSurface > & pbuf)
{
        ScreenCoord c1(current_ctrans(cbbox.get_southwest())), c2(current_ctrans(cbbox.get_northeast()));
        int pbufx = pbuf->get_width(), pbufy = pbuf->get_height();
        if (pbufx <= 0 || pbufy <= 0)
                return;
        cr->save();
        cr->translate(c1.getx(), c2.gety());
        cr->scale((c2.getx() - c1.getx()) / (double)pbufx, (c1.gety() - c2.gety()) / (double)pbufy);
        //std::cerr << "scaling " << (c2.getx() - c1.getx()) / (double)pbufx << " / " << (c1.gety() - c2.gety()) / (double)pbufy
        //          << " translate (" << c1.getx() << " / " << c2.gety() << ") size " << pbufx << " / " << pbufy << std::endl;
        //std::cerr << "cbbox " << cbbox << " cursfccenter " << m_current_surfacecenter << std::endl;
        Cairo::RefPtr<Cairo::SurfacePattern> pat(Cairo::SurfacePattern::create(pbuf));
        cr->set_source(pat);
        cr->paint();
        cr->restore();
}
void VectorMapRenderer::DrawThread::draw_scale(Cairo::Context * cr, float target_pixel)
{
        float nmi = target_pixel * m_invmapscale * (Point::to_deg * 60.0);
        nmi = powf(10, floorf(log10f(nmi)));
        float pix1 = nmi * m_mapscale * (Point::from_deg * (1.0 / 60.0));
        if (pix1 < target_pixel * (1.0 / 2.0)) {
                pix1 *= 5;
                nmi *= 5;
        } else if (pix1 < target_pixel) {
                pix1 *= 2;
                nmi *= 2;
        }
        int pix = (int)pix1;
        //std::cerr << "draw_scale: nmi " << nmi << " pix " << pix << std::endl;
        std::ostringstream oss;
        oss << nmi;
        cr->save();
        cr->set_line_width(2.0);
        cr->set_source_rgb(0, 0, 0);
        cr->move_to(0, -5);
        cr->rel_line_to(0, 5);
        cr->rel_line_to(pix, 0);
        cr->rel_line_to(0, -5);
        cr->stroke();
        Cairo::TextExtents ext;
        cr->get_text_extents(oss.str(), ext);
#if 0
        cr->move_to((pix - ext.width) / 2, -3);
        cr->show_text(oss.str());
#else
        cr->set_source_rgb(1.0, 1.0, 1.0);
        cr->move_to((pix - ext.width) / 2, -2);
        cr->show_text(oss.str());
        cr->move_to((pix - ext.width) / 2, -4);
        cr->show_text(oss.str());
        cr->move_to((pix - ext.width) / 2 - 1, -3);
        cr->show_text(oss.str());
        cr->move_to((pix - ext.width) / 2 + 1, -3);
        cr->show_text(oss.str());
        cr->move_to((pix - ext.width) / 2, -3);
        cr->set_source_rgb(0, 0, 0);
        cr->show_text(oss.str());
#endif
        cr->stroke();
        cr->restore();
}

void VectorMapRenderer::DrawThread::draw_route(Cairo::Context * cr, const FPlanRoute * route, ScreenCoord xy, ScreenCoord offs)
{
        if (!route)
                return;
        if (route->get_nrwpt() < 2)
                return;
        cr->save();
        cr->set_line_width(2.0);
        cr->set_source_rgb(0xff / 255.0, 0x00 / 255.0, 0xc8 / 255.0);
        ScreenCoord sc1(offs + xy);
        ScreenCoordFloat sc(get_offset_float() + ScreenCoordFloat(sc1.getx(), sc1.gety()));
        ScreenCoordFloat px(cftrans((*route)[0].get_coord()) - sc);
        cr->move_to(px.getx(), px.gety());
        for (unsigned int i = 1; i < route->get_nrwpt(); i++) {
                px = cftrans((*route)[i].get_coord()) - sc;
                cr->line_to(px.getx(), px.gety());
        }
        cr->stroke();
        cr->restore();
}

void VectorMapRenderer::DrawThread::draw_track(Cairo::Context * cr, const TracksDb::Track * track, ScreenCoord xy, ScreenCoord offs)
{
        if (!track)
                return;
        ScreenCoord sc1(offs + xy);
        ScreenCoordFloat sc(get_offset_float() + ScreenCoordFloat(sc1.getx(), sc1.gety()));
        if (m_drawflags & drawflags_track_points) {
                cr->save();
                Cairo::RefPtr<Cairo::ImageSurface> pats(Cairo::ImageSurface::create(Cairo::FORMAT_ARGB32, 4, 4));
                int path(pats->get_height()), patw(pats->get_width());
                {
                        Cairo::RefPtr<Cairo::Context> crp(Cairo::Context::create(pats));
                        crp->set_source_rgba(1.0, 1.0, 1.0, 0.0);
                        crp->paint();
                        crp->set_source_rgba(255.0/255.0, 128.0/255.0, 0.0/255.0, 1.0);
                        crp->arc(patw / 2, path / 2, 2.0, 0.0, 2.0 * M_PI);
                        crp->fill();
                        //crp->paint();
                }
                Cairo::RefPtr<Cairo::SurfacePattern> pat(Cairo::SurfacePattern::create(pats));
                pat->set_extend(Cairo::EXTEND_NONE);
                pat->set_filter(Cairo::FILTER_FAST);
                cr->set_source(pat);
                Cairo::Matrix matrix;
                pat->get_matrix(matrix);
                cairo_matrix_translate(&matrix, patw/2, path/2);
                for (unsigned int i = 0; i < track->size(); i++) {
                        ScreenCoordFloat px(cftrans((*track)[i].get_coord()) - sc);
                        Cairo::Matrix matrix2(matrix);
                        cairo_matrix_translate(&matrix2, -px.getx(), -px.gety());
                        pat->set_matrix(matrix2);
                        //std::cerr << "Translate: " << px.getx() << ", " << px.gety() << std::endl;
                        cr->paint();
                }
                cr->restore();
                return;
        }
        if (track->size() < 2)
                return;
        cr->save();
        cr->set_line_width(2.0);
        cr->set_source_rgb(255.0/255.0, 128.0/255.0, 0.0/255.0);
        ScreenCoordFloat px(cftrans((*track)[0].get_coord()) - sc);
        cr->move_to(px.getx(), px.gety());
        for (unsigned int i = 1; i < track->size(); i++) {
                px = cftrans((*track)[i].get_coord()) - sc;
                cr->line_to(px.getx(), px.gety());
        }
        cr->stroke();
        cr->restore();
}

void VectorMapRenderer::DrawThread::draw_label(Cairo::Context * cr, const Glib::ustring & text, const ScreenCoord & pos, NavaidsDb::Navaid::label_placement_t lp, int dist)
{
        if (text.empty() || lp == NavaidsDb::Navaid::label_off)
                return;
        ScreenCoord px(pos);
        cr->save();
        Cairo::TextExtents ext;
        cr->get_text_extents(text, ext);
        ScreenCoord tdim((int)ext.width, (int)ext.height);
        switch (lp) {
                default:
                        cr->restore();
                        return;

                case NavaidsDb::Navaid::label_n:
                        px -= ScreenCoord(tdim.getx() / 2, dist);
                        break;

                case NavaidsDb::Navaid::label_e:
                        px += ScreenCoord(dist, tdim.gety() / 2);
                        break;

                case NavaidsDb::Navaid::label_s:
                        px += ScreenCoord(-tdim.getx() / 2, dist + tdim.gety());
                        break;

                case NavaidsDb::Navaid::label_w:
                        px += ScreenCoord(-tdim.getx() - dist, tdim.gety() / 2);
                        break;

                case NavaidsDb::Navaid::label_center:
                case NavaidsDb::Navaid::label_any:
                        px += ScreenCoord(-tdim.getx() / 2, tdim.gety() / 2);
                        break;
        }
        cr->move_to(px.getx(), px.gety());
        cr->show_text(text);
        //std::cerr << "drawing text " << text << " at " << px << std::endl;
        cr->restore();
}

void VectorMapRenderer::DrawThread::draw_label( Cairo::Context * cr, const Glib::ustring & text, const Point & pt, NavaidsDb::Navaid::label_placement_t lp, int dist )
{
        if (text.empty() || lp == NavaidsDb::Navaid::label_off)
                return;
        draw_label(cr, text, current_ctrans(pt), lp, dist);
}

void VectorMapRenderer::DrawThread::draw_waypoint( Cairo::Context * cr, const Point & pt, const Glib::ustring& str, bool mandatory, NavaidsDb::Navaid::label_placement_t lp )
{
        cr->save();
        cr->set_source_rgb(0.0, 0.0, 0.0);
        cr->set_line_width(2.0);
        cr->set_fill_rule(Cairo::FILL_RULE_EVEN_ODD);
        ScreenCoordFloat px(current_cftrans(pt));
        cr->move_to(px.getx() + 0.000 * 7, px.gety() - 1.000 * 7);
        cr->line_to(px.getx() - 0.866 * 7, px.gety() + 0.500 * 7);
        cr->line_to(px.getx() + 0.866 * 7, px.gety() + 0.500 * 7);
        cr->close_path();
        cr->arc(px.getx(), px.gety(), 2, 0, 2.0*M_PI);
        cr->close_path();
        if (mandatory)
                cr->fill();
        else
                cr->stroke();
        if (!str.empty())
                draw_label(cr, str, pt, lp, 8);
        cr->restore();
}

VectorMapRenderer::DrawThread::AirportRoutePoint::AirportRoutePoint(const Glib::ustring & name, const Point & coord, AirportsDb::Airport::label_placement_t lp)
        : m_name(name), m_coord(coord), m_labelplacement(lp)
{
}

VectorMapRenderer::DrawThread::AirportRoutePoint::AirportRoutePoint(const AirportsDb::Airport::VFRRoute::VFRRoutePoint & pt)
        : m_name(pt.get_name()), m_coord(pt.get_coord()), m_labelplacement(pt.get_label_placement())
{
}

bool VectorMapRenderer::DrawThread::AirportRoutePoint::operator<(const AirportRoutePoint & p) const
{
        if (m_coord.get_lon() < p.m_coord.get_lon())
                return true;
        if (m_coord.get_lon() > p.m_coord.get_lon())
                return false;
        if (m_coord.get_lat() < p.m_coord.get_lat())
                return true;
        if (m_coord.get_lat() > p.m_coord.get_lat())
                return false;
        return m_name < p.m_name;
}

void VectorMapRenderer::DrawThread::draw_airport_vfrreppt(Cairo::Context * cr, const AirportsDb::Airport & arpt)
{
        typedef std::map<AirportRoutePoint,bool> points_t;
        points_t pts;
        for (unsigned int i = 0; i < arpt.get_nrvfrrte(); i++) {
                const AirportsDb::Airport::VFRRoute& vfrrte(arpt.get_vfrrte(i));
                for (unsigned int j = 0; j < vfrrte.size(); j++) {
                        const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrrtept(vfrrte[j]);
                        if (vfrrtept.is_at_airport())
                                continue;
                        Glib::ustring name(vfrrtept.get_name());
                        if (name.empty())
                                continue;
                        std::pair<points_t::iterator,bool> ins(pts.insert(points_t::value_type(AirportRoutePoint(vfrrtept), vfrrtept.is_compulsory())));
                        if (vfrrtept.is_compulsory() && ins.first != pts.end())
                                ins.first->second = true;
                }
        }
        for (points_t::const_iterator pi(pts.begin()), pe(pts.end()); pi != pe; pi++)
                draw_waypoint(cr, pi->first.get_coord(), pi->first.get_name(), pi->second, pi->first.get_label_placement());
}

void VectorMapRenderer::DrawThread::draw_navaid(Cairo::Context * cr, const NavaidsDb::Navaid & n, const ScreenCoordFloat & pos)
{
        int dist = 2;
        NavaidsDb::Navaid::navaid_type_t t(n.get_navaid_type());
        switch (t) {
                default:
                        return;

                case NavaidsDb::Navaid::navaid_ndb:
                {
                        cr->arc(pos.getx(), pos.gety(), 1, 0, 2.0*M_PI);
                        cr->fill();
                        cr->arc(pos.getx(), pos.gety(), 5, 0, 2.0*M_PI);
                        cr->stroke();
                        std::valarray<double> dash(1);
                        dash[0] = 1;
                        cr->set_dash(dash, 0);
                        for (double r = 7; r < 12; r += 2) {
                                cr->arc(pos.getx(), pos.gety(), r, 0, 2.0*M_PI);
                                cr->stroke();
                        }
                        std::valarray<double> nodash(0);
                        cr->set_dash(nodash, 0);
                        dist = 11;
                        break;
                }

                case NavaidsDb::Navaid::navaid_vor:
                case NavaidsDb::Navaid::navaid_vordme:
                case NavaidsDb::Navaid::navaid_dme:
                {
                        cr->arc(pos.getx(), pos.gety(), 2, 0, 2.0*M_PI);
                        cr->fill();
                        cr->move_to(pos.getx() + 1.000 * 6, pos.gety() + 0.000 * 6);
                        cr->line_to(pos.getx() + 0.500 * 6, pos.gety() + 0.866 * 6);
                        cr->line_to(pos.getx() - 0.500 * 6, pos.gety() + 0.866 * 6);
                        cr->line_to(pos.getx() - 1.000 * 6, pos.gety() + 0.000 * 6);
                        cr->line_to(pos.getx() - 0.500 * 6, pos.gety() - 0.866 * 6);
                        cr->line_to(pos.getx() + 0.500 * 6, pos.gety() - 0.866 * 6);
                        cr->close_path();
                        cr->stroke();
                        if (t == NavaidsDb::Navaid::navaid_vordme || t == NavaidsDb::Navaid::navaid_dme) {
                                cr->move_to(pos.getx() + 1.000 * 6, pos.gety() + 0.866 * 6);
                                cr->line_to(pos.getx() + 1.000 * 6, pos.gety() - 0.866 * 6);
                                cr->line_to(pos.getx() - 1.000 * 6, pos.gety() - 0.866 * 6);
                                cr->line_to(pos.getx() - 1.000 * 6, pos.gety() + 0.866 * 6);
                                cr->close_path();
                                cr->stroke();
                        }
                        if (t == NavaidsDb::Navaid::navaid_vor || t == NavaidsDb::Navaid::navaid_vordme) {
                                cr->arc(pos.getx(), pos.gety(), 20, 0, 2.0*M_PI);
                                cr->stroke();
                                cr->move_to(pos.getx() + 10, pos.gety());
                                cr->line_to(pos.getx() + 20, pos.gety());
                                cr->stroke();
                                cr->move_to(pos.getx() + 15, pos.gety() + 2);
                                cr->line_to(pos.getx() + 20, pos.gety());
                                cr->line_to(pos.getx() + 15, pos.gety() - 2);
                                cr->stroke();
                                cr->move_to(pos.getx() - 10, pos.gety());
                                cr->line_to(pos.getx() - 20, pos.gety());
                                cr->stroke();
                                cr->move_to(pos.getx() - 15, pos.gety() + 2);
                                cr->line_to(pos.getx() - 20, pos.gety());
                                cr->line_to(pos.getx() - 15, pos.gety() - 2);
                                cr->stroke();
                                cr->move_to(pos.getx(), pos.gety() + 10);
                                cr->line_to(pos.getx(), pos.gety() + 20);
                                cr->stroke();
                                cr->move_to(pos.getx() + 2, pos.gety() + 15);
                                cr->line_to(pos.getx(), pos.gety() + 20);
                                cr->line_to(pos.getx() - 2, pos.gety() + 15);
                                cr->stroke();
                                cr->move_to(pos.getx(), pos.gety() - 10);
                                cr->line_to(pos.getx(), pos.gety() - 20);
                                cr->stroke();
                                cr->move_to(pos.getx() + 2, pos.gety() - 15);
                                cr->line_to(pos.getx(), pos.gety() - 20);
                                cr->line_to(pos.getx() - 2, pos.gety() - 15);
                                cr->stroke();
                        }
                        dist = 20;
                        break;
                }
        }
        draw_label(cr, n.get_icao() + ' ' + freqstr(n.get_frequency()), ScreenCoord((int)pos.getx(), (int)pos.gety()), n.get_label_placement(), dist);
}

VectorMapRenderer::DrawThreadVMap::DrawThreadVMap(Engine & eng, int depth, const ScreenCoord & scrsize, const Point & center, int alt)
        : DrawThread(eng, depth, scrsize, center, alt),
          m_current_drawflags(drawflags_all), m_current_surfacerect(center, center), m_current_dbrect(center, center), m_abort_cache(false)
{
}

VectorMapRenderer::DrawThreadVMap::~DrawThreadVMap()
{
        async_cancel();
        Glib::Mutex::Lock lock(m_mutex);
}

void VectorMapRenderer::DrawThreadVMap::airspaces_changed(void)
{
        Engine::AirspaceResult::cancel(m_airspacequery);
        if (m_drawflags & drawflags_airspaces)
                m_mapscalechange = true;
        DrawThread::airspaces_changed();
}

void VectorMapRenderer::DrawThreadVMap::airports_changed(void)
{
        Engine::AirportResult::cancel(m_airportquery);
        if (m_drawflags & drawflags_airports)
                m_mapscalechange = true;
        DrawThread::airports_changed();
}

void VectorMapRenderer::DrawThreadVMap::navaids_changed(void)
{
        Engine::NavaidResult::cancel(m_navaidquery);
        if (m_drawflags & drawflags_navaids)
                m_mapscalechange = true;
        DrawThread::navaids_changed();
}

void VectorMapRenderer::DrawThreadVMap::waypoints_changed(void)
{
        Engine::WaypointResult::cancel(m_waypointquery);
        if (m_drawflags & drawflags_waypoints)
                m_mapscalechange = true;
        DrawThread::waypoints_changed();
}

void VectorMapRenderer::DrawThreadVMap::airways_changed(void)
{
        Engine::AirwayResult::cancel(m_airwayquery);
        if (m_drawflags & (drawflags_airways_low | drawflags_airways_high))
                m_mapscalechange = true;
        DrawThread::airways_changed();
}

void VectorMapRenderer::DrawThreadVMap::async_cancel(void)
{
        Engine::ElevationMapCairoResult::cancel(m_topocairoquery);
        Engine::MapelementResult::cancel(m_mapelementquery);
        Engine::NavaidResult::cancel(m_navaidquery);
        Engine::WaypointResult::cancel(m_waypointquery);
        Engine::AirportResult::cancel(m_airportquery);
        Engine::AirspaceResult::cancel(m_airspacequery);
}

void VectorMapRenderer::DrawThreadVMap::async_done(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_cond.broadcast();
}

bool VectorMapRenderer::DrawThreadVMap::check_abort(void)
{
        if (m_hidden || m_terminate || m_abort_cache)
                return true;
        if (!m_abort)
                return false;
        if (m_mapscalechange) {
                m_abort_cache = true;
                return true;
        }
        if ((m_current_drawflags ^ m_drawflags) & (drawflags_all & ~(drawflags_route | drawflags_track))) {
                m_abort_cache = true;
                return true;
        }
        Rect scrbbox(Point(), Point(Point::round<Point::coord_t,float>(m_screensize.getx() * m_invmapscalelon),
                                    Point::round<Point::coord_t,float>(m_screensize.gety() * m_invmapscale)));
        scrbbox += (m_center - scrbbox.boxcenter());
        if (!m_current_surfacerect.is_inside(scrbbox)) {
                m_abort_cache = true;
                return true;
        }
        std::cerr << "cancelling abort: new screen rect " << scrbbox << " surface rect " << m_current_surfacerect << std::endl;
        m_abort = false;
        return false;
}

void VectorMapRenderer::DrawThreadVMap::compute_pixmap( void )
{
        DrawFlags current_drawflags;
        Point surfacecenter;
        ScreenCoord screensize(0, 0);
        bool need_redraw;
        {
                Glib::Mutex::Lock lock(m_mutex);
                current_drawflags = m_drawflags;
                surfacecenter = m_center;
                screensize = m_screensize;
                need_redraw = check_abort() || !m_surface;
        }
        if (!need_redraw)
                return;
        m_done = false;
        Rect screenbbox(Point(), Point(Point::round<Point::coord_t,float>(screensize.getx() * m_invmapscalelon),
                                       Point::round<Point::coord_t,float>(screensize.gety() * m_invmapscale)));
        screenbbox += (surfacecenter - screenbbox.boxcenter());
        screensize = ScreenCoord(screensize.getx() * 2, screensize.gety() * 2);
        bool need_dbquery;
        {
                Glib::Mutex::Lock lock(m_mutex);
                need_dbquery = (!m_current_dbrect.is_inside(screenbbox))
                                || ((current_drawflags & drawflags_topo) && (!m_topocairoquery || m_topocairoquery->is_error()))
                                || ((current_drawflags & drawflags_terrain) && (!m_mapelementquery || m_mapelementquery->is_error()))
                                || ((current_drawflags & drawflags_navaids) && (!m_navaidquery || m_navaidquery->is_error()))
                                || ((current_drawflags & drawflags_waypoints) && (!m_waypointquery || m_waypointquery->is_error()))
                                || ((current_drawflags & drawflags_airports) && (!m_airportquery || m_airportquery->is_error()))
                                || ((current_drawflags & drawflags_airspaces) && (!m_airspacequery || m_airspacequery->is_error()));
        }
        screenbbox = Rect(Point(), Point(Point::round<Point::coord_t,float>(screensize.getx() * m_invmapscalelon),
                                         Point::round<Point::coord_t,float>(screensize.gety() * m_invmapscale)));
        screenbbox += (surfacecenter - screenbbox.boxcenter());
        Rect dbquerybbox(screenbbox);
        TimeMeasurement tmquery;
        if (need_dbquery) {
                Point oversize(dbquerybbox.get_northeast() - dbquerybbox.get_southwest());
                oversize.set_lon(oversize.get_lon() >> 2);
                oversize.set_lat(oversize.get_lat() >> 2);
                dbquerybbox = Rect(dbquerybbox.get_southwest() - oversize, dbquerybbox.get_northeast() + oversize);
                std::cerr << "bbox: " << dbquerybbox << " (center " << surfacecenter << ')' << std::endl;
                async_cancel();
                if (current_drawflags & drawflags_topo)
                        m_topocairoquery = m_engine.async_elevation_map_cairo(dbquerybbox);
                if (current_drawflags & drawflags_terrain)
                        m_mapelementquery = m_engine.async_mapelement_find_bbox(dbquerybbox, ~0, MapelementsDb::element_t::subtables_all);
                if (current_drawflags & drawflags_navaids)
                        m_navaidquery = m_engine.async_navaid_find_bbox(dbquerybbox, ~0, NavaidsDb::element_t::subtables_all);
                if (current_drawflags & drawflags_waypoints)
                        m_waypointquery = m_engine.async_waypoint_find_bbox(dbquerybbox, ~0, WaypointsDb::element_t::subtables_all);
                if (current_drawflags & drawflags_airports)
                        m_airportquery = m_engine.async_airport_find_bbox(dbquerybbox, ~0, AirportsDb::element_t::subtables_runways | AirportsDb::element_t::subtables_vfrroutes);
                if (current_drawflags & drawflags_airspaces)
                        m_airspacequery = m_engine.async_airspace_find_bbox(dbquerybbox, ~0, AirspacesDb::element_t::subtables_none);
                if (current_drawflags & (drawflags_airways_low | drawflags_airways_high))
                        m_airwayquery = m_engine.async_airway_find_bbox(dbquerybbox, ~0, AirwaysDb::element_t::subtables_none);
                if (m_topocairoquery)
                        m_topocairoquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                if (m_mapelementquery)
                        m_mapelementquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                if (m_navaidquery)
                        m_navaidquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                if (m_waypointquery)
                        m_waypointquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                if (m_airportquery)
                        m_airportquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                if (m_airspacequery)
                        m_airspacequery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                if (m_airwayquery)
                        m_airwayquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadVMap::async_done));
                m_current_dbrect = dbquerybbox;
        }
        // Wait until topo db is ready
        {
                Glib::Mutex::Lock lock(m_mutex);
                m_abort = m_abort_cache = m_mapscalechange = false;
                while (!check_abort() && m_topocairoquery && !m_topocairoquery->is_done())
                        m_cond.wait(m_mutex);
        }
        // Allocate Pixmap
        {
                Cairo::RefPtr<Cairo::ImageSurface> surface = Cairo::ImageSurface::create(/*Cairo::FORMAT_ARGB32*/Cairo::FORMAT_RGB24, screensize.getx(), screensize.gety());
                Glib::Mutex::Lock lock(m_mutex);
                m_current_drawflags = current_drawflags;
                m_surface = surface;
                m_surfacecenter = surfacecenter;
                m_current_surfacecenter = surfacecenter;
                m_current_surfacerect = screenbbox;
                m_current_surfacesize = screensize;
                if (!m_surface) {
                        m_done = true;
                        return;
                }
        }
        // start drawing
        Cairo::RefPtr<Cairo::Context> cr = Cairo::Context::create(m_surface);
        cr->rectangle(0, 0, screensize.getx(), screensize.gety());
        cr->set_source_rgb(1.0, 1.0, 1.0);
        cr->fill_preserve();
        cr->clip();
        //cr->set_tolerance(1.0);

        // draw topography
        if ((current_drawflags & drawflags_topo) && !check_abort() && m_topocairoquery && m_topocairoquery->is_done() && !m_topocairoquery->is_error()) {
                std::cerr << "topo: cairo time " << tmquery << std::endl;
                tmquery.start();
                draw_gtopo(cr.operator->(), m_topocairoquery->get_bbox(), m_topocairoquery->get_surface());
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_mapelementquery && !m_mapelementquery->is_done())
                        m_cond.wait(m_mutex);
        }
        if ((current_drawflags & drawflags_terrain) && !check_abort() && m_mapelementquery && m_mapelementquery->is_done() && !m_mapelementquery->is_error()) {
                std::cerr << "mapel: " << m_mapelementquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                mapel_render_comp comp;
                TimeMeasurement tm;
                sort(m_mapelementquery->get_result().begin(), m_mapelementquery->get_result().end(), comp);
                std::cerr << "mapel sort: " << m_mapelementquery->get_result().size() << " time " << tm << std::endl;
                tm.start();
                unsigned int sz(0);
                for (MapelementsDb::elementvector_t::const_iterator i1(m_mapelementquery->get_result().begin()), ie(m_mapelementquery->get_result().end()); i1 != ie; i1++) {
                        if (check_abort())
                                break;
                        if (false && !(*i1).get_name().casefold().compare(0, 4, Glib::ustring("Biel").casefold()))
                                std::cerr << "Mapel: " << (*i1).get_name() << " bbox " << (*i1).get_bbox() << std::endl;
                        draw(cr.operator->(), *i1);
                        if (!i1->get_name().empty())
                                sz++;
                }
                if ((m_current_drawflags & drawflags_terrain_names) && (sz < 100)) {
                        cr->save();
                        cr->select_font_face("Sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
                        cr->set_font_size(12);
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        for (MapelementsDb::elementvector_t::const_iterator i1(m_mapelementquery->get_result().begin()), ie(m_mapelementquery->get_result().end()); i1 != ie; i1++) {
                                if (check_abort())
                                        break;
                                draw_text(cr.operator->(), *i1);
                        }
                        cr->restore();
                }
                std::cerr << "mapel drawing time " << tm << std::endl;
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_navaidquery && !m_navaidquery->is_done())
                        m_cond.wait(m_mutex);
        }
        if ((current_drawflags & drawflags_navaids) && !check_abort() && m_navaidquery && m_navaidquery->is_done() && !m_navaidquery->is_error()) {
                std::cerr << "navaids: " << m_navaidquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                draw(cr.operator->(), m_navaidquery->get_result());
                std::cerr << "navaids drawing time " << tm << std::endl;
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_waypointquery && !m_waypointquery->is_done())
                        m_cond.wait(m_mutex);
        }
        if ((current_drawflags & drawflags_waypoints) && !check_abort() && m_waypointquery && m_waypointquery->is_done() && !m_waypointquery->is_error()) {
                std::cerr << "waypoints: " << m_waypointquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                draw(cr.operator->(), m_waypointquery->get_result());
                std::cerr << "waypoints drawing time " << tm << std::endl;
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_airportquery && !m_airportquery->is_done())
                        m_cond.wait(m_mutex);
        }
        if ((current_drawflags & drawflags_airports) && !check_abort() && m_airportquery && m_airportquery->is_done() && !m_airportquery->is_error()) {
                std::cerr << "airports: " << m_airportquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                draw(cr.operator->(), m_airportquery->get_result());
                std::cerr << "airports drawing time " << tm << std::endl;
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_airspacequery && !m_airspacequery->is_done())
                        m_cond.wait(m_mutex);
        }
        if ((current_drawflags & drawflags_airspaces) && !check_abort() && m_airspacequery && m_airspacequery->is_done() && !m_airspacequery->is_error()) {
                std::cerr << "airspaces: " << m_airspacequery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                draw(cr.operator->(), m_airspacequery->get_result());
                std::cerr << "airspaces drawing time " << tm << std::endl;
                m_dispatch();
        }
        if ((current_drawflags & (drawflags_airways_low | drawflags_airways_high)) && !check_abort() && m_airwayquery && m_airwayquery->is_done() && !m_airwayquery->is_error()) {
                std::cerr << "airways: " << m_airwayquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                draw(cr.operator->(), m_airwayquery->get_result());
                std::cerr << "airways drawing time " << tm << std::endl;
                m_dispatch();
        }
        if (check_abort())
                return;
        m_done = true;
        m_dispatch();
}

void VectorMapRenderer::DrawThreadVMap::draw(Cairo::Context *cr, const MapelementsDb::Mapelement& mapel)
{
        if (false && !mapel.get_name().casefold().compare(0, 4, Glib::ustring("Biel").casefold())) {
                std::cerr << "Mapel: " << mapel.get_name() << " bbox " << mapel.get_bbox() << " type " << mapel.get_typecode() << std::endl;
                PolygonHole ph(mapel.get_polygon());
                Polygon poly(ph.get_exterior());
                std::cerr << "  Exterior:" << std::endl;
                for (Polygon::const_iterator pi(poly.begin()), pe(poly.end()); pi != pe; ++pi)
                        std::cerr << "    " << pi->get_lat_str() << ", " << pi->get_lon_str() << std::endl;
                for (unsigned int i = 0; i < ph.get_nrinterior(); ++i) {
                        Polygon poly(ph[i]);
                        std::cerr << "  Interior: " << i << std::endl;
                        for (Polygon::const_iterator pi(poly.begin()), pe(poly.end()); pi != pe; ++pi)
                                std::cerr << "    " << pi->get_lat_str() << ", " << pi->get_lon_str() << std::endl;
                }
        }
        switch (mapel.get_typecode()) {
                default:
                case MapelementsDb::Mapelement::maptyp_invalid:
                        break;

                case MapelementsDb::Mapelement::maptyp_railway:
                        cr->save();
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->set_line_width(2.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_railway_d:
                        cr->save();
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->set_line_width(3.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_aerial_cable:
                        cr->save();
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->set_line_width(1.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_highway:
                        cr->save();
                        cr->set_source_rgb(1.0, 0.0, 0.0);
                        cr->set_line_width(3.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke_preserve();
                        cr->set_source_rgb(1.0, 1.0, 1.0);
                        cr->set_line_width(1.0);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_road:
                        cr->save();
                        cr->set_source_rgb(1.0, 0.0, 0.0);
                        cr->set_line_width(2.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_trail:
                        cr->save();
                        cr->set_source_rgb(1.0, 0.0, 0.0);
                        cr->set_line_width(1.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_river:
                case MapelementsDb::Mapelement::maptyp_river_t:
                        cr->save();
                        cr->set_source_rgb(0.0, 0.0, 1.0);
                        cr->set_line_width(3.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_canal:
                        cr->save();
                        cr->set_source_rgb(0.0, 0.0, 1.0);
                        cr->set_line_width(2.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_lake:
                case MapelementsDb::Mapelement::maptyp_lake_t:
                        cr->save();
                        cr->set_source_rgb(0.0, 0.0, 1.0);
                        polypath(cr, mapel.get_polygon(), true);
                        cr->fill();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_forest:
                        cr->save();
                        cr->set_source_rgba(0.0, 1.0, 0.0, 0.5);
                        polypath(cr, mapel.get_polygon(), true);
                        cr->fill();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_pack_ice:
                case MapelementsDb::Mapelement::maptyp_glacier:
                        cr->save();
                        cr->set_source_rgba(1.0, 1.0, 1.0, 0.5);
                        polypath(cr, mapel.get_polygon(), true);
                        cr->fill();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_city:
                        cr->save();
                        cr->set_source_rgb(1.0, 1.0, 0.0);
                        polypath(cr, mapel.get_polygon(), true);
                        cr->fill();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_village:
                case MapelementsDb::Mapelement::maptyp_spot:
                case MapelementsDb::Mapelement::maptyp_landmark:
                        break;

                case MapelementsDb::Mapelement::maptyp_powerline:
                        cr->save();
                        cr->set_source_rgba(1.0, 0.0, 0.0, 0.5);
                        cr->set_line_width(2.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;

                case MapelementsDb::Mapelement::maptyp_border:
                        if (!(m_current_drawflags & drawflags_terrain_borders))
                                break;
                        cr->save();
                        cr->set_source_rgb(0.5, 0.22, 0.0);
                        cr->set_line_width(2.0);
                        polypath(cr, mapel.get_polygon(), false);
                        cr->stroke();
                        cr->restore();
                        break;
        }
}

void VectorMapRenderer::DrawThreadVMap::draw_text( Cairo::Context * cr, const MapelementsDb::Mapelement & mapel )
{
        if (mapel.get_name().empty())
                return;
#if 0
        Rect bbox(mapel.get_bbox());
        ScreenCoord dim(current_ctrans(bbox.get_northeast()) - current_ctrans(bbox.get_southwest()));
        if (dim.getx() < 10 || dim.gety() < 10)
        return;
#endif
        draw_label(cr, mapel.get_name(), mapel.get_labelcoord(), NavaidsDb::Navaid::label_center);
}

void VectorMapRenderer::DrawThreadVMap::draw( Cairo::Context * cr, const std::vector< NavaidsDb::Navaid > & navaids )
{
        cr->save();
        cr->set_source_rgb(0.0, 0.0, 0.0);
        cr->set_line_width(1.0);
        cr->set_fill_rule(Cairo::FILL_RULE_EVEN_ODD);
        for (std::vector<NavaidsDb::Navaid>::const_iterator i1(navaids.begin()), ie(navaids.end()); i1 != ie && !check_abort(); i1++) {
                const NavaidsDb::Navaid& n(*i1);
                NavaidsDb::Navaid::navaid_type_t t(n.get_navaid_type());
                switch (t) {
                        default:
                                continue;

                        case NavaidsDb::Navaid::navaid_ndb:
                                if (!(m_current_drawflags & drawflags_navaids_ndb))
                                        continue;
                                break;

                        case NavaidsDb::Navaid::navaid_vor:
                        case NavaidsDb::Navaid::navaid_vordme:
                        case NavaidsDb::Navaid::navaid_dme:
                        {
                                bool ok(false);
                                ok |= (m_current_drawflags & drawflags_navaids_vor) && (t == NavaidsDb::Navaid::navaid_vor || t == NavaidsDb::Navaid::navaid_vordme);
                                ok |= (m_current_drawflags & drawflags_navaids_dme) && (t == NavaidsDb::Navaid::navaid_dme || t == NavaidsDb::Navaid::navaid_vordme);
                                if (!ok)
                                        continue;
                                break;
                        }
                }
                draw_navaid(cr, n, current_cftrans(n.get_coord()));
        }
        cr->restore();
}

bool VectorMapRenderer::DrawThreadVMap::is_ground(const AirspacesDb::Airspace & a)
{
        uint8_t fl(a.get_altlwrflags());
        int32_t alt(a.get_altlwr());
        uint8_t tc(a.get_typecode());
        char bc(a.get_bdryclass());

        // do not fill class E, F or G airspaces
        if (tc != 255 && bc >= 'E' && bc <= 'G')
                return false;
        if (tc == 255 && (bc == 'A' || bc == 'D' || bc == 'M' || bc == 'P' || bc == 'R' || bc == 'T' || bc == 'W'))
                return false;
        if (fl & (AirspacesDb::Airspace::altflag_sfc | AirspacesDb::Airspace::altflag_gnd))
                return true;
        if (fl & AirspacesDb::Airspace::altflag_agl)
                return (alt < 500);
        return (a.get_altlwr_corr() <= a.get_gndelevmax());
}

void VectorMapRenderer::DrawThreadVMap::draw( Cairo::Context * cr, const std::vector< AirspacesDb::Airspace > & airspaces )
{
        cr->save();
        cr->set_line_width(2.0);
        cr->set_fill_rule(Cairo::FILL_RULE_EVEN_ODD);
        //std::cerr << "Airspaces: draw flags: "
        //          << ((m_current_drawflags & drawflags_airspaces_specialuse) ? "SUAS " : "")
        //          << ((m_current_drawflags & drawflags_airspaces_ab) ? "AB" : "")
        //          << ((m_current_drawflags & drawflags_airspaces_cd) ? "CD" : "")
        //          << ((m_current_drawflags & drawflags_airspaces_efg) ? "EFG" : "") << std::endl;
        for (std::vector<AirspacesDb::Airspace>::const_iterator i1(airspaces.begin()), ie(airspaces.end()); i1 != ie && !check_abort(); i1++) {
                const AirspacesDb::Airspace& a(*i1);
                // check for limit
                if (a.get_altlwr_corr() >= m_maxalt)
                        continue;
                if (a.get_typecode() == 255) {
                        // special use airspace
                        if (!(m_current_drawflags & drawflags_airspaces_specialuse))
                                continue;
                } else if (a.get_bdryclass() <= 'B') {
                        if (!(m_current_drawflags & drawflags_airspaces_ab))
                                continue;
                } else if (a.get_bdryclass() <= 'D') {
                        if (!(m_current_drawflags & drawflags_airspaces_cd))
                                continue;
                } else {
                        if (!(m_current_drawflags & drawflags_airspaces_efg))
                                continue;
                }
                Gdk::Color col = a.get_color();
                polypath(cr, a.get_polygon(), true);
                if ((m_current_drawflags & drawflags_airspaces_fill_ground) && is_ground(a)) {
                        cr->set_source_rgba(col.get_red_p(), col.get_green_p(), col.get_blue_p(), 0.5);
                        cr->fill_preserve();
                }
                cr->set_source_rgb(col.get_red_p(), col.get_green_p(), col.get_blue_p());
                cr->stroke();
                Glib::ustring buf1(a.get_name());
                Glib::ustring buf2(a.get_altlwr_string());
                Glib::ustring buf3(a.get_commname());
                Glib::ustring tmp1(a.get_class_string());
                if (!buf1.empty() && !tmp1.empty())
                        buf1 += ' ';
                buf1 += tmp1;
                tmp1 = a.get_altupr_string();
                if (!buf2.empty() && !tmp1.empty())
                        buf2 += ' ';
                buf2 += tmp1;
                tmp1 = freqstr(a.get_commfreq());
                if (!buf3.empty() && !tmp1.empty())
                        buf3 += ' ';
                buf3 += tmp1;
                Cairo::TextExtents ext1, ext2, ext3;
                cr->get_text_extents(buf1, ext1);
                cr->get_text_extents(buf2, ext2);
                cr->get_text_extents(buf3, ext3);
                double width = std::max(std::max(ext1.width, ext2.width), ext3.width);
                double height = ext1.height + ext2.height + ext3.height + 4;
                ScreenCoordFloat px(current_cftrans(a.get_labelcoord()));
                cr->move_to(px.getx() - width * 0.5, px.gety() + height * 0.5);
                cr->show_text(buf3);
                cr->move_to(px.getx() - width * 0.5, px.gety() + height * 0.5 - ext3.height - 2);
                cr->show_text(buf2);
                cr->move_to(px.getx() - width * 0.5, px.gety() + height * 0.5 - ext3.height - ext2.height - 4);
                cr->show_text(buf1);
        }
        cr->restore();
}

void VectorMapRenderer::DrawThreadVMap::draw( Cairo::Context * cr, const std::vector< AirportsDb::Airport > & airports )
{
        cr->save();
        cr->set_line_width(1.0);
        for (std::vector<AirportsDb::Airport>::const_iterator i1(airports.begin()), ie(airports.end()); i1 != ie && !check_abort(); i1++) {
                const AirportsDb::Airport& a(*i1);
                bool ismil(false);
                switch (a.get_typecode()) {
                        case 'A':
                                if (!(m_current_drawflags & drawflags_airports_civ))
                                        continue;
                                break;

                        case 'C':
                                if (!(m_current_drawflags & drawflags_airports_mil))
                                        continue;
                                ismil = true;
                                break;

                        default:
                                break;
                }
                int rwydir = -1, rwylen = -1;
                bool rwyasp = false;
                for (unsigned int i = 0; i < a.get_nrrwy(); i++) {
                        const AirportsDb::Airport::Runway& rwy(a.get_rwy(i));
                        bool thisrwyasp = rwy.is_concrete();
                        if (rwy.get_length() > rwylen || (thisrwyasp && !rwyasp)) {
                                rwylen = rwy.get_length();
                                rwydir = rwy.get_he_hdg();
                                rwyasp |= thisrwyasp;
                        }
                        // draw localizer cones
                        if (!(m_current_drawflags & drawflags_airports_localizercone))
                                continue;
                        if (rwy.is_he_usable()) {
                                //std::cerr << a.get_icao() << " RWY " << rwy.get_ident_he() << std::endl;
                                Point pt(rwy.get_he_coord());
                                float ang((rwy.get_he_hdg() + 0x8000) * FPlanLeg::to_deg);
                                Point ptc(pt.spheric_course_distance_nmi(ang, aiport_loccone_centerlength));
                                Point pts0(pt.spheric_course_distance_nmi(ang - aiport_loccone_sideangle, aiport_loccone_sidelength));
                                Point pts1(pt.spheric_course_distance_nmi(ang + aiport_loccone_sideangle, aiport_loccone_sidelength));
                                ScreenCoordFloat px(current_cftrans(pt));
                                ScreenCoordFloat pxc(current_cftrans(ptc));
                                ScreenCoordFloat pxs0(current_cftrans(pts0));
                                ScreenCoordFloat pxs1(current_cftrans(pts1));
                                ScreenCoordFloat pxd = (pxc - px);
                                if (pxd.getx() * pxd.getx() + pxd.gety() * pxd.gety() >= 100.0f) {
                                        cr->set_source_rgb(0.0 / 255, 148.0 / 255, 0.0 / 255);
                                        cr->move_to(px.getx(), px.gety());
                                        cr->line_to(pxs0.getx(), pxs0.gety());
                                        cr->line_to(pxc.getx(), pxc.gety());
                                        cr->line_to(pxs1.getx(), pxs1.gety());
                                        cr->line_to(px.getx(), px.gety());
                                        cr->line_to(pxc.getx(), pxc.gety());
                                        cr->stroke();
                                }
                        }
                        if (rwy.is_le_usable()) {
                                //std::cerr << a.get_icao() << " RWY " << rwy.get_ident_le() << std::endl;
                                Point pt(rwy.get_le_coord());
                                float ang((rwy.get_le_hdg() + 0x8000) * FPlanLeg::to_deg);
                                Point ptc(pt.spheric_course_distance_nmi(ang, aiport_loccone_centerlength));
                                Point pts0(pt.spheric_course_distance_nmi(ang - aiport_loccone_sideangle, aiport_loccone_sidelength));
                                Point pts1(pt.spheric_course_distance_nmi(ang + aiport_loccone_sideangle, aiport_loccone_sidelength));
                                ScreenCoordFloat px(current_cftrans(pt));
                                ScreenCoordFloat pxc(current_cftrans(ptc));
                                ScreenCoordFloat pxs0(current_cftrans(pts0));
                                ScreenCoordFloat pxs1(current_cftrans(pts1));
                                ScreenCoordFloat pxd = (pxc - px);
                                if (pxd.getx() * pxd.getx() + pxd.gety() * pxd.gety() >= 100.0f) {
                                        cr->set_source_rgb(0.0 / 255, 148.0 / 255, 0.0 / 255);
                                        cr->move_to(px.getx(), px.gety());
                                        cr->line_to(pxs0.getx(), pxs0.gety());
                                        cr->line_to(pxc.getx(), pxc.gety());
                                        cr->line_to(pxs1.getx(), pxs1.gety());
                                        cr->line_to(px.getx(), px.gety());
                                        cr->line_to(pxc.getx(), pxc.gety());
                                        cr->stroke();
                                }
                        }
                }
                if (ismil)
                        cr->set_source_rgb(167.0 / 255, 101.0 / 255, 160.0 / 255);
                else
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                ScreenCoordFloat px(current_cftrans(a.get_coord()));
                if (rwydir == -1) {
                        cr->arc(px.getx(), px.gety(), 15, 0, 2.0*M_PI);
                        cr->stroke();
                } else {
                        //if (a.get_icao() == "LSZK")
                        //        std::cerr << a.get_icao() << " rwydir " << rwydir << std::endl;
                        double phi(rwydir * FPlanLeg::to_rad);
                        double c(cos(phi)), s(sin(phi));
                        ScreenCoordFloat p0(-2 * c + 20 * s, -2 * s - 20 * c);
                        ScreenCoordFloat p1(2 * c + 20 * s, 2 * s - 20 * c);
                        cr->move_to(px.getx() + p0.getx(), px.gety() + p0.gety());
                        cr->line_to(px.getx() + p1.getx(), px.gety() + p1.gety());
                        cr->line_to(px.getx() - p0.getx(), px.gety() - p0.gety());
                        cr->line_to(px.getx() - p1.getx(), px.gety() - p1.gety());
                        cr->close_path();
                        if (rwyasp)
                                cr->fill_preserve();
                        cr->stroke();
                        cr->arc(px.getx(), px.gety(), 15, phi + (0.5 * M_PI + 0.133), phi + (1.5 * M_PI - 0.133));
                        cr->stroke();
                        cr->arc(px.getx(), px.gety(), 15, phi + (1.5 * M_PI + 0.133), phi + (2.5 * M_PI - 0.133));
                        cr->stroke();
                }
                Glib::ustring buf1(a.get_icao());
                Glib::ustring buf2(a.get_name());
                if (!buf1.empty() && !buf2.empty())
                        buf1 += ' ';
                buf1 += buf2;
                draw_label(cr, buf1, a.get_coord(), a.get_label_placement(), 20);
                draw_airport_vfrreppt(cr, a);
        }
        cr->restore();
}

void VectorMapRenderer::DrawThreadVMap::draw( Cairo::Context * cr, const std::vector< WaypointsDb::Waypoint > & waypoints )
{
        for (std::vector<WaypointsDb::Waypoint>::const_iterator i1(waypoints.begin()), ie(waypoints.end()); i1 != ie && !check_abort(); i1++) {
                const WaypointsDb::Waypoint& w(*i1);
                switch (w.get_usage()) {
                        case WaypointsDb::Waypoint::usage_highlevel:
                                if (!(m_current_drawflags & drawflags_waypoints_high))
                                        continue;
                                break;

                        case WaypointsDb::Waypoint::usage_lowlevel:
                                if (!(m_current_drawflags & drawflags_waypoints_low))
                                        continue;
                                break;

                        case WaypointsDb::Waypoint::usage_bothlevel:
                                if (!(m_current_drawflags & (drawflags_waypoints_low | drawflags_waypoints_high)))
                                        continue;
                                break;

                        case WaypointsDb::Waypoint::usage_rnav:
                                if (!(m_current_drawflags & drawflags_waypoints_rnav))
                                        continue;
                                break;

                        case WaypointsDb::Waypoint::usage_terminal:
                                if (!(m_current_drawflags & drawflags_waypoints_terminal))
                                        continue;
                                break;

                        case WaypointsDb::Waypoint::usage_vfr:
                                if (!(m_current_drawflags & drawflags_waypoints_vfr))
                                        continue;
                                break;

                        default:
                        case WaypointsDb::Waypoint::usage_user:
                                if (!(m_current_drawflags & drawflags_waypoints_user))
                                        continue;
                                break;
                }
                draw_waypoint(cr, w.get_coord(), w.get_name(), true, w.get_label_placement());
        }
}

void VectorMapRenderer::DrawThreadVMap::draw(Cairo::Context * cr, const std::vector< AirwaysDb::Airway > & airways)
{
        cr->save();
        cr->set_line_width(1.0);
        cr->set_source_rgb(0.0 / 255, 0.0 / 255, 0.0 / 255);
        for (std::vector<AirwaysDb::Airway>::const_iterator i1(airways.begin()), ie(airways.end()); i1 != ie && !check_abort(); i1++) {
                const AirwaysDb::Airway& awy(*i1);
                switch (awy.get_type()) {
                        case AirwaysDb::Airway::airway_high:
                                if (!(m_current_drawflags & drawflags_airways_high))
                                        continue;
                                break;

                        case AirwaysDb::Airway::airway_low:
                                if (!(m_current_drawflags & drawflags_airways_low))
                                        continue;
                                break;

                        case AirwaysDb::Airway::airway_invalid:
                                continue;

                        default:
                                break;
                }
                {
                        ScreenCoordFloat pb(current_cftrans(awy.get_begin_coord()));
                        ScreenCoordFloat pe(current_cftrans(awy.get_end_coord()));
                        cr->move_to(pb.getx(), pb.gety());
                        cr->line_to(pe.getx(), pe.gety());
                        cr->stroke();
                        draw_label(cr, awy.get_name(), awy.get_labelcoord(), awy.get_label_placement(), 20);
                }
        }
        cr->restore();
}

bool VectorMapRenderer::DrawThreadVMap::mapel_render_comp::operator ( )( const MapelementsDb::Mapelement & m1, const MapelementsDb::Mapelement & m2 ) const
{
        if (m1.get_typecode() == MapelementsDb::Mapelement::maptyp_invalid && m2.get_typecode() != MapelementsDb::Mapelement::maptyp_invalid)
                return false;
        if (m1.get_typecode() != MapelementsDb::Mapelement::maptyp_invalid && m2.get_typecode() == MapelementsDb::Mapelement::maptyp_invalid)
                return true;
        if (m1.get_typecode() == MapelementsDb::Mapelement::maptyp_pack_ice && m2.get_typecode() != MapelementsDb::Mapelement::maptyp_pack_ice)
                return true;
        if (m1.get_typecode() != MapelementsDb::Mapelement::maptyp_pack_ice && m2.get_typecode() == MapelementsDb::Mapelement::maptyp_pack_ice)
                return false;
        if (m1.get_typecode() == MapelementsDb::Mapelement::maptyp_glacier && m2.get_typecode() != MapelementsDb::Mapelement::maptyp_glacier)
                return true;
        if (m1.get_typecode() != MapelementsDb::Mapelement::maptyp_glacier && m2.get_typecode() == MapelementsDb::Mapelement::maptyp_glacier)
                return false;
        if (m1.get_typecode() == MapelementsDb::Mapelement::maptyp_forest && m2.get_typecode() != MapelementsDb::Mapelement::maptyp_forest)
                return true;
        if (m1.get_typecode() != MapelementsDb::Mapelement::maptyp_forest && m2.get_typecode() == MapelementsDb::Mapelement::maptyp_forest)
                return false;
        if ((m1.get_typecode() == MapelementsDb::Mapelement::maptyp_lake || m1.get_typecode() == MapelementsDb::Mapelement::maptyp_lake_t) &&
             (m2.get_typecode() != MapelementsDb::Mapelement::maptyp_lake && m2.get_typecode() != MapelementsDb::Mapelement::maptyp_lake_t))
                return true;
        if ((m1.get_typecode() != MapelementsDb::Mapelement::maptyp_lake && m1.get_typecode() != MapelementsDb::Mapelement::maptyp_lake_t) &&
             (m2.get_typecode() == MapelementsDb::Mapelement::maptyp_lake || m2.get_typecode() == MapelementsDb::Mapelement::maptyp_lake_t))
                return false;
        if (m1.get_typecode() == MapelementsDb::Mapelement::maptyp_city && m2.get_typecode() != MapelementsDb::Mapelement::maptyp_city)
                return true;
        if (m1.get_typecode() != MapelementsDb::Mapelement::maptyp_city && m2.get_typecode() == MapelementsDb::Mapelement::maptyp_city)
                return false;
        return false;
}

VectorMapRenderer::DrawThreadTerrain::DrawThreadTerrain(Engine & eng, int depth, const ScreenCoord & scrsize, const Point & center, int alt)
        : DrawThread(eng, depth, scrsize, center, alt),
          m_current_surfacerect(center, center), m_current_dbrect(center, center), m_abort_cache(false)
{
}

VectorMapRenderer::DrawThreadTerrain::~DrawThreadTerrain()
{
        async_cancel();
        Glib::Mutex::Lock lock(m_mutex);
}

void VectorMapRenderer::DrawThreadTerrain::airspaces_changed(void)
{
}

void VectorMapRenderer::DrawThreadTerrain::airports_changed(void)
{
}

void VectorMapRenderer::DrawThreadTerrain::navaids_changed(void)
{
}

void VectorMapRenderer::DrawThreadTerrain::waypoints_changed(void)
{
}

void VectorMapRenderer::DrawThreadTerrain::airways_changed(void)
{
}

void VectorMapRenderer::DrawThreadTerrain::async_cancel(void)
{
        Engine::ElevationMapResult::cancel(m_topoquery);
}

void VectorMapRenderer::DrawThreadTerrain::async_done(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_cond.broadcast();
}

bool VectorMapRenderer::DrawThreadTerrain::check_abort(void)
{
        if (m_hidden || m_terminate || m_abort_cache)
                return true;
        if (!m_abort)
                return false;
        if (m_mapscalechange) {
                m_abort_cache = true;
                return true;
        }
        if (abs(m_altitude - m_current_surfacealtitude) > 10) {
                m_abort_cache = true;
                return true;
        }
        Rect scrbbox(Point(), Point(Point::round<Point::coord_t,float>(m_screensize.getx() * m_invmapscalelon),
                                    Point::round<Point::coord_t,float>(m_screensize.gety() * m_invmapscale)));
        scrbbox += (m_center - scrbbox.boxcenter());
        if (!m_current_surfacerect.is_inside(scrbbox)) {
                m_abort_cache = true;
                return true;
        }
        std::cerr << "cancelling abort: new screen rect " << scrbbox << " surface rect " << m_current_surfacerect << std::endl;
        m_abort = false;
        return false;
}

void VectorMapRenderer::DrawThreadTerrain::compute_pixmap(void)
{
        int currentalt;
        Point surfacecenter;
        ScreenCoord screensize(0, 0);
        bool need_redraw;
        {
                Glib::Mutex::Lock lock(m_mutex);
                currentalt = m_altitude;
                surfacecenter = m_center;
                screensize = m_screensize;
                need_redraw = check_abort() || !m_surface;
        }
        if (!need_redraw)
                return;
        m_done = false;
        Rect screenbbox(Point(), Point(Point::round<Point::coord_t,float>(screensize.getx() * m_invmapscalelon),
                                       Point::round<Point::coord_t,float>(screensize.gety() * m_invmapscale)));
        screenbbox += (surfacecenter - screenbbox.boxcenter());
        screensize = ScreenCoord((screensize.getx() * 3) >> 1, (screensize.gety() * 2) >> 1);
        bool need_dbquery;
        {
                Glib::Mutex::Lock lock(m_mutex);
                need_dbquery = (!m_current_dbrect.is_inside(screenbbox))
                                || (!m_topoquery || m_topoquery->is_error());
        }
        screenbbox = Rect(Point(), Point(Point::round<Point::coord_t,float>(screensize.getx() * m_invmapscalelon),
                                         Point::round<Point::coord_t,float>(screensize.gety() * m_invmapscale)));
        screenbbox += (surfacecenter - screenbbox.boxcenter());
        Rect dbquerybbox(screenbbox);
        TimeMeasurement tmquery;
        if (need_dbquery) {
                //Point oversize(dbquerybbox.get_northeast() - dbquerybbox.get_southwest());
                //oversize.set_lon(oversize.get_lon() >> 2);
                //oversize.set_lat(oversize.get_lat() >> 2);
                //dbquerybbox = Rect(dbquerybbox.get_southwest() - oversize, dbquerybbox.get_northeast() + oversize);
                std::cerr << "bbox: " << dbquerybbox << " (center " << surfacecenter << ')' << std::endl;
                async_cancel();
                m_topoquery = m_engine.async_elevation_map(dbquerybbox);
                if (m_topoquery)
                        m_topoquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadTerrain::async_done));
        }
        // Wait until topo db is ready
        {
                Glib::Mutex::Lock lock(m_mutex);
                m_abort = m_abort_cache = m_mapscalechange = false;
                while (!check_abort() && m_topoquery && !m_topoquery->is_done())
                        m_cond.wait(m_mutex);
        }
        // Allocate Pixmap
        m_current_surfacecenter = surfacecenter;
        m_current_surfacerect = screenbbox;
        m_current_surfacesize = screensize;
        m_current_surfacealtitude = currentalt;
        Cairo::RefPtr<Cairo::ImageSurface> surface = Cairo::ImageSurface::create(/*Cairo::FORMAT_ARGB32*/Cairo::FORMAT_RGB24, screensize.getx(), screensize.gety());
        if (!surface) {
                Glib::Mutex::Lock lock(m_mutex);
                m_surface = Cairo::RefPtr<Cairo::ImageSurface>();
                m_surfacecenter = surfacecenter;
                m_done = true;
                return;
        }
        // start drawing
        Cairo::RefPtr<Cairo::Context> cr = Cairo::Context::create(surface);
        cr->rectangle(0, 0, screensize.getx(), screensize.gety());
        cr->set_source_rgb(1.0, 1.0, 1.0);
        cr->fill_preserve();
        cr->clip();
        //cr->set_tolerance(1.0);

        // draw topography
        if (!check_abort() && m_topoquery && m_topoquery->is_done() && !m_topoquery->is_error()) {
                std::cerr << "topo: time " << tmquery << std::endl;
                tmquery.start();
                unsigned int width(m_topoquery->get_width()), height(m_topoquery->get_height());
                Cairo::RefPtr<Cairo::ImageSurface> surface = Cairo::ImageSurface::create(/*Cairo::FORMAT_ARGB32*/Cairo::FORMAT_RGB24, width, height);
                if (surface) {
                        unsigned int stride(surface->get_stride());
                        unsigned char *data(surface->get_data());
                        std::cerr << "W " << width << " H " << height << " stride " << stride << std::endl;
                        for (unsigned int y = 0; y < height; y++) {
                                for (unsigned int x = 0; x < width; x++) {
                                        unsigned char *d = data + 4 * x + stride * y;
                                        TopoDb30::elev_t gndelev(m_topoquery.operator->()->operator()(x, y));
                                        //std::cerr << "x " << x << " y " << y << " gndelev " << gndelev << " alt " << currentalt << std::endl;
                                        d[3] = 0;
                                        if (gndelev == TopoDb30::nodata) {
                                                d[2] = 64; // red
                                                d[1] = 0; // green
                                                d[0] = 0; // blue
                                                continue;
                                        }
                                        if (gndelev == TopoDb30::ocean)
                                                gndelev = 0;
                                        int elev = currentalt - TopoDb30::m_to_ft(gndelev);
                                        //std::cerr << "x " << x << " y " << y << " gndelev " << gndelev << " alt " << currentalt << " elev " << elev << std::endl;
                                        if (elev < 200) {
                                                d[2] = 255; // red
                                                d[1] = 0; // green
                                                d[0] = 0; // blue
                                        } else if (elev < 1000) {
                                                d[2] = 255; // red
                                                d[1] = 255; // green
                                                d[0] = 0; // blue
                                        } else {
                                                d[2] = 0; // red
                                                d[1] = 0; // green
                                                d[0] = 0; // blue
                                        }
                                }
                        }
                        draw_gtopo(cr.operator->(), m_topoquery->get_bbox(), surface);
                        std::cerr << "topo: draw time " << tmquery << std::endl;
                }
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                if (check_abort())
                        return;
                m_done = true;
                m_surface = surface;
                m_surfacecenter = surfacecenter;
        }
        std::cerr << "updated image: center " << surfacecenter << " bbox " << screenbbox << " alt " << currentalt << std::endl;
        m_dispatch();
}

VectorMapRenderer::DrawThreadAirportDiagram::DrawThreadAirportDiagram(Engine & eng, int depth, const ScreenCoord & scrsize, const Point & center, int alt)
        : DrawThread(eng, depth, scrsize, center, alt),
          m_current_surfacerect(center, center), m_current_dbrect(center, center), m_abort_cache(false)
{
}

VectorMapRenderer::DrawThreadAirportDiagram::~DrawThreadAirportDiagram()
{
        async_cancel();
        Glib::Mutex::Lock lock(m_mutex);
}

void VectorMapRenderer::DrawThreadAirportDiagram::airspaces_changed(void)
{
}

void VectorMapRenderer::DrawThreadAirportDiagram::airports_changed(void)
{
        Engine::AirportResult::cancel(m_airportquery);
        m_mapscalechange = true;
        DrawThread::airports_changed();
}

void VectorMapRenderer::DrawThreadAirportDiagram::navaids_changed(void)
{
}

void VectorMapRenderer::DrawThreadAirportDiagram::waypoints_changed(void)
{
}

void VectorMapRenderer::DrawThreadAirportDiagram::airways_changed(void)
{
}

bool VectorMapRenderer::DrawThreadAirportDiagram::check_abort(void)
{
        if (m_hidden || m_terminate || m_abort_cache)
                return true;
        if (!m_abort)
                return false;
        if (m_mapscalechange) {
                m_abort_cache = true;
                return true;
        }
        Rect scrbbox(Point(), Point(Point::round<Point::coord_t,float>(m_screensize.getx() * m_invmapscalelon),
                                    Point::round<Point::coord_t,float>(m_screensize.gety() * m_invmapscale)));
        scrbbox += (m_center - scrbbox.boxcenter());
        if (!m_current_surfacerect.is_inside(scrbbox)) {
                m_abort_cache = true;
                return true;
        }
        std::cerr << "cancelling abort: new screen rect " << scrbbox << " surface rect " << m_current_surfacerect << std::endl;
        m_abort = false;
        return false;
}

void VectorMapRenderer::DrawThreadAirportDiagram::compute_pixmap(void)
{
        int currentalt;
        Point surfacecenter;
        ScreenCoord screensize(0, 0);
        bool need_redraw;
        {
                Glib::Mutex::Lock lock(m_mutex);
                currentalt = m_altitude;
                surfacecenter = m_center;
                screensize = m_screensize;
                need_redraw = check_abort() || !m_surface;
        }
        if (!need_redraw)
                return;
        m_done = false;
        Rect screenbbox(Point(), Point(Point::round<Point::coord_t,float>(screensize.getx() * m_invmapscalelon),
                                       Point::round<Point::coord_t,float>(screensize.gety() * m_invmapscale)));
        screenbbox += (surfacecenter - screenbbox.boxcenter());
        screensize = ScreenCoord((screensize.getx() * 3) >> 1, (screensize.gety() * 2) >> 1);
        bool need_dbquery;
        {
                Glib::Mutex::Lock lock(m_mutex);
                need_dbquery = (!m_current_dbrect.is_inside(screenbbox))
                                || (!m_airportquery || m_airportquery->is_error());
        }
        screenbbox = Rect(Point(), Point(Point::round<Point::coord_t,float>(screensize.getx() * m_invmapscalelon),
                                         Point::round<Point::coord_t,float>(screensize.gety() * m_invmapscale)));
        screenbbox += (surfacecenter - screenbbox.boxcenter());
        Rect dbquerybbox(screenbbox);
        TimeMeasurement tmquery;
        if (need_dbquery) {
                Point oversize(dbquerybbox.get_northeast() - dbquerybbox.get_southwest());
                oversize.set_lon(oversize.get_lon() >> 2);
                oversize.set_lat(oversize.get_lat() >> 2);
                dbquerybbox = Rect(dbquerybbox.get_southwest() - oversize, dbquerybbox.get_northeast() + oversize);
                std::cerr << "bbox: " << dbquerybbox << " (center " << surfacecenter << ')' << std::endl;
                async_cancel();
                m_airportquery = m_engine.async_airport_find_bbox(dbquerybbox, ~0, AirportsDb::element_t::subtables_all);
                m_navaidquery = m_engine.async_navaid_find_bbox(dbquerybbox, ~0, NavaidsDb::element_t::subtables_all);
                m_waypointquery = m_engine.async_waypoint_find_bbox(dbquerybbox, ~0, WaypointsDb::element_t::subtables_all);
                if (m_airportquery)
                        m_airportquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadAirportDiagram::async_done));
                if (m_navaidquery)
                        m_navaidquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadAirportDiagram::async_done));
                if (m_waypointquery)
                        m_waypointquery->connect(sigc::mem_fun(*this, &VectorMapRenderer::DrawThreadAirportDiagram::async_done));
                m_current_dbrect = dbquerybbox;
        }
        // Wait until airport db is ready
        {
                Glib::Mutex::Lock lock(m_mutex);
                m_abort = m_abort_cache = m_mapscalechange = false;
                while (!check_abort() && m_airportquery && !m_airportquery->is_done())
                        m_cond.wait(m_mutex);
        }
        // Allocate Pixmap
        m_current_surfacecenter = surfacecenter;
        m_current_surfacerect = screenbbox;
        m_current_surfacesize = screensize;
        Cairo::RefPtr<Cairo::ImageSurface> surface = Cairo::ImageSurface::create(/*Cairo::FORMAT_ARGB32*/Cairo::FORMAT_RGB24, screensize.getx(), screensize.gety());
        if (!surface) {
                Glib::Mutex::Lock lock(m_mutex);
                m_surface = Cairo::RefPtr<Cairo::ImageSurface>();
                m_surfacecenter = surfacecenter;
                m_done = true;
                return;
        }
        // start drawing
        Cairo::RefPtr<Cairo::Context> cr = Cairo::Context::create(surface);
        cr->rectangle(0, 0, screensize.getx(), screensize.gety());
        cr->set_source_rgb(1.0, 1.0, 1.0);
        cr->fill_preserve();
        cr->clip();
        //cr->set_tolerance(1.0);

        // draw topography
        if (!check_abort() && m_airportquery && m_airportquery->is_done() && !m_airportquery->is_error()) {
                std::cerr << "airport: time " << tmquery << std::endl;
                tmquery.start();
                const AirportsDb::elementvector_t& ev(m_airportquery->get_result());
                for (AirportsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee && !check_abort(); ++ei)
                        draw(cr.operator->(), *ei);
                std::cerr << "airport: draw time " << tmquery << std::endl;
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_navaidquery && !m_navaidquery->is_done())
                        m_cond.wait(m_mutex);
        }
        if (!check_abort() && m_navaidquery && m_navaidquery->is_done() && !m_navaidquery->is_error()) {
                std::cerr << "navaids: " << m_navaidquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                const NavaidsDb::elementvector_t& ev(m_navaidquery->get_result());
                for (NavaidsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee && !check_abort(); ++ei)
                        draw(cr.operator->(), *ei);
                std::cerr << "navaids drawing time " << tm << std::endl;
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                while (!check_abort() && m_waypointquery && !m_waypointquery->is_done())
                        m_cond.wait(m_mutex);
        }
        if (!check_abort() && m_waypointquery && m_waypointquery->is_done() && !m_waypointquery->is_error()) {
                std::cerr << "waypoints: " << m_waypointquery->get_result().size() << " time " << tmquery << std::endl;
                tmquery.start();
                TimeMeasurement tm;
                const WaypointsDb::elementvector_t& ev(m_waypointquery->get_result());
                for (WaypointsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee && !check_abort(); ++ei)
                        draw(cr.operator->(), *ei);
                std::cerr << "waypoints drawing time " << tm << std::endl;
                m_dispatch();
        }
        {
                Glib::Mutex::Lock lock(m_mutex);
                if (check_abort())
                        return;
                m_done = true;
                m_surface = surface;
                m_surfacecenter = surfacecenter;
        }
        std::cerr << "updated image: center " << surfacecenter << " bbox " << screenbbox << " alt " << currentalt << std::endl;
        m_dispatch();
}

void VectorMapRenderer::DrawThreadAirportDiagram::async_cancel(void)
{
        Engine::AirportResult::cancel(m_airportquery);
        Engine::NavaidResult::cancel(m_navaidquery);
        Engine::WaypointResult::cancel(m_waypointquery);
}

void VectorMapRenderer::DrawThreadAirportDiagram::async_done(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_cond.broadcast();
}

void VectorMapRenderer::DrawThreadAirportDiagram::polyline_to_path(Cairo::Context *cr, const AirportsDb::Airport::Polyline& pl)
{
        if (pl.empty())
                return;
        const AirportsDb::Airport::PolylineNode *pnfirst = 0;
        ScreenCoordFloat sc0, sc1, scfirst;
        for (AirportsDb::Airport::Polyline::const_iterator ni(pl.begin()), ni0, ne(pl.end()); ni != ne; ni0 = ni, sc0 = sc1, ++ni) {
                const AirportsDb::Airport::PolylineNode& pn1(*ni);
                sc1 = current_cftrans(pn1.get_point());
                if (!pnfirst) {
                        pnfirst = &pn1;
                        scfirst = sc1;
                        cr->move_to(sc1.getx(), sc1.gety());
                        continue;
                }
                const AirportsDb::Airport::PolylineNode& pn0(*ni0);
                if (pn0.is_bezier_next() || pn1.is_bezier_prev()) {
                        ScreenCoordFloat scc0(sc0);
                        if (pn0.is_bezier_next())
                                scc0 = current_cftrans(pn0.get_next_controlpoint());
                        ScreenCoordFloat scc1(sc1);
                        if (pn1.is_bezier_prev())
                                scc1 = current_cftrans(pn1.get_prev_controlpoint());
                        cr->curve_to(scc0.getx(), scc0.gety(), scc1.getx(), scc1.gety(), sc1.getx(), sc1.gety());
                } else {
                        cr->line_to(sc1.getx(), sc1.gety());
                }
                if (pn1.is_closepath()) {
                        if (pn0.is_bezier_next() || pn1.is_bezier_prev()) {
                                ScreenCoordFloat scc0(sc1);
                                if (pn1.is_bezier_next())
                                        scc0 = current_cftrans(pn1.get_next_controlpoint());
                                ScreenCoordFloat scc1(scfirst);
                                if (pnfirst->is_bezier_prev())
                                        scc1 = current_cftrans(pnfirst->get_prev_controlpoint());
                                cr->curve_to(scc0.getx(), scc0.gety(), scc1.getx(), scc1.gety(), scfirst.getx(), scfirst.gety());
                        } else {
                                cr->line_to(scfirst.getx(), scfirst.gety());
                        }
                        cr->close_path();
                }
                if (pn1.is_closepath() || pn1.is_endpath()) {
                        pnfirst = 0;
                }
        }
}

void VectorMapRenderer::DrawThreadAirportDiagram::draw(Cairo::Context *cr, const AirportsDb::Airport& arpt)
{
        float featurescale(nmi_to_pixel(Point::km_to_nmi * 0.002));
        cr->save();
        // draw line features
        cr->set_line_width(1.0);
        cr->set_fill_rule(Cairo::FILL_RULE_WINDING);
        // first pass: area features
        for (unsigned int r = 0; r < arpt.get_nrlinefeatures(); ++r) {
                const AirportsDb::Airport::Polyline& pl(arpt.get_linefeature(r));
                if (!pl.is_area())
                        continue;
                polyline_to_path(cr, pl);
                if (pl.is_concrete())
                        cr->set_source_rgba(192.0/256.0, 192.0/256.0, 192.0/256.0, 1.0);
                else
                        cr->set_source_rgba(0.0/256.0, 150.0/256.0, 0.0/256.0, 1.0);
                cr->fill_preserve();
                if (pl.is_airportboundary()) {
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->stroke_preserve();
                }
                cr->begin_new_path();
        }
        cr->set_line_width(1.0);
        for (unsigned int r = 0; r < arpt.get_nrrwy(); ++r) {
                const AirportsDb::Airport::Runway& rwy(arpt.get_rwy(r));
                float halfwidth(rwy.get_width() * (0.5f * Point::ft_to_m * (1.0f/1000.f) * Point::km_to_nmi));
                Point rept0(rwy.get_le_coord().spheric_course_distance_nmi((rwy.get_le_hdg() - 0x4000) * FPlanLeg::to_deg, halfwidth));
                Point rept1(rwy.get_le_coord().spheric_course_distance_nmi((rwy.get_le_hdg() + 0x4000) * FPlanLeg::to_deg, halfwidth));
                Point rept2(rwy.get_he_coord().spheric_course_distance_nmi((rwy.get_he_hdg() - 0x4000) * FPlanLeg::to_deg, halfwidth));
                Point rept3(rwy.get_he_coord().spheric_course_distance_nmi((rwy.get_he_hdg() + 0x4000) * FPlanLeg::to_deg, halfwidth));
                ScreenCoordFloat repx0(current_cftrans(rept0));
                ScreenCoordFloat repx1(current_cftrans(rept1));
                ScreenCoordFloat repx2(current_cftrans(rept2));
                ScreenCoordFloat repx3(current_cftrans(rept3));
                if (rwy.is_concrete())
                        cr->set_source_rgba(192.0/256.0, 192.0/256.0, 192.0/256.0, 1.0);
                else
                        cr->set_source_rgba(0.0/256.0, 150.0/256.0, 0.0/256.0, 1.0);
                cr->move_to(repx0.getx(), repx0.gety());
                cr->line_to(repx1.getx(), repx1.gety());
                cr->line_to(repx2.getx(), repx2.gety());
                cr->line_to(repx3.getx(), repx3.gety());
                cr->line_to(repx0.getx(), repx0.gety());
                cr->fill();
                float rwyhalfwidth(nmi_to_pixel(rwy.get_width() * (Point::ft_to_m / 2000.0f * Point::km_to_nmi)));
                // LE headend
                {
                        cr->save();
                        {
                                Point ptcenter(rwy.get_le_coord());
                                if (rwy.get_le_disp())
                                        ptcenter = ptcenter.spheric_course_distance_nmi(rwy.get_le_hdg() * FPlanLeg::to_deg,
                                                                                        rwy.get_le_disp() * (Point::ft_to_m * (1.0f/1000.f) * Point::km_to_nmi));
                                ScreenCoordFloat pxcenter(current_cftrans(ptcenter));
                                cr->translate(pxcenter.getx(), pxcenter.gety());
                        }
                        cr->rotate((rwy.get_le_hdg() + 0x8000) * FPlanLeg::to_rad);
                        //std::cerr << "scale: " << rwyhalfwidth << " width(nmi): " << (rwy.get_width() * (Point::ft_to_m / 2000.0f * Point::km_to_nmi)) << std::endl;
                        cr->scale(rwyhalfwidth, rwyhalfwidth);
                        cr->set_line_width(0.2);
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->move_to(-1.0, 0.0);
                        cr->line_to(1.0, 0.0);
                        cr->stroke();
                        if (rwy.is_le_usable()) {
                                for (unsigned int i = 0; i < 5; ++i) {
                                        cr->move_to((1.0/11.0) * (1 + 2 * i), 0.2);
                                        cr->rel_line_to((1.0/11.0), 0.0);
                                        cr->rel_line_to(0.0, 1.0);
                                        cr->rel_line_to(-(1.0/11.0), 0.0);
                                        cr->close_path();
                                        cr->fill();
                                        cr->move_to(-(1.0/11.0) * (1 + 2 * i), 0.2);
                                        cr->rel_line_to(-(1.0/11.0), 0.0);
                                        cr->rel_line_to(0.0, 1.0);
                                        cr->rel_line_to((1.0/11.0), 0.0);
                                        cr->close_path();
                                        cr->fill();
                                }
                        } else {
                                for (unsigned int i = 0; i < 4; ++i) {
                                        cr->move_to(-1.0 + 0.5 * i, 0.2);
                                        cr->rel_line_to(0.5, 0.5);
                                        cr->rel_move_to(-0.5, 0.0);
                                        cr->rel_line_to(0.5, -0.5);
                                        cr->stroke();
                                }
                        }
                        if (rwy.get_le_disp()) {
                                float rwyend(-nmi_to_pixel(rwy.get_le_disp() * (Point::ft_to_m / 1000.0f * Point::km_to_nmi)) / rwyhalfwidth);
                                cr->save();
                                cr->move_to(-1.0, rwyend);
                                cr->rel_line_to(2.0, 0.0);
                                cr->line_to(1.0, 0.0);
                                cr->rel_line_to(-2.0, 0.0);
                                cr->close_path();
                                cr->clip();
                                for (float curp = -0.4; curp >= rwyend; curp -= 1.4) {
                                        cr->move_to(-0.8, curp-1.6);
                                        cr->rel_line_to(0.8, 1.6);
                                        cr->rel_line_to(0.8, -1.6);
                                        cr->stroke();
                                }
                                cr->restore();
                        }
                        if (!rwy.get_ident_le().empty()) {
                                cr->select_font_face("Sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
                                cr->set_font_size(1);
                                Cairo::TextExtents ext;
                                cr->get_text_extents(rwy.get_ident_le(), ext);
                                if (ext.width > 0.0) {
                                        cr->set_font_size(1.8 / ext.width);
                                        cr->save();
                                        cr->translate(0.9, 1.5);
                                        cr->rotate(M_PI);
                                        cr->move_to(0.0, 0.0);
                                        cr->show_text(rwy.get_ident_le());
                                        cr->restore();
                                }
                        }
                        cr->restore();
                }
                // HE headend
                {
                        cr->save();
                        {
                                Point ptcenter(rwy.get_he_coord());
                                if (rwy.get_he_disp())
                                        ptcenter = ptcenter.spheric_course_distance_nmi(rwy.get_he_hdg() * FPlanLeg::to_deg,
                                                                                        rwy.get_he_disp() * (Point::ft_to_m * (1.0f/1000.f) * Point::km_to_nmi));
                                ScreenCoordFloat pxcenter(current_cftrans(ptcenter));
                                cr->translate(pxcenter.getx(), pxcenter.gety());
                        }
                        cr->rotate((rwy.get_he_hdg() + 0x8000) * FPlanLeg::to_rad);
                        //std::cerr << "scale: " << rwyhalfwidth << " width(nmi): " << (rwy.get_width() * (Point::ft_to_m / 2000.0f * Point::km_to_nmi)) << std::endl;
                        cr->scale(rwyhalfwidth, rwyhalfwidth);
                        cr->set_line_width(0.2);
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->move_to(-1.0, 0.0);
                        cr->line_to(1.0, 0.0);
                        cr->stroke();
                        if (rwy.is_he_usable()) {
                                for (unsigned int i = 0; i < 5; ++i) {
                                        cr->move_to((1.0/11.0) * (1 + 2 * i), 0.2);
                                        cr->rel_line_to((1.0/11.0), 0.0);
                                        cr->rel_line_to(0.0, 1.0);
                                        cr->rel_line_to(-(1.0/11.0), 0.0);
                                        cr->close_path();
                                        cr->fill();
                                        cr->move_to(-(1.0/11.0) * (1 + 2 * i), 0.2);
                                        cr->rel_line_to(-(1.0/11.0), 0.0);
                                        cr->rel_line_to(0.0, 1.0);
                                        cr->rel_line_to((1.0/11.0), 0.0);
                                        cr->close_path();
                                        cr->fill();
                                }
                        } else {
                                for (unsigned int i = 0; i < 4; ++i) {
                                        cr->move_to(-1.0 + 0.5 * i, 0.2);
                                        cr->rel_line_to(0.5, 0.5);
                                        cr->rel_move_to(-0.5, 0.0);
                                        cr->rel_line_to(0.5, -0.5);
                                        cr->stroke();
                                }
                        }
                        if (rwy.get_he_disp()) {
                                float rwyend(-nmi_to_pixel(rwy.get_he_disp() * (Point::ft_to_m / 1000.0f * Point::km_to_nmi)) / rwyhalfwidth);
                                cr->save();
                                cr->move_to(-1.0, rwyend);
                                cr->rel_line_to(2.0, 0.0);
                                cr->line_to(1.0, 0.0);
                                cr->rel_line_to(-2.0, 0.0);
                                cr->close_path();
                                cr->clip();
                                for (float curp = -0.4; curp >= rwyend; curp -= 1.4) {
                                        cr->move_to(-0.8, curp-1.6);
                                        cr->rel_line_to(0.8, 1.6);
                                        cr->rel_line_to(0.8, -1.6);
                                        cr->stroke();
                                }
                                cr->restore();
                        }
                        if (!rwy.get_ident_he().empty()) {
                                cr->select_font_face("Sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
                                cr->set_font_size(1);
                                Cairo::TextExtents ext;
                                cr->get_text_extents(rwy.get_ident_he(), ext);
                                if (ext.width > 0.0) {
                                        cr->set_font_size(1.8 / ext.width);
                                        cr->save();
                                        cr->translate(0.9, 1.5);
                                        cr->rotate(M_PI);
                                        cr->move_to(0.0, 0.0);
                                        cr->show_text(rwy.get_ident_he());
                                        cr->restore();
                                }
                        }
                        cr->restore();
                }
                // draw centerline
                {
                        float halfwidth(rwy.get_width() * (0.5f * Point::ft_to_m * (1.0f/1000.f) * Point::km_to_nmi));
                        Point pt1(rwy.get_le_coord().spheric_course_distance_nmi(rwy.get_le_hdg() * FPlanLeg::to_deg,
                                  (rwy.get_le_disp() + 2 * rwy.get_width()) * (Point::ft_to_m * (1.0f/1000.f) * Point::km_to_nmi)));
                        Point pt2(rwy.get_he_coord().spheric_course_distance_nmi(rwy.get_he_hdg() * FPlanLeg::to_deg,
                                  (rwy.get_he_disp() + 2 * rwy.get_width()) * (Point::ft_to_m * (1.0f/1000.f) * Point::km_to_nmi)));
                        ScreenCoordFloat px1(current_cftrans(pt1));
                        ScreenCoordFloat px2(current_cftrans(pt2));
                        cr->set_line_width(0.05 * rwyhalfwidth);
                        std::vector<double> dashes;
                        dashes.push_back(rwyhalfwidth);
                        cr->set_dash(dashes, 0.0);
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->move_to(px1.getx(), px1.gety());
                        cr->line_to(px2.getx(), px2.gety());
                        cr->stroke();
                        dashes.clear();
                        cr->set_dash(dashes, 0.0);
                }
        }
        // draw line features
        cr->set_line_width(1.0);
        // second pass: line features
        for (unsigned int r = 0; r < arpt.get_nrlinefeatures(); ++r) {
                const AirportsDb::Airport::Polyline& pl(arpt.get_linefeature(r));
                if (pl.is_area())
                        continue;
                if (pl.is_airportboundary()) {
                        polyline_to_path(cr, pl);
                        cr->set_source_rgb(0.0, 0.0, 0.0);
                        cr->stroke();
                }
        }
        // draw heliports
        cr->set_line_width(1.0);
        cr->set_source_rgb(0.0, 0.0, 0.0);
        for (unsigned int r = 0; r < arpt.get_nrhelipads(); ++r) {
                const AirportsDb::Airport::Helipad& hp(arpt.get_helipad(r));
                ScreenCoordFloat px(current_cftrans(hp.get_coord()));
                double r1(nmi_to_pixel(hp.get_width() * (Point::ft_to_m * 0.0005f * Point::km_to_nmi)));
                double r2(nmi_to_pixel(hp.get_length() * (Point::ft_to_m * 0.0005f * Point::km_to_nmi)));
                cr->save();
                cr->translate(px.getx(), px.gety());
                cr->scale(r1, r2);
                cr->begin_new_path();
                cr->arc(0.0, 0.0, 1.0, 0.0, 2.0 * M_PI);
                if (hp.is_concrete())
                        cr->set_source_rgba(192.0/256.0, 192.0/256.0, 192.0/256.0, 1.0);
                else
                        cr->set_source_rgba(0.0/256.0, 150.0/256.0, 0.0/256.0, 1.0);
                cr->fill_preserve();
                cr->set_source_rgb(0.0, 0.0, 0.0);
                cr->set_line_width(0.1);
                cr->stroke();
                cr->restore();
                cr->select_font_face("Sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
                cr->set_font_size(100.0);
                Cairo::TextExtents ext;
                cr->get_text_extents(hp.get_ident(), ext);
                double sc = std::min(r1 / ext.width, r2 / ext.height);
                //std::cerr << "Helipad: Font: r1 " << r1 << " r2 " << r2 << " W " << ext.width << " H " << ext.height << " sc " << sc << std::endl;
                cr->set_font_size(sc * 100.0);
                cr->get_text_extents(hp.get_ident(), ext);
                cr->move_to(px.getx() - 0.5 * ext.width, px.gety() + 0.5 * ext.height);
                cr->show_text(hp.get_ident());
        }
        // draw ARP
        {
                cr->save();
                ScreenCoordFloat px(current_cftrans(arpt.get_coord()));
                cr->translate(px.getx(), px.gety());
                cr->scale(featurescale, featurescale);
                cr->set_line_width(1.0);
                cr->set_source_rgb(0.0, 0.0, 0.0);
                cr->move_to(-10, 0);
                cr->line_to(+10, 0);
                cr->move_to(0, -10);
                cr->line_to(0, +10);
                cr->stroke();
                cr->arc(0, 0, 5.0, 0.0, 2.0 * M_PI);
                cr->stroke();
                cr->restore();
        }
        // draw point features
        cr->set_line_width(1.0);
        cr->select_font_face("Sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
        cr->set_font_size(12);
        for (unsigned int r = 0; r < arpt.get_nrpointfeatures(); ++r) {
                const AirportsDb::Airport::PointFeature& pf(arpt.get_pointfeature(r));
                switch (pf.get_feature()) {
                        case AirportsDb::Airport::PointFeature::feature_windsock:
                        {
                                ScreenCoordFloat px(current_cftrans(pf.get_coord()));
                                cr->save();
                                cr->set_source_rgb(0.0, 0.0, 0.0);
                                cr->set_line_width(0.25);
                                cr->translate(px.getx(), px.gety());
                                cr->scale(featurescale, featurescale);
                                cr->move_to(0.0, -1.0);
                                cr->line_to(-2.0, 0.0);
                                cr->line_to(2.0, 0.0);
                                cr->close_path();
                                cr->move_to(0.0, -10.0);
                                cr->line_to(4.0, -11.0);
                                cr->line_to(4.0, -19.0);
                                cr->line_to(0.0, -20.0);
                                cr->close_path();
                                cr->move_to(8.0, -12.0);
                                cr->line_to(12.0, -13.0);
                                cr->line_to(12.0, -17.0);
                                cr->line_to(8.0, -18.0);
                                cr->close_path();
                                cr->fill();
                                cr->move_to(0.0, 0.0);
                                cr->line_to(0.0, -20.0);
                                cr->line_to(12.0, -17.0);
                                cr->line_to(12.0, -13.0);
                                cr->line_to(0.0, -10.0);
                                cr->stroke();
                                cr->restore();
                                break;
                        }

                        case AirportsDb::Airport::PointFeature::feature_vasi:
                        {
                                uint16_t hdg(0);
                                //std::cerr << "VASI: RWY Ident " << pf.get_rwyident() << std::endl;
                                for (unsigned int r = 0; r < arpt.get_nrrwy(); ++r) {
                                        const AirportsDb::Airport::Runway& rwy(arpt.get_rwy(r));
                                        if (rwy.get_ident_he() == pf.get_rwyident()) {
                                                hdg = rwy.get_he_hdg();
                                                break;
                                        }
                                        if (rwy.get_ident_le() == pf.get_rwyident()) {
                                                hdg = rwy.get_le_hdg();
                                                break;
                                        }
                                }
                                ScreenCoordFloat px(current_cftrans(pf.get_coord()));
                                cr->save();
                                cr->set_source_rgb(0.0, 0.0, 0.0);
                                cr->translate(px.getx(), px.gety());
                                cr->rotate(hdg * FPlanLeg::to_rad);
                                cr->scale(featurescale, featurescale);
                                cr->arc(-9.0, 0.0, 2.0, 0.0, 2.0 * M_PI);
                                cr->begin_new_sub_path();
                                cr->arc(-3.0, 0.0, 2.0, 0.0, 2.0 * M_PI);
                                cr->begin_new_sub_path();
                                cr->arc(3.0, 0.0, 2.0, 0.0, 2.0 * M_PI);
                                cr->begin_new_sub_path();
                                cr->arc(9.0, 0.0, 2.0, 0.0, 2.0 * M_PI);
                                cr->fill();
                                cr->restore();
                                break;
                        }

                        case AirportsDb::Airport::PointFeature::feature_tower:
                        {
                                static const float starx[5] = {
                                        0.0000,
                                        -9.5106,
                                        -5.8779,
                                        5.8779,
                                        9.5106
                                };
                                static const float stary[5] = {
                                        10.0000,
                                        3.0902,
                                        -8.0902,
                                        -8.0902,
                                        3.0902
                                };
                                ScreenCoordFloat px(current_cftrans(pf.get_coord()));
                                cr->save();
                                cr->set_source_rgb(0.0, 0.0, 0.0);
                                cr->set_line_width(1.0);
                                cr->translate(px.getx(), px.gety());
                                cr->rotate(pf.get_hdg() * FPlanLeg::to_rad);
                                cr->scale(featurescale, featurescale);
                                cr->arc(0.0, 0.0, 10.0, 0.0, 2.0 * M_PI);
                                cr->close_path();
                                cr->move_to(starx[0], -stary[0]);
                                for (unsigned int i = 0;;) {
                                        i += 2;
                                        if (i >= 5)
                                                i -= 5;
                                        cr->line_to(starx[i], -stary[i]);
                                        if (!i)
                                                break;
                                }
                                cr->close_path();
                                cr->stroke();
                                cr->restore();
                                break;
                        }

                        case AirportsDb::Airport::PointFeature::feature_taxiwaysign:
                        {
                                ScreenCoordFloat px(current_cftrans(pf.get_coord()));
                                cr->save();
                                cr->set_source_rgb(0.0, 0.0, 0.0);
                                cr->translate(px.getx(), px.gety());
                                cr->rotate(pf.get_hdg() * FPlanLeg::to_rad);
                                float sc(ldexpf(featurescale, -1));
                                cr->scale(sc, sc);
                                draw_taxiwaysign(cr, pf.get_name());
                                cr->restore();
                                break;
                        }

                        case AirportsDb::Airport::PointFeature::feature_startuploc:
                        {
                                ScreenCoordFloat px(current_cftrans(pf.get_coord()));
                                cr->save();
                                cr->set_source_rgb(0.0, 0.0, 0.0);
                                cr->translate(px.getx(), px.gety());
                                cr->rotate(pf.get_hdg() * FPlanLeg::to_rad);
                                float sc(ldexpf(featurescale, -1));
                                cr->scale(sc, sc);
                                draw_label(cr, pf.get_name(), ScreenCoord(0, 0), AirportsDb::Airport::label_center, 0);
                                cr->restore();
                                break;
                        }

                        default:
                                break;
                }
        }
        draw_airport_vfrreppt(cr, arpt);
        cr->restore();
}

void VectorMapRenderer::DrawThreadAirportDiagram::set_map_scale(float nmi_per_pixel)
{
        m_invmapscale = std::max(0.01f/20, std::min(100.0f/20, nmi_per_pixel)) * (Point::from_deg / 60.0);
        m_mapscale = 1.0 / m_invmapscale;
        update_mapscalelon();
        m_mapscalechange = true;
        restart();
}

void VectorMapRenderer::DrawThreadAirportDiagram::draw(Cairo::Context * cr, const NavaidsDb::Navaid & n)
{
        cr->save();
        cr->set_source_rgb(0.0, 0.0, 0.0);
        cr->set_line_width(1.0);
        cr->set_fill_rule(Cairo::FILL_RULE_EVEN_ODD);
        ScreenCoordFloat pos(current_cftrans(n.get_coord()));
        cr->translate(pos.getx(), pos.gety());
        float featurescale(nmi_to_pixel(Point::km_to_nmi * 0.002));
        cr->scale(featurescale, featurescale);
        draw_navaid(cr, n, ScreenCoordFloat());
        cr->restore();
}

void VectorMapRenderer::DrawThreadAirportDiagram::draw(Cairo::Context * cr, const WaypointsDb::Waypoint & w)
{
        draw_waypoint(cr, w.get_coord(), w.get_name(), true, w.get_label_placement());
}

void VectorMapRenderer::DrawThreadAirportDiagram::draw_taxiwaysign(Cairo::Context * cr, const Glib::ustring & text)
{
        std::vector<LabelPart> lbl;
        lbl.push_back(LabelPart());
        for (Glib::ustring::size_type i = 0; i < text.size();) {
                {
                        Glib::ustring::value_type ch(text[i]);
                        if (ch != '{') {
                                lbl.back().push_back(ch);
                                ++i;
                                continue;
                        }
                }
                ++i;
                // in glyph mode, find glyphs separated by comma or terminated by right curly brace
                while (i < text.size()) {
                        Glib::ustring::size_type j = text.find_first_of(",}", i);
                        Glib::ustring glyph(text, i, (j == Glib::ustring::npos) ? j : (j - i));
                        i = j;
                        if (glyph == "@Y") {
                                // (black lettering on yellow - the most common taxiway sign).
                                if (!lbl.back().empty())
                                        lbl.push_back(LabelPart());
                                lbl.back().set_fg(0.0, 0.0, 0.0);
                                lbl.back().set_bg(1.0, 1.0, 0.0);
                        } else if (glyph == "@R") {
                                // (white lettering on a black background - usually a mandatory sign, such as warning of a runway crossing).
                                if (!lbl.back().empty())
                                        lbl.push_back(LabelPart());
                                lbl.back().set_fg(1.0, 1.0, 1.0);
                                lbl.back().set_bg(0.0, 0.0, 0.0);
                        } else if (glyph == "@L") {
                                // (yellow lettering on a black background - informs you of your current location).
                                if (!lbl.back().empty())
                                        lbl.push_back(LabelPart());
                                lbl.back().set_fg(1.0, 1.0, 0.0);
                                lbl.back().set_bg(0.0, 0.0, 0.0);
                        } else if (glyph == "@B") {
                                // (white lettering on a black background - used for runway 'distance-remaining' signs).
                                if (!lbl.back().empty())
                                        lbl.push_back(LabelPart());
                                lbl.back().set_fg(1.0, 1.0, 1.0);
                                lbl.back().set_bg(0.0, 0.0, 0.0);
                        } else if (glyph == "@@") {
                                // backside - just ignore
                        } else if (glyph == "_") {
                                lbl.back().push_back(' ');
                        } else if (glyph == "*") {
                                lbl.back().push_back('.');
                        } else if (glyph == "|") {
                                // frame
                                lbl.back().push_back('|');
                        } else if (glyph == "^u") {
                                // up arrow
                                lbl.back().push_back((gunichar)0x2191);
                        } else if (glyph == "^d") {
                                // down arrow
                                lbl.back().push_back((gunichar)0x2193);
                        } else if (glyph == "^l") {
                                // left arrow
                                lbl.back().push_back((gunichar)0x2190);
                        } else if (glyph == "^r") {
                                // right arrow
                                lbl.back().push_back((gunichar)0x2192);
                        } else if (glyph == "^lu") {
                                lbl.back().push_back((gunichar)0x2196);
                        } else if (glyph == "^ru") {
                                lbl.back().push_back((gunichar)0x2197);
                        } else if (glyph == "^ld") {
                                lbl.back().push_back((gunichar)0x2199);
                        } else if (glyph == "^rd") {
                                lbl.back().push_back((gunichar)0x2198);
                        } else if (glyph == "no-entry") {
                                lbl.back().append("no entry");
                        } else {
                                lbl.back().append(glyph);
                        }
                        if (i == Glib::ustring::npos || i >= text.size())
                                break;
                        if (text[i] == '}') {
                                ++i;
                                break;
                        }
                        ++i;
                }
        }
        double toth(0), totw(0);
        for (std::vector<LabelPart>::const_iterator li(lbl.begin()), le(lbl.end()); li != le; ++li) {
                Cairo::TextExtents ext;
                cr->get_text_extents(*li, ext);
                totw += ext.width + 1.0;
                toth = std::max(toth, ext.height);
        }
        double posw(-0.5 * totw), posh(-0.5 * toth);
        for (std::vector<LabelPart>::const_iterator li(lbl.begin()), le(lbl.end()); li != le; ++li) {
                Cairo::TextExtents ext;
                cr->get_text_extents(*li, ext);
                li->set_bg(cr);
                cr->rectangle(posw, posh - 1.0, ext.width + 1.0, toth + 2.0);
                cr->fill();
                li->set_fg(cr);
                cr->move_to(posw, -posh);
                cr->show_text(*li);
                posw += ext.width + 1.0;
        }
}

