/*
 * updategndelev.cc:  Compute ground elevations from TopoDb data
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* AIX requires this to be the first thing in the file.  */
#ifndef __GNUC__
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  ifdef _AIX
#pragma alloca
#  else
#   ifndef alloca /* predefined by HP cc +Olibcalls */
char *alloca ();
#   endif
#  endif
# endif
#endif

#include <getopt.h>
#include <iostream>

#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#define EX_USAGE   64
#define EX_DATAERR 65
#define EX_OK      0
#endif

#include "geom.h"
#include "dbobj.h"

/* ---------------------------------------------------------------------- */

static bool update_airspaces(AirspacesDb::Airspace& e, TopoDb30& topodb)
{
        if (e.get_gndelevmin() != AirspacesDb::Airspace::gndelev_unknown &&
            e.get_gndelevmax() != AirspacesDb::Airspace::gndelev_unknown)
                return false;
        TopoDb30::minmax_elev_t mme(topodb.get_minmax_elev(e.get_polygon()));
        if (e.get_gndelevmin() == AirspacesDb::Airspace::gndelev_unknown && mme.first != TopoDb30::nodata)
                e.set_gndelevmin(Point::round<AirspacesDb::Airspace::gndelev_t,float>(mme.first * Point::m_to_ft));
        if (e.get_gndelevmax() == AirspacesDb::Airspace::gndelev_unknown && mme.second != TopoDb30::nodata)
                e.set_gndelevmax(Point::round<AirspacesDb::Airspace::gndelev_t,float>(mme.second * Point::m_to_ft));
        return true;
}

/* ---------------------------------------------------------------------- */

static void process_airspaces(const Glib::ustring& output_dir)
{
        TopoDb30 topodb;
        topodb.open(output_dir);
        topodb.sync_off();
        AirspacesDb db;
        db.open(output_dir);
        db.sync_off();
        AirspacesDb::Airspace e;
        db.loadfirst(e);
        while (e.is_valid()) {
                if (update_airspaces(e, topodb)) {
                        std::cerr << "Saving Airspace " << e.get_icao() << ' ' << e.get_name() << ' ' << e.get_ident() << ' ' << e.get_sourceid()
                                  << " elev min: " << e.get_gndelevmin() << " max: " << e.get_gndelevmax() << std::endl;
                        db.save(e);
                }
                db.loadnext(e);
        }
        db.analyze();
        db.vacuum();
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
        static struct option long_options[] = {
                { "help",               no_argument,       NULL, 'h' },
                { "version",            no_argument,       NULL, 'v' },
                { "quiet",              no_argument,       NULL, 'q' },
                { "output-dir",         required_argument, NULL, 'o' },
                { NULL,                 0,                 NULL, 0 }
        };
        int c, err(0);
        bool quiet(false);
        Glib::ustring output_dir(".");

        while ((c = getopt_long(argc, argv, "hvqo:", long_options, NULL)) != -1) {
                switch (c) {
                        case 'v':
                                std::cout << argv[0] << ": (C) 2007 Thomas Sailer" << std::endl;
                                return 0;

                        case 'o':
                                output_dir = optarg;
                                break;

                        case 'q':
                                quiet = true;
                                break;

                        case 'h':
                        default:
                                err++;
                                break;
                }
        }
        if (err) {
                std::cerr << "usage: updategndelev [-o <dir>]" << std::endl
                          << "     -q                do not prompt for HotSync button press" << std::endl
                          << "     -h, --help        Display this information" << std::endl
                          << "     -v, --version     Display version information" << std::endl << std::endl;
                return EX_USAGE;
        }
        try {
                process_airspaces(output_dir);
        } catch (const std::exception& e) {
                std::cerr << "Error: " << e.what() << std::endl;
                return EX_DATAERR;
        } catch (const Glib::Exception& e) {
                std::cerr << "Glib Error: " << e.what() << std::endl;
                return EX_DATAERR;
        }
        return 0;
}
